#include "coord.h"
#include "face.h"
#include "enum.h"
#include "assert.h"
#include <stdlib.h>

//----------------------------------------------------------------------------
//   class ShapeRef 
//----------------------------------------------------------------------------

//---ShapeRef::ShapeRef-------------------------------------------------------
// Input : Shape *ashape     : init
//         ShapeRef *anext   : init
// Output:

ShapeRef::ShapeRef(Shape *ashape, ShapeRef *anext)
{
    shape=ashape;
    next=anext;
}

//----------------------------------------------------------------------------
//   class Coord : Data
//----------------------------------------------------------------------------

//---Coord::Coord-------------------------------------------------------------
// Input : float ax, ay, az  : init
//         int alod          : init
// Output:

Coord::Coord(float ax, float ay, float az, int alod) : Data()
{
    x = ax;
    y = ay;
    z = az;
    represent = NULL;
    selected = NULL;
    lod = alod;
    count = 0;
    shapeRef = NULL;
}

//---Coord::~Coord--------------------------------------------------------------
// Input :
// Output:

Coord::~Coord()
{
    deleteShapeRefs();
}

//---Coord::print-------------------------------------------------------------
// Input : FILE *fd          : VRML output file descriptor
//         int fid           : field id
// Output:
//write x y z separated by blanks to fd
//increases global variable global_cpl at 3. This enables the method
//DataSet::printSet making carriage return s if global_cpl >= global_floatcpl

void Coord::print(FILE *fd, int fid)
{
    printFloat(fd,x); fprintf(fd," ");
    printFloat(fd,y); fprintf(fd," ");
    printFloat(fd,z);
    global_cpl+=3;
}

//---Coord::deleteShapeRefs---------------------------------------------------
// Input :
// Output:
//delete the list of references to shapes
//which uses this coordinate

void Coord::deleteShapeRefs()
{
    ShapeRef *help;

    while (shapeRef)
    {
        help=shapeRef;
        shapeRef=shapeRef->next;
        delete help;
    }
}

//---Coord::addShapeRef-------------------------------------------------------
// Input :
// Output:
//add a new reference to shape to the list

void Coord::addShapeRef(Shape *f)
{
    ShapeRef *frh = new ShapeRef(f,shapeRef);
    shapeRef = frh;
}

//---Coord::resetShapeDoneFlags-----------------------------------------------
// Input :
// Output:
//some algorithms e.g. OctreeCoordSet::calcError uses the shapes in the
//shapeRef list - to prevent accessing some shapes multiple these algorithms
//mark the shapes by setting a done-flag - this function resets this flags

void Coord::resetShapeDoneFlags()
{
    ShapeRef *srh = shapeRef;
    while (srh)
    {
        srh->shape->done = 0;
        srh=srh->next;
    }
}

//---Coord::getRepresent------------------------------------------------------
// Input : int targetlod       : LOD
// Output: Coord *             : representative coordinate
//get the representative of this coordinate at Level of Detail targetlod

Coord *Coord::getRepresent(int targetlod)
{
    Coord *ch = this;
    Coord *r;

    r=ch->represent;
    while (r && r->lod>=0 && r->lod<=targetlod)
    {
        ch=r;
        r=ch->represent;
    } 
    return ch;
}

//---Coord::getSelected-------------------------------------------------------
// Input : int targetlod       : LOD 
// Output: Coord *             : selected coordinate
//get the coord which was selected for the representant at the
//Level of Detail with number targetlod

Coord *Coord::getSelected(int targetlod)
{
   Coord *ch = this;
   Coord *s;

   s=ch->selected;
   while (s && s->lod>=targetlod)
   {
       ch=s;
       s=ch->selected;
   }
   return ch;
}

//----------------------------------------------------------------------------
//   class OctreeCoord : Coord : Data
//----------------------------------------------------------------------------

//---OctreeCoord::OctreeCoord-------------------------------------------------
// Input : float ax, ay, az    : init
//         int alod            : init
// Output:

OctreeCoord::OctreeCoord(float ax, float ay, float az, int alod) :
             Coord(ax,ay,az,alod)
{
    int i;
    for (i=0; i<8; i++) succ[i]=NULL;
    depth=1;
}

//----------------------------------------------------------------------------
//   class CoordSet : DataSet
//----------------------------------------------------------------------------

//---CoordSet::CoordSet-------------------------------------------------------
// Input :
// Output:

float CoordSet::errorRange = 2.0;
float CoordSet::cameraHeightAngle = 0.785398;
float CoordSet::minDiff = 0.2;

CoordSet::CoordSet() : DataSet()
{
    bounds = Bounds(0,0,0, 0,0,0); 
    currentlod = 0;
}

//---CoordSet::insert---------------------------------------------------------
// Input : Coord **c            : new coordinate
// Output:

void CoordSet::insert(Coord **c)
{
    //virtual - nothing to do in this class
}

//---CoordSet::print----------------------------------------------------------
// Input : FILE *fd             : VRML output file descriptor
//         int indent           : indent for file
// Output:

void CoordSet::print(FILE *fd, int indent)
{
    //virtual - nothing to do in this class
}

//---CoordSet::setBounds------------------------------------------------------
// Input : Bounds b             : new bounds
// Output:
//before any coord can be inserted this function must be called
//b must contain the bounds of all coordinates you want to insert
//the method Bounds::makeCubic is called to ensure that the elongation
//in all directions is the same 

void CoordSet::setBounds(Bounds b)
{
    bounds = b;
    bounds.makeCubic();
}

//---CoordSet::setShapeType---------------------------------------------------
// Input : int type             : new type
// Output:
//before any lod can be generated this function must be called to
//determine the type of the shapes (faces or lines) which uses the
//coordinates in this object

void CoordSet::setShapeType(int type)
{
    shapeType = type;
}

//---CoordSet::getCurrentLOD--------------------------------------------------
// Input :
// Output:
//some functions returns different results depending on the current active
//lod - this function returns the number the lod which is currently active

int CoordSet::getCurrentLOD()
{
    return currentlod;
}

//---CoordSet::generateNextLOD------------------------------------------------
// Input :
// Output:

void CoordSet::generateNextLOD()
{
    //virtual - nothing to do in this class
}

//---CoordSet::getRange-------------------------------------------------------
// Input :
// Output:

float CoordSet::getRange()
{
    return 0;    //virtual - nothing to do in this class
}

//---CoordSet::deleteShapeRefs------------------------------------------------
// Input :
// Output:

void CoordSet::deleteShapeRefs()
{
    //virtual - nothing to do in this class
}

//---CoordSet::resetLODs------------------------------------------------------
// Input :
// Output:

void CoordSet::resetLODs()
{
    //virtual - nothing to do in this class
}

//---CoordSet::getCountUsed---------------------------------------------------
// Input :
// Output:

int CoordSet::getCountUsed()
{
    return 0;   //virtual - nothing to do in this class
}

//----------------------------------------------------------------------------
//   class OctreeCoordSet : CoordSet : DataSet
//----------------------------------------------------------------------------

//---OctreeCoordSet::OctreeCoordSet-------------------------------------------
// Input :
// Output:

float OctreeCoordSet::weightAlgErrorArea    = 50;
float OctreeCoordSet::weightAlgErrorVolume  = 30;
float OctreeCoordSet::weightAlgWeightedMean = 20;

OctreeCoordSet::OctreeCoordSet() : CoordSet()
{
    root = NULL;
    icount = 0;
}

//---OctreeCoordSet::~OctreeCoordSet------------------------------------------
// Input :
// Output:

OctreeCoordSet::~OctreeCoordSet()
{
    deleteOctree(root);
}

//---OctreeCoordSet::insert---------------------------------------------------
// Input : Coord **c          : new coordinate
// Output:
//insert c in the OctreeCoordSet. c is inserted internally in an octree with
//adaptive depth using insertOctree. if an equal coord already exists c will
//be deleted and the pointer to c will be redirected to the existing cooord
//additionally a reference to c will be appended to DataSet::refArray
//NOTE: c must be from type OctreeCoord !!

void OctreeCoordSet::insert(Coord **c)
{
    insertOctree((OctreeCoord **)c,&root,bounds); 
    insertArray(*c);       
}

//---OctreeCoordSet::gerateNextLOD--------------------------------------------
// Input :
// Output:
//generate the next lod - all lods are stored in the same data structure
//the function calls generateOctreeLODs until the depth of the tree has
//changed and the number of the coords is decreased by a specified amount
//(decreased at minDiff*count of old lod)

void OctreeCoordSet::generateNextLOD()
{
    if (root==NULL) return;

    calcOctreeDepthOfUsedPoints(root);
    int olddepth = root->depth;
    if (olddepth<=1) return;

    int oldcount = calcOctreeCountOfUsedPoints(root);
    int diffcount = (int)((float)oldcount*minDiff);
    if (diffcount==0) diffcount=1;
    int maxcount = oldcount-diffcount;
    if (maxcount<=0) maxcount=1;

    int newcount = oldcount;

    while (root->depth==olddepth || newcount>maxcount)
        generateOctreeLODs(root,newcount);

    currentlod++;
}

//---OctreeCoordSet::getRange-------------------------------------------------
// Input : none
// Output: float  : range
//get the maximum distance to the camera for lod-1 before
//switching to this lod - this may be confusing: set the range of
//an FaceSet(LineSet) for lod n to getRange() of an OctreeCoordSet at
//state lod n+1 !
//
//you can switch to this lod if the diagonal of the nodes (which can
//be assumed as cubes) at root->depth projected into 2D is less than a specified
//value (less than errorRange percent of the height of the viewing window)
//
//the diagonal is: bounds.getMaxElongation() * sqrt(3) / pow(2,root->depth-1)
//                 (the elongations of the cube are all the same (makeCubic))
//the calculation of the range is made by:
//                 diagonal / (2*tan(cameraHeightAngle/2)*errorRange/100)

float OctreeCoordSet::getRange()
{
    if (root==NULL) return 0;

    float h = tan(cameraHeightAngle/2)*2;
    if (h==0) 
    {
        fprintf(stderr,"error: illegal cameraheightangle >%f<\n",cameraHeightAngle);
        exit(1);
    }
    float fac = h*errorRange/100;
    if (fac==0)
    {
        fprintf(stderr,"error: can't calculate lod-rangefactor\n\
       with heightAngle >%f< and errorrange >%f<\n",cameraHeightAngle,errorRange);
        exit(1);
    }

    return bounds.getMaxElongation() * sqrt(3) / pow(2,root->depth-1) / fac;
}

//---OctreeCoordSet::print----------------------------------------------------
// Input : FILE *fd         : VRML output file descriptor
//         int indent       : indent for file
// Output:
//writes "point [ ... ]" to fd where ... are the coordinates separated by
//commas. These coordinates are printed using DataSet::printSet
//additionally DataSet::setIndices is called

void OctreeCoordSet::print(FILE *fd, int indent)
{
    setIndices();

    printIndent(fd,indent);
    fprintf(fd,"%s [ ",fieldIdString[FID_POINT]);
    printSet(fd,indent+8,0);
    fprintf(fd," ]\n");
}

//---OctreeCoordSet::deleteShapeRefs------------------------------------------
// Input :
// Output:
//delete the references to shapes of all inserted coordinates

void OctreeCoordSet::deleteShapeRefs()
{
    deleteOctreeShapeRefs(root);
}

//---OctreeCoordSet::resetLODs------------------------------------------------
// Input :
// Output:
//reset data structures to lod 0 - all pseudo leaf-nodes will be converted
//back to intermediate-nodes

void OctreeCoordSet::resetLODs()
{
    resetOctreeLODs(root);
    calcOctreeDepthOfUsedPoints(root);
    calcOctreeCountOfUsedPoints(root);
    currentlod=0;
}

//---OctreeCoordSet::getCountUsed---------------------------------------------
// Input :
// Output:
//get the count of the coordinates which are used by shapes
//in the current lod

int OctreeCoordSet::getCountUsed()
{
    calcOctreeCountOfUsedPoints(root);
    return root->count;
}

//---OctreeCoordSet::insertOctree---------------------------------------------
// Input : OctreeCoord **c      : insert octree
//         OctreeCoord **root   : octree root
//         Bounds b             : bounds
// Output:
//insert c into the octree specified by root if the tree is not deep
//enought new intermediate nodes are created
//algorithm: if root is null - ok insert
//           if root is an intermediate-node search again recursively
//           if root is a leaf-node -> tree is not deep enought - convert
//               node into an intermediate-node and append old leaf-node
//               and c to the intermediate-node

void OctreeCoordSet::insertOctree(OctreeCoord **c,
                                  OctreeCoord **root, Bounds b)
{
    OctreeCoord *r = *root;
    OctreeCoord *h = *c;

    if (r == NULL)
    {
        //found a place to insert

        *root = *c;
        count++;
        return;
    }
    
    if (r->lod < 0)
    {
        //found an intermediate-node - search again

        int i;
        if (h->x <= r->x) { b.xmax = r->x; i=0; }
                     else { b.xmin = r->x; i=4; }
        if (h->y <= r->y) { b.ymax = r->y; }
                     else { b.ymin = r->y; i+=2; }
        if (h->z <= r->z) { b.zmax = r->z; }
                     else { b.zmin = r->z; i++; }

        insertOctree(c, &(r->succ[i]) ,b);
    }    

    else  //leaf-node reached
 
    {
        if (h->x == r->x && h->y == r->y && h->z == r->z)
        {
            //a node with the coordinates of c already exists - c will be
            //deleted and a pointer to this leaf-node will be returned instead

            delete *c;
            *c = *root;
            return;
        }

        else

        {
            //a leaf-node is reached with coordinates different to c -
            //this means the octree is not deep enough -> a new intermediate-
            //node will be created and placed on the position of the leaf-node
            //then the leaf-node and c will be appended to the new
            //intermediate-node using recursive calls to insertOctree

            OctreeCoord *ocOld = *root;
            *root = new OctreeCoord((b.xmax + b.xmin) / 2,
                                    (b.ymax + b.ymin) / 2,
                                    (b.zmax + b.zmin) / 2, -1);
            icount++;
            count--;                //because ocOld will be inserted again !
            insertOctree(&ocOld,root,b);
            insertOctree(c,root,b);
        }
    }
}

//---OctreeCoordSet::deleteOctree---------------------------------------------
// Input : OctreeCoord *root        : tree root
// Output:
//delete the octree specified by root

void OctreeCoordSet::deleteOctree(OctreeCoord *root)
{
    int i;

    if (root == NULL) return;
    for (i=0; i<8; i++) deleteOctree(root->succ[i]);
    delete root;
}

//---OctreeCoordSet::calcOctreeDepthOfUsedPoints------------------------------
// Input : OctreeCoord * root       : tree root 
// Output: int                      : depth of used points
//get the depth of the part of the tree which is determined by the used
//coordinates (which have shapeRefs) in the current lod
//the function also sets the variable OctreeCoord::depth in the nodes
//appropriated to the subtree-depths

int OctreeCoordSet::calcOctreeDepthOfUsedPoints(OctreeCoord *root)
{
    int i;
    int max = 0;
    int d;

    if (root==NULL) return 0;

    if (root->lod<0)
        for (i=0; i<8; i++)
        {
            d=calcOctreeDepthOfUsedPoints(root->succ[i]);
            if (d>max) max=d;
        }    

    if (max>0 || root->shapeRef) max++;    //if intermediate or used
    root->depth = max;
    return max;
}

//---OctreeCoordSet::calcOctreeCountOfUsedPoints------------------------------
// Input : OctreeCoord * root        : tree root
// Output: int                       : count of used points
//get count of the coordinates which are used by shapes in the current lod
//the function also sets the variable Coord::count int the nodes
//appropriated to the counts in the subtrees

int OctreeCoordSet::calcOctreeCountOfUsedPoints(OctreeCoord *root)
{
    int i;
    int c = 0;

    if (root==NULL) return 0;

    if (root->lod<0) 
        for (i=0; i<8; i++) c+=calcOctreeCountOfUsedPoints(root->succ[i]);

    if (c==0 && root->shapeRef) c++;
    root->count = c;
    return c;
}

//---OctreeCoordSet::generateOctreeLODs---------------------------------------
// Input : OctreeCoord *root         : tree root
// Output: int &newcount             : new count
//this method searches for the leaves in the subtree with the greatest
//depth and most coords - this leaves will be combined by selecting an
//representant with method selectOctreeRepresent


void OctreeCoordSet::generateOctreeLODs(OctreeCoord *root, int &newcount)
{
    int d,i;
    OctreeCoord *soc;

    if (root==NULL) return;

    assert(root->lod<0);
    
    //---find out the subnode with the greatest depth and most coords 

    int maxdepth=0;
    int maxcount=0;
    OctreeCoord *socmax=NULL;
    for (i=0; i<8; i++)
    {
        soc = root->succ[i];
        if (soc)
        {
            d=soc->depth;
            if (d>maxdepth) 
            { 
                maxdepth=d; 
                maxcount=soc->count;
                socmax=soc;
            }
            else if (d==maxdepth && soc->count>maxcount)
            { 
                maxcount=soc->count;
                socmax=soc;        
            }
        }
    }
    assert(maxcount>0);

    //---if maxdepth>1 then search again - after return check changed depths

    if (maxdepth>1)
    {
        generateOctreeLODs(socmax,newcount);

        d=socmax->depth;
        if (d<maxdepth) 
        {
            maxdepth=d;
            for (i=0; i<8; i++)
            {
                soc = root->succ[i];
                if (soc) { d=soc->depth; if (d>maxdepth) maxdepth=d; }
            }
            root->depth = maxdepth+1;
        }
    }

    else

    //---change this intermediate-node to a representative node

    {
        int num=0;
        for (i=0; i<8; i++)
        {
            soc = root->succ[i];
            if (soc && soc->count>0)
            {
                num++;
                soc->represent = root;
            }
        }
        root->lod = currentlod+1;

        OctreeCoord *best = selectOctreeRepresent(root);

        root->x = best->x;
        root->y = best->y;
        root->z = best->z;
        root->selected = best;
        root->index = best->index;
        root->depth = 1;
        newcount=newcount-num+1;
    }
}

//---OctreeCoordSet::selectOctreeRepresent------------------------------------
// Input : Octree * root         : tree root
// Output: OctreeCoord *         : sub nodes representant
//selects a representant of the sub-nodes of root (the sub-nodes must be
//leaf-nodes). If the coords are used by faces the method uses a weighted
//meanvalue of the (normalized) errors produced by the algorithms
//errorarea, errorvolume, weightedmean for selecting the representant
//If the coords are used by lines only the algorithm weightedmean is used
//These algorithms are realized in calcError

OctreeCoord *OctreeCoordSet::selectOctreeRepresent(OctreeCoord *root)
{
    int i;
    float err[8],e[8];

    if (shapeType==SHAPE_TYPE_FACE) 
    {
        for (i=0; i<8; i++) err[i]=0;

        if (weightAlgErrorArea>0)
        {
            calcError(root,COORDSET_METHOD_ERRORAREA,e);
            for (i=0; i<8; i++) 
                if (e[i]<0) err[i]=e[i]; else err[i] += e[i]*weightAlgErrorArea;
        }

        if (weightAlgErrorVolume>0)
        {
            calcError(root,COORDSET_METHOD_ERRORVOLUME,e);
            for (i=0; i<8; i++) 
                if (e[i]<0) err[i]=e[i]; else err[i] += e[i]*weightAlgErrorVolume;
        }

        if (weightAlgWeightedMean>0)
        {
            calcError(root,COORDSET_METHOD_WEIGHTEDMEAN,e);
            for (i=0; i<8; i++) 
                if (e[i]<0) err[i]=e[i]; else err[i] += e[i]*weightAlgWeightedMean;
        }
    }
    else calcError(root,COORDSET_METHOD_WEIGHTEDMEAN,err);

    float besterror = -1;
    int bestindex = -1;
    for (i=0; i<8; i++) 
        if (err[i]>=0)
            if (bestindex==-1 || err[i]<besterror) { besterror=err[i]; bestindex=i; }

    assert(bestindex!=-1);
    return root->succ[bestindex];
}

//---OctreeCoordSet::calcError------------------------------------------------
// Input : OctreeCoord *root     : sub tree
//         int method            : algorithm methode
//         float err[8]          : error
// Output:
//calc the errors which will occur if a sub-node is selected for the
//representant. if sub-node x doesn't exist the corresponding err[x] is 
//set to -1 parameter method specifies the algorithm which should be used

void OctreeCoordSet::calcError(OctreeCoord *root, int method, float err[8])
{ 
    int targetlod = currentlod+1;
    OctreeCoord *soc;
    int def[8];                               
    int i,j;

    assert(shapeType==SHAPE_TYPE_FACE || method!=COORDSET_METHOD_ERRORAREA);
    assert(shapeType==SHAPE_TYPE_FACE || method!=COORDSET_METHOD_ERRORVOLUME);

    for (i=0; i<8; i++) 
    {
        soc=root->succ[i];
        if (soc && soc->count>0) def[i]=1; else def[i]=0;
        err[i]=-1;
    }

    //----------------------------------------------------------------------

    switch (method)
    {

        case COORDSET_METHOD_ERRORAREA:
        case COORDSET_METHOD_ERRORVOLUME:
        {
            for (i=0; i<8; i++)
            {
                if (def[i]) 
                {
                    OctreeCoord *selected=root->succ[i];
  
                    root->x = selected->x;   //test this candidat
                    root->y = selected->y;
                    root->z = selected->z;

                    for (j=0; j<8; j++) if (def[j]) root->succ[j]->resetShapeDoneFlags();

                    float error = 0; 
                    for (j=0; j<8; j++)
                    {
                        if (def[j])
                        {
                            ShapeRef *srh = root->succ[j]->shapeRef;
                            while (srh)
                            {
                                Shape *shape = srh->shape;
                                if (shape->done==0)
                                {
                                    switch(method)
                                    {
                                        case COORDSET_METHOD_ERRORVOLUME:
                                            error+=((Face *)shape)->calcErrorVolume(root,targetlod);
                                            break;
                                        case COORDSET_METHOD_ERRORAREA:
                                            error+=((Face *)shape)->calcErrorArea(root,targetlod);
                                            break;
                                        default: assert(0);
                                    }
                                    shape->done=1;
                                }
                                srh=srh->next;
                            }
                        }
                    }
                    err[i] = fabs(error);
                }
            }
        }
        break;

    //----------------------------------------------------------------------

        case COORDSET_METHOD_WEIGHTEDMEAN:
        {
            //---find out the mean value of the sub-points

            int c=0;
            Vec m;
            for (i=0; i<8; i++)
            {
                soc=root->succ[i];
                if (def[i])
                {
                    int csoc=soc->count;
                    m.x+=soc->x*csoc;
                    m.y+=soc->y*csoc;
                    m.z+=soc->z*csoc;
                    c+=csoc;
                }
            }
            m/=c;

            //---find out the sub-point which is nearest to the mean value 

            for (i=0; i<8; i++)
            {
                soc = root->succ[i];
                if (def[i])
                {
                    Vec d;
                    d.x = m.x-soc->x;
                    d.y = m.y-soc->y;
                    d.z = m.z-soc->z;
                    err[i] = d.getLength();
                }
            }
        }
        break;

        default: assert(0);
    }

    normalizeErrors(err);
}

//---OctreeCoordSet::normalizeErrors------------------------------------------
// Input : float err[8]          : error to normalize
// Output:
//scale the errors to [0,1]
//an error<0 is unchanged (undefined)

void OctreeCoordSet::normalizeErrors(float err[8])
{
    float max=0;
    int i;
    for (i=0; i<8; i++) if (err[i]>max) max=err[i];
    for (i=0; i<8; i++) if (err[i]>0) err[i] /= max;
}

//---OctreeCoordSet::deleteOctreeShapeRefs------------------------------------
// Input : OctreeCoord *root       : tree to delete
// Output:
//delete the references to shapes of all coordinates in 
//the octree specified by root

void OctreeCoordSet::deleteOctreeShapeRefs(OctreeCoord *root)
{
    int i;
    if (root==NULL) return;
    for (i=0; i<8; i++) deleteOctreeShapeRefs(root->succ[i]);
    root->deleteShapeRefs();
}

//---OctreeCoordSet::resetOctreeLODs------------------------------------------
// Input : OctreeCoord *root        : tree
// Output:
//reset data structures to lod 0 - all pseudo leaf-nodes will be 
//converted back to intermediate-nodes

void OctreeCoordSet::resetOctreeLODs(OctreeCoord *root)
{
    if (root==NULL) return;

    int i;
    int n=0;
    for (i=0; i<8; i++) 
        if (root->succ[i])
        {
            resetOctreeLODs(root->succ[i]);
            n++;
        }
    if (n==0) root->lod=0; else root->lod=-1;
}

//----------------------------------------------------------------------------

