// ===========================================================================
//  Copyright 1996 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       coord
//  TYPE:       c++ header
//  PROJECT:    lodestar
//  CONTENT:    class CoordSet, class OctreeCoordSet,
//              class Coord, class OctreeCoord,
//              class ShapeRef
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rs      Reinhard Sainitzer
//		hb      Herbert Buchegger
// ===========================================================================
//  HISTORY:
//
// ===========================================================================
//  MANUAL:
//
//  The class OctreeCoordSet is used to store coordinates defined in VRML-
//  Nodes from type Coordinate3. It removes double defined coordinates.
//  Coordinates which will be inserted in this class must be from type
//  OctreeCoord. The coordinates are inserted using the method OctreeCoordSet::
//  insert(Coord **c). ATTENTION: some of the methods in OctreeCoordSet are 
//  virtual and therefore some parameters are from type Coord BUT the parameters
//  MUST be from type OctreeCoord - if not no mercy ! 
//  The coordinates are inserted internally in an octree with adaptive depth.
//  Additionally References to the coordinates are appended to ReferenceArray
//  (see also DataSet) to get the coordinates by index in insertion order
//  The OctreeCoordSet is also used to produce Levels of Detail 
//
//  To produce a higher LOD (less coords) the following alogrithm is applied
//  to all leaf-nodes (whose depths have the depth of the tree) :
//
//  Combine leaf-nodes into parent-nodes. This is made by selecting a
//  representant of the leaf-nodes, setting the data of the parent-node to the
//  data of the representant and converting the parent-node into a quasi
//  leaf-node (if in a node the variable Coord::lod >= 0 the node is supposed
//  to be a leaf-node. It doesn't matter that possibly the node have subnodes -
//  these subnodes will be ignored by the methods)
//  the selection of the representant is made using various algorithms
//  (errorarea, errorvolume, weighted mean)
//
//  The depth of the resulting tree is decreased by one (because the subnodes of 
//  the new quasi leaf-nodes are ignored)
//  if the difference in the count of the coordinates of the new lod and the 
//  old lod is less then minDiff*(count of coordinates old lod) then the
//  algorithm is applied again to the tree
//  The method CoordSet::resetLODs resets the object to lod 0
//
//  NOTE: before you can insert any coordinate you must call OctreeCoordSet::
//        setBounds(Bounds b) where b are the bounds of all coordinates to insert
//
//        before you can generate LODs you must call CoordSet::
//        setShapeType(int type) (see SHAPE_TYPE_xxxxx constants in shape.h)
//        This function tells the OctreeCoordSet which shapes uses the coordinates
//        (lines or faces)
//
//        LODs are created only for the coordinates which are really used by
//        shapes (this coordinates have a list of shapeRefs - this shapeRefs are
//        added by a FaceSet or LineSet automatically)
//
//
//  The class CoordSet is a superclass of OctreeCoordSet it can't be used
//  directly (abstract). It makes only sense if someone wants to implement 
//  an class for coordinates with another datastructur than an octree. 
//
//  SEE ALSO
//      DataSet, Shape (, Face, Line, FaceSet, LineSet)
//
// ===========================================================================

#ifndef _COORD_H_
#define _COORD_H_

#include "global.h"
#include "dataset.h"
#include "shape.h"
#include <stdio.h>

//----------------------------------------------------------------------------
//   constants for parameter method
//----------------------------------------------------------------------------

#define COORDSET_METHOD_ERRORAREA    1
#define COORDSET_METHOD_ERRORVOLUME  2
#define COORDSET_METHOD_WEIGHTEDMEAN 3

//---forward definitions------------------------------------------------------

class Shape;      //defined in shape.h

//----------------------------------------------------------------------------
//simple class for storing references to shapes in a list
//(classes Face and Line are derived from Shape)
//classes derived from Coord can contain a list of these ShapeRefs

class ShapeRef
{
public:

    Shape *shape;      //a reference to object with a type derived from Shape
    ShapeRef *next;    //pointer to the next ShapeRef

    ShapeRef(Shape *ashape, ShapeRef *anext);
};

//----------------------------------------------------------------------------
//Coord can't be used directly (abstract)
//it contains all data and methods which are not related
//to a specific data structure such as an octree

class Coord : public Data
{
public:

    float x;
    float y;
    float z;

    Coord *represent;  //pointer to the representative - this isn't really the 
                       //representative of this coordinate in the next lod
                       //because the node *represent points to can also have
                       //an representative at the same lod
                       //so always use function getRepresent
 
    Coord *selected;   //same as represent in other direction -
                       //points to the subnode which was selected for the
                       //representative - same as above... always use
                       //function getSelected

    int lod;           //number of the lod this Coord is related to

    int count;         //this Coord is representant of count Coords

    ShapeRef *shapeRef;   //pointer to a list of references to shapes
                          //this coordinate is used by this shapes

    Coord(float ax, float ay, float az, int alod);
    virtual ~Coord();

    virtual void print(FILE *fd, int fid);  //prints x y z to fd
                                            //fid is not used

    void deleteShapeRefs();      //delete the list of references to shapes
                                 //which uses this coordinate

    void addShapeRef(Shape *f);  //add a new reference to a shape to the list

    void resetShapeDoneFlags();  //some algorithms e.g. OctreeCoordSet::calcError
                                 //uses the shapes in the shapeRef list - 
                                 //to prevent accessing some shapes multiple
                                 //these algogrithms mark the shapes by setting
                                 //a done-flag - this function resets this flags

    Coord *getRepresent(int targetlod); //get the representative of
                                        //this coord at Level of Detail
                                        //with number targetlod

    Coord *getSelected(int targetlod);  //get the coord which was 
                                        //selected for the representant
                                        //at the Level of Detail
                                        //with number targetlod
};

//----------------------------------------------------------------------------
//used to store a coordinate in an OctreeCoordSet
//Coord::lod<0 means this is an intermediate node

class OctreeCoord : public Coord
{
public:

    int depth;            //depth of the octree with this object as root
    OctreeCoord *succ[8]; //pointer to the 8 subtrees of this OctreeCoord 

    OctreeCoord(float ax, float ay, float az, int alod);
};

//----------------------------------------------------------------------------
//CoordSet can't be used directly (abstract)
//it contains all data and methods which are not related
//to a specific data structure such as an octree

class CoordSet : public DataSet
{
public:

    static float errorRange;           //used by method getRange
                                       //getRange calculates the distance of 
                                       //switching from the current lod to the
                                       //next - the distance is calculated so
                                       //that points projected to the screen
                                       //are misplaced at maximal errorRange %
                                       //of the height of the viewing window

    static float cameraHeightAngle;    //used by method getRange
                                       //heightAngle of the camera used for viewing

    static float minDiff;              //minimum difference between two lods
                                       //the count of a new lod must be less
                                       //than (1-minDiff)*(count of old lod)
                                       //used by generateNextLOD

    CoordSet();

    virtual void insert(Coord **c);            //abstract - no effect
    virtual void print(FILE *fd, int indent);  //abstract - no effect

    void setBounds(Bounds b);      //before you can insert any Coord this 
                                   //function must be called. b must contain
                                   //the bounds of all coordinates you want
                                   //to insert.

    void setShapeType(int type);   //before you can generate any lod this 
                                   //function must be called. This function
                                   //tells this class which shapes uses
                                   //the coordinates (lines or faces)

    int getCurrentLOD();             //some functions returns different results
                                     //depending on the current active lod
                                     //this function returns the number
                                     //of the current active lod

    virtual void generateNextLOD();  //generate the next lod - all lods are
                                     //store in the same data structure
                                     //abstract - no effect

    virtual float getRange();        //abstract - no effect
    virtual void deleteShapeRefs();  //abstract - no effect
    virtual void resetLODs();        //abstract - no effect
    virtual int getCountUsed();      //abstract - no effect
     
protected:
    Bounds bounds;    //bounds of the coordinates - the function setBounds
                      //calls Bounds::makeCubic to make sure that the elongation
                      //in all directions is the same
    int shapeType;    //the type of the shapes (faces or lines) which uses
                      //the coordinates in this object
    int currentlod;   //number of the lod which if currently active
};

//----------------------------------------------------------------------------
//The class OctreeCoordSet is used to store coordinates defined in VRML-
//Nodes from type Coordinate3. It removes double defined coordinates.
//This class is also used to generate lods of coordinates (see top of this file)

class OctreeCoordSet : public CoordSet
{
public:

    static float weightAlgErrorArea;    //weights of the three algorithms
    static float weightAlgErrorVolume;  //used for selecting a representant
    static float weightAlgWeightedMean; //of some leaf-nodes

    int icount;     //number of intermediate nodes in the tree
                    //not affected by lods or whether coords are used or not
                    //only for statistic purposes

    OctreeCoordSet();
    virtual ~OctreeCoordSet();

    virtual void insert(Coord **c);  //insert c in the OctreeCoordSet. c is
                                     //inserted internally in an octree using
                                     //insertOctree. if an equal Coord already
                                     //exists c will be deleted and the pointer
                                     //to c is redirected to the existing coord
                                     //additionally a reference to c will be
                                     //appended to DataSet::refArray
                                     //NOTE: c must be from type OctreeCoord !!

    virtual void print(FILE *fd, int indent); //writes "point [ .. ] to fd where
                                              //... are the coords separated by
                                              //commas. These coords are printed
                                              //using DataSet::printSet
                                              //additionally DataSet::setIndices
                                              //is called

    virtual void generateNextLOD();  //generate the next lod - all lods are
                                     //stored in the same data structure
                                     //some data and functions are affected by
                                     //the current lod. The function 
                                     //getCurrentLOD return the number of the
                                     //current lod. the function resetLODs
                                     //resets the object to lod 0

    virtual float getRange();        //get the maximum distance to the camera
                                     //for the current lod before switching
                                     //to the next lod

    virtual void deleteShapeRefs();  //delete the references to shapes of all
                                     //inserted coordinates

    virtual void resetLODs();        //reset data structures to lod 0
                                     //all pseudo leaf-nodes will be
                                     //converted back to intermediate-nodes
                                     //uses resetOctreeLODs

    virtual int getCountUsed();      //get count of the coordinates which
                                     //are used by shapes in the current lod
                                     //uses calcOctreeCountOfUsedPoints
     
protected:

    OctreeCoord *root;    //root of the octree

    void insertOctree(OctreeCoord **c,    //insert c into the octree specified by root
                      OctreeCoord **root, //if the tree is not deep enought new
                      Bounds b);          //intermediate nodes are created

    void deleteOctree(OctreeCoord *root);  //delete the octree specified by root

    int  calcOctreeDepthOfUsedPoints(OctreeCoord *root); //get the depth of the
                                                         //part of the tree which
                                                         //is determined by the
                                                         //used coordinates and
                                                         //the current lod

    int  calcOctreeCountOfUsedPoints(OctreeCoord *root); //get count of the
                                                         //coordinates which are
                                                         //used by shapes in the
                                                         //current lod

    void generateOctreeLODs(OctreeCoord *root, //this method searches for the
                            int &newcount);    //leaves in the subtree with
                                               //the greatest depth and most
                                               //coords - this leaves will be
                                               //combined by selecting an
                                               //representant with method 
                                               //selectOctreeRepresent

    OctreeCoord *selectOctreeRepresent(OctreeCoord *root);
                          //selects a representant of the sub-nodes of root
                          //(the sub-nodes must be leaf-nodes). If the
                          //coords are used by faces the method uses a weighted
                          //meanvalue of the (normalized) errors produced by the
                          //algorithms errorarea, errorvolume, weightedmean
                          //for selecting the representant
                          //if the coords are used by lines only the algorithm
                          //weightedmean is used
                          //These algorithms are realized in calcError

    void calcError(OctreeCoord *root, int method, float err[8]);
                          //calc the errors which will occur if a sub-node is
                          //selected for the representant. if sub-node x
                          //doesn't exist the corresponding err[x] is set to -1
                          //parameter method specifies the algorithm which 
                          //should be used

    void normalizeErrors(float err[8]); //scale the errors to [0,1]
                                        //an -1 is unchanged (undefined)

    void deleteOctreeShapeRefs(OctreeCoord *root); //delete the references to
                                                   //shapes of all coordinates
                                                   //in the octree specified
                                                   //by root

    void resetOctreeLODs(OctreeCoord *root); //reset data structures to lod 0
                                             //all pseudo leaf-nodes will be
                                             //converted back to intermediate-
                                             //nodes
};

//----------------------------------------------------------------------------

#endif

