// ===========================================================================
//  Copyright 1996 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       dataset
//  TYPE:       c++ header
//  PROJECT:    lodestar
//  CONTENT:    class DataSet, class Data
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rs      Reinhard Sainitzer
//		hb      Herbert Buchegger
// ===========================================================================
//  HISTORY:
//
// ===========================================================================
//  MANUAL:
//
//  The class is a predecessor of the classes CoordSet, MaterialSet, NormalSet,
//  TextureCoordSet, FaceSet and LineSet. It contains variables and methods
//  which are common to all of these classes. The insert-methods of all these
//  classes are calling the method DataSet::insertArray(Data *d). This method
//  appends a reference to d to the dynamic array refArray. So you can access
//  the inserted Datas by the order of insertion with the method getIndex.
//  The classes CoordSet, MaterialSet and all others remove double defined
//  Datas, so it can happen that *d points to the same element. The method
//  printSet prints double defined elements only one time. To ensure this it 
//  is necessary to mark already printed elements. This is made by setting 
//  the bit DATASET_STAT_DONE in variable status of class Data.
//  In some cases it can't be allowed printing double defined Datas only on
//  time within the first x elements (insertion order). (e.g. a VRML-Cube with
//  material binding PER_FACE uses the first 6 Materials so within this 6 
//  Materials removing doublets is not allowed) To avoid this the
//  variable DataSet::countUncompressed must be set to x.
//  The method setIndices sets the variable Data::index according to the
//  printing-order generated by printSet. With this mechanism e.g. a FaceSet
//  can find out the index of a Normal in a NormalSet
//
//  Elements which will be inserted must be derived from class Data
//
//  SEE ALSO
//      CoordSet, MaterialSet, NormalSet, TextureCoordSet, FaceSet, LineSet
//
// ===========================================================================

#ifndef _DATASET_H_
#define _DATASET_H_

#include "global.h"
#include "fields.h"
#include "refarray.h"
#include <stdio.h>

//----------------------------------------------------------------------------
//   constants for Data::index
//----------------------------------------------------------------------------

#define DATASET_STAT_DONE    0x80

//----------------------------------------------------------------------------

class Data
{
public:

    int index;          //index of element in printing order !! (printSet)
    int status;

    Data();
    virtual void print(FILE *fd, int fid);
};

//----------------------------------------------------------------------------
//   class DataSet
//----------------------------------------------------------------------------

class DataSet
{
public:

    int count;                    //count of elements in the set
                                  //without doublets !

    int countUncompressed;        //prevents printSet from printing double
                                  //defined elements only one time within
                                  //the first countUncompressed elements

    DataSet();
    ~DataSet();
    Data *getIndex(int index);    //get element by index (insertion order !!) 
    int getMaxIndex();            //returns count of inserted elements-1
                                  //(with doublets !)
     
protected:
    ReferenceArray *refArray;   //pointer to an object of class ReferenceArray
                                //created in the constructor

    void insertArray(Data *d);  //append a Reference to d to the dynamic array
                                //refArray - this allows picking an element
                                //by index in insertion order

    void printSet(FILE *fd, int indent, //prints out the contents of the
                  int fid);             //inserted elements by calling their
                                        //print-method. The elements are 
                                        //printed generally in insertion
                                        //order BUT double defined elements
                                        //are printed only once. exception:
                                        //if countUncompressed>0 then doublets
                                        //are printed until the count of the
                                        //elements printed is >= then 
                                        //countUncompressed
                                        //at this time only class MaterialSet
                                        //uses the field-identifier fid
                                        //all other classes ignores the
                                        //parameter fid because the have only 
                                        //one field

    void setIndices();  //The method setIndices sets the variable Data::index
                        //in all inserted Datas according to the
                        //printing-order generated by printSet. With this
                        //mechanism e.g. a FaceSet can find out the index
                        //of a Normal in a NormalSet
};

//----------------------------------------------------------------------------

#endif

