// ===========================================================================
//  Copyright 1996 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       face
//  TYPE:       c++ code
//  PROJECT:    lodestar
//  CONTENT:    implementation of class FaceSet, class Face
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rs      Reinhard Sainitzer
//		hb      Herbert Buchegger
// ===========================================================================
//  HISTORY:
//
// ===========================================================================

#include "face.h"
#include "enum.h"
#include "global.h"
#include "tri.h"
#include <assert.h>

//----------------------------------------------------------------------------
//   class Face : Shape
//----------------------------------------------------------------------------

float Face::splitCosinus = 0.984808; //cos(10 degree)
                                     //cosinus of the maximum allowed angle of 
                                     //the vertexNormals in opposite side of an
                                     //Face with 4 vertices - used in function
                                     //isIrregularQuadrangle 
  
//---Face::Face----------------------------------------------------------------
// Input : Vertex v1  : init  the vertices of the face -
//         Vertex v2  : init  if Vertex::coord==NULL this means
//         Vertex v3  : init  this vertex is undefined
//         Vertex v4  : init
// Output: none
// a vertex is not defined if Vertex::coord==NULL the constructor closes this
// 'NULL-holes' by shifting the vertices right from the 'hole' one position to
// the left - also the constructor removes vertices with double defined coords
// further the constructor calculates the values of the variables vertexArea
// and vertexNormal

Face::Face(Vertex v1, Vertex v2, Vertex v3, Vertex v4) : Shape()
{
    int i,j;
    vertices=4;

    vertex[0]=v1;
    vertex[1]=v2;
    vertex[2]=v3;
    vertex[3]=v4;

    //--- close 'NULL-holes' :-)

    i=0;
    while (i<vertices)
        if (vertex[i].coord==NULL) removeVertex(i); else i++;


    //--- remove double defined coordinates

    i=0;                             //!!! verschiedene Materialien ???
    while (i<vertices)
        if (vertex[i].coord==vertex[(i+1)%vertices].coord) removeVertex(i); else i++;

    //--- calculate some characteristics

    float edgeLength[4];
    Vec nnull;

    switch (vertices)
    {
        case 0: 
        case 1:
            break; //nothing to do

        case 2:
            vertexArea[0]=vertexArea[1] =0;
            vertexNormal[0]=vertexNormal[1] =nnull;
            break;

        default: 
        {
            Vec vec[4],v,w;
            for (i=0; i<vertices; i++) 
            {
                Coord *c1=vertex[i].coord;
                Coord *c2=vertex[(i+1)%vertices].coord;
                v.x=c2->x-c1->x;
                v.y=c2->y-c1->y;
                v.z=c2->z-c1->z;
                vec[i]=v;
                edgeLength[i]=v.getLength();
            }
            for (i=0; i<vertices; i++)
            {
                j=(i-1+vertices)%vertices;
                v=vec[i];
                w=vec[j];
                float area = edgeLength[i]*edgeLength[j];
                if (area==0) vertexArea[i] = 0;
                else
                {
                    float cosinus = (v.x*w.x + v.y*w.y + v.z*w.z) /area;
                    float sin2 = 1-cosinus*cosinus;
                    if (sin2<0) sin2=0;
                    vertexArea[i] = area * sqrt(sin2) /2;
                }

                Vec nn;
                nn.x = v.y*w.z - v.z*w.y;
                nn.y = v.z*w.x - v.x*w.z;
                nn.z = v.x*w.y - v.y*w.x;
                float nnlen = nn.getLength(); 
                if (nnlen==0) nn=nnull; else nn /= nnlen;
                vertexNormal[i] = nn;
            }
            for (i=0; i<vertices; i++)
                if (vertexNormal[i].isNullVector())
                    vertexNormal[i]=vertexNormal[(i+1)%vertices];
        }
    }

}

//---Face::removeVertex-------------------------------------------------------
// Input : int index : index of vertex to remove
// Output: none
//remove vertex at index and shift the vertices right
//from index one position to the left

void Face::removeVertex(int index)
{
    int i;
    for (i=index; i<3; i++) vertex[i]=vertex[i+1];
    vertex[3].coord=NULL;
    vertices--;
}

//---Face::createRepresent----------------------------------------------------
// Input : int targetlod : number of the lod of the representant
// Output: Face *        : pointer to the new Face
//creates a new Face whose vertices are the representatives of the
//vertices of this Face in the next lod

Face *Face::createRepresent(int targetlod)
{
    Vertex v[4];
    Coord *r,*c;
    Coord *represent[4];
    Coord *selected[4];
    int i,j;

    for (i=0; i<vertices; i++)
    {
        c = vertex[i].coord;
        r = c->getRepresent(targetlod);
        represent[i] = r;
        selected[i] = r->getSelected(c->lod);
    }

    for (i=0; i<vertices; i++)
    {
        //--------------------------------------------------------
        //point has no representative or point is selected for the 
        //representative - this means this point is unchanged
        //--------------------------------------------------------

        if (selected[i]==vertex[i].coord)
        {
            v[i] = vertex[i];
            v[i].coord = represent[i];
        }

        //--------------------------------------------------------

        else
        { 
            //----------------------------------------------------
            //check if selected is another point of the polygon
            //----------------------------------------------------

            int elim = 0;
            for (j=0; j<vertices && !elim; j++)
                if (selected[i]==vertex[j].coord) elim=1;

            //----------------------------------------------------
            //the selected point is not a part of this polygon. 
            //----------------------------------------------------

            if (elim==0)
            {                              
                v[i] = vertex[i];
                v[i].coord = represent[i];

            }
        }
    }

    return new Face(v[0],v[1],v[2],v[3]);
}

//---Face::getVertexCountOfRepresent------------------------------------------
// Input : int targetlod : number of the lod of the representant
// Output: int           : the number of vertices of the representant
//gets the count of the Face mapped to the representatives
//of the coordinates (count of the vertices of the Face the function
//createRepresent would produce)

int Face::getVertexCountOfRepresent(int targetlod)
{
    Coord *represent[4];
    int vnew = vertices;
    int i;

    for (i=0; i<vertices; i++)
        represent[i] = vertex[i].coord->getRepresent(targetlod);

    for (i=0; i<vertices; i++) 
        if (represent[i]==represent[(i+1)%vertices]) vnew--;

   return vnew;
}

//---Face::isEquivOrSubsetOf--------------------------------------------------
// Input : Face *f            : compare this Face with f
//         int vertexOrdering : see enum.h
// Output: int                : TRUE or FALSE
//compare this Face with f. If (a subset of) the coordinates of this Face
//(is)are equivalent to f then this function returns 1
//e.g. if this Face is a triangle and f is a quadrangle and the triangle
//is part of the quadrangle the function returns 1.
//If vertexOrdering is VRML_UNKNOWN_ORDERING then the orientation of the 
//faces will be ignored

int Face::isEquivOrSubsetOf(Face *f, int vertexOrdering)
{
    int i;

    int fnv = f->vertices;

    Coord *c1 = vertex[0].coord;
    Coord *c2 = vertex[1].coord;
    Coord *c3 = vertex[2].coord;
    Coord *c4 = vertex[3].coord;

    if (vertices<=1) return 1;

    if (vertices==2)
        for (i=0; i<fnv; i++)
            if (c1==f->vertex[i].coord)
            {
                if (c2==f->vertex[(i+1)%fnv].coord) return 1;
                if (c2==f->vertex[(i-1+fnv)%fnv].coord) return 1;
                if (fnv==4 && c2==f->vertex[(i+2)%fnv].coord) return 1;
            }

    if (vertices==3)
        if (fnv>=3)
        {
            for (i=0; i<fnv; i++)
                if (c1==f->vertex[i].coord &&
                    c2==f->vertex[(i+1)%fnv].coord &&
                    c3==f->vertex[(i+2)%fnv].coord) return 1;

            if (vertexOrdering==VRML_UNKNOWN_ORDERING)
                for (i=0; i<fnv; i++)
                    if (c3==f->vertex[i].coord &&
                        c2==f->vertex[(i+1)%fnv].coord &&
                        c1==f->vertex[(i+2)%fnv].coord) return 1;
        }

    if (vertices==4)
        if (fnv==4)
        {
            for (i=0; i<fnv; i++)
                if (c1==f->vertex[i].coord &&
                    c2==f->vertex[(i+1)%fnv].coord &&
                    c3==f->vertex[(i+2)%fnv].coord &&
                    c4==f->vertex[(i+3)%fnv].coord) return 1;

            if (vertexOrdering==VRML_UNKNOWN_ORDERING)
                if (c4==f->vertex[i].coord &&
                    c3==f->vertex[(i+1)%fnv].coord &&
                    c2==f->vertex[(i+2)%fnv].coord &&
                    c1==f->vertex[(i+3)%fnv].coord) return 1;
        }

    return 0;
}

//---Face::calcErrorVolume----------------------------------------------------
// Input : Coord *newrepresent : pointer to a new point (representant)
//         int targetlod       : number of the lod of the representant
// Output: float               : the error vomlume
//calcs the volume of the pyramid which base is defined by the area of
//the face (vertexAreas) and the top is newrepresent uses calcVolume
//NOTE:if newrepresent isn't the represent of at least one of
//     the vertices of the face (in lod targetlod) then zero is returned

float Face::calcErrorVolume(Coord *newrepresent, int targetlod)
{
    if (vertices<=2) return 0;

    float volume;
    if (vertices==3) volume = calcVolume(0,1,2,newrepresent,targetlod);
    else
    {
        int si;
        isIrregularQuadrangle(si);    //the return-value doesn't matter
        if (si==0) volume = calcVolume(0,1,2,newrepresent,targetlod)+
                            calcVolume(0,2,3,newrepresent,targetlod);
              else volume = calcVolume(1,2,3,newrepresent,targetlod)+
                            calcVolume(1,3,0,newrepresent,targetlod);
    }

    return volume;
}

//---Face::calcVolume---------------------------------------------------------
// Input : int i1,i2,i3  : indices of vertices in this Face
//         Coord *cnew   : pointer to a new point (representant)
//         int targetlod : number of the lod of the representant
// Output: float         : volume of the tetraeder
//calcs the volume of the tetraeder which base is defined by the
//coordinates with indices i1,i2,i3 and the top is cnew
//NOTE:if cnew isn't the represent of at least one of
//     i1,i2,i3 (in lod targetlod) then zero is returned

float Face::calcVolume(int i1, int i2, int i3, Coord *cnew, int targetlod)
{
    Coord *c1 = vertex[i1].coord;
    Coord *c2 = vertex[i2].coord;
    Coord *c3 = vertex[i3].coord;

    if (c1->getRepresent(targetlod) != cnew &&
        c2->getRepresent(targetlod) != cnew &&
        c3->getRepresent(targetlod) != cnew) return 0;

    Vec t;
    t.x = cnew->x - c1->x;
    t.y = cnew->y - c1->y;
    t.z = cnew->z - c1->z;

    Vec nn = vertexNormal[i1];
    float height = (t.x*nn.x + t.y*nn.y * t.z*nn.z);

    return fabs(vertexArea[i1]*height/3);
}

//---Face::calcErrorArea------------------------------------------------------
// Input : Coord *newrepresent : pointer to a new point (representant)
//         int targetlod       : number of the lod of the representant
// Output: float               : error area
//if newrepresent is the representative of at least one of the coords
//of this Face then: The difference of the area of the face and the area
//of the face where all coords which have newrepresent as representative
//in lod targetlod are replaced by newrepresent

float Face::calcErrorArea(Coord *newrepresent, int targetlod)
{
    if (vertices<=2) return 0;

    Vertex nv[4];
    Coord *ch;
    int i;

    for (i=0; i<vertices; i++)
    {
        nv[i]=vertex[i];
        ch = vertex[i].coord->getRepresent(targetlod);
        if (ch==newrepresent) nv[i].coord = ch;
    }

    Face newface = Face(nv[0],nv[1],nv[2],nv[3]);

    //--- calc old area

    float oldarea;
    if (vertices==3) oldarea = vertexArea[0];
    else
    {
        int si;
        isIrregularQuadrangle(si);
        if (si==0) oldarea = vertexArea[1]+vertexArea[3];
              else oldarea = vertexArea[0]+vertexArea[2];
    }

    //--- calc new area

    float newarea;
    if (newface.vertices<=2) newarea = 0;
    else if (newface.vertices==3) newarea = newface.vertexArea[0];
    else
    {
        int si;
        newface.isIrregularQuadrangle(si);    //the return-value doesn't matter
        if (si==0) newarea = newface.vertexArea[1]+newface.vertexArea[3];
              else newarea = newface.vertexArea[0]+newface.vertexArea[2];
    }

    return newarea - oldarea;
}

//---Face::isIrregularQuadrangle----------------------------------------------
// Input : none
// Output: int &splitindex : vertex-index where a quadrangle has to be split
//returns 1 if the face is a quadrangle and the face isn't convex or
//it is too much distorted in 3D
//in param splitangle the function returns the the index where the 
//quadrangle has to be split into triangles (so that the angle between
//the 2 new triangles is minimal)
//NOTE: the function always sets parameter splitangle
//      independend of the return value

int Face::isIrregularQuadrangle(int &splitindex)
{
    if (vertices<=3)
    {
        splitindex=0;
        return 0;
    } 

    Vec nn0 = vertexNormal[0];
    Vec nn1 = vertexNormal[1];
    Vec nn2 = vertexNormal[2];
    Vec nn3 = vertexNormal[3];

    float cos02 = nn0.x*nn2.x + nn0.y*nn2.y + nn0.z*nn2.z;
    float cos13 = nn1.x*nn3.x + nn1.y*nn3.y + nn1.z*nn3.z;

    if (cos02<cos13) splitindex = 0;
                else splitindex = 1;

    if (cos02<splitCosinus || cos13<splitCosinus) return 1;
    return 0;
}

//----------------------------------------------------------------------------
//   class FaceSet
//----------------------------------------------------------------------------

int FaceSet::nolines = 0; //if !=0 the FaceSet throws away lines
int FaceSet::noquads = 0; //if !=0 quadrangles are always split into triangles

//---FaceSet::FaceSet---------------------------------------------------------
//Input : int alod             : init
//        CoordSet *cs         : init
//        int amaterialBinding : init
//        int anormalBinding   : init
//        int avertexOrdering  : init
//Output: none

FaceSet::FaceSet(int alod, CoordSet *cs, int amaterialBinding,
                 int anormalBinding, int avertexOrdering)
{
    lod = alod;
    coordSet = cs;
    materialBinding = amaterialBinding;
    normalBinding = anormalBinding;
    vertexOrdering = avertexOrdering;

    range = 0;
    count = 0;
    linecount = 0;
    polycount = 0;
    tricount = 0;
    quadcount = 0;

    firstFace = NULL;
    lastFace = NULL;

    next = NULL;
}

//---FaceSet::~FaceSet--------------------------------------------------------
// Input : none
// Output: none

FaceSet::~FaceSet()
{
    Face *fh;
    while (firstFace)
    {
        fh=firstFace;
        firstFace=(Face *)firstFace->next; 
        delete fh;
    }
}

//---FaceSet::insert----------------------------------------------------------
// Input : Face **f : p to p to the Face to insert
// Output: Face **f : possible modified if face is split or deleted
//insert f in the FaceSet.  f is inserted internally in a list.
//if f is a point f is deleted
//if f is a line and FaceSet::nolines!=0 f is deleted
//if f is a quadrangle and f is not convex or too much distorted
//    in 3D (Face::splitCosinus) the quadrangle will be splitted into
//    2 triangles which will be inserted recursivly
//if f is equal or part of an already inserted  face f is deleted
//NOTE: faces with more vertices must be inserted before faces with
//      less vertices - otherwise the FaceSet is unable to remove
//      all faces which are part of others

void FaceSet::insert(Face **f)
{
    Face *fh = *f;

    if (fh->vertices==4)
    {
        if (fh->vertex[0].coord==fh->vertex[2].coord ||
            fh->vertex[1].coord==fh->vertex[3].coord)
        {
            Face *fnew = new Face(fh->vertex[0],fh->vertex[1],Vertex(),Vertex());
            fh = new Face(fh->vertex[2],fh->vertex[3],Vertex(),Vertex()); 
            delete *f;
            *f = fh;
            insert(&fnew);
        }
        else
	{
            //if quadrangle and not convex or quadrangles not allowed (noquads==1)  ->
            //split into 2 triangles - recursive call to insert for 2nd tri.
 
            int si;
            if (fh->isIrregularQuadrangle(si) || noquads)
            {
                Face *fnew;
                if (si==0)
                {
                    fnew = new Face(fh->vertex[0],fh->vertex[1],fh->vertex[2],Vertex());
                    fh   = new Face(fh->vertex[0],fh->vertex[2],fh->vertex[3],Vertex()); 
                }
                else
                {
                    fnew = new Face(fh->vertex[1],fh->vertex[2],fh->vertex[3],Vertex());
                    fh   = new Face(fh->vertex[0],fh->vertex[1],fh->vertex[3],Vertex()); 
                }
                delete *f;
                *f = fh;
                insert(&fnew);
            }
	}
    }

    if (insertOk(fh))
    {
        if (firstFace)
        {
            lastFace->next=fh;
            lastFace=fh;
        }
        else firstFace=lastFace=fh;

        int fnv = fh->vertices;
        int i;
        for (i=0; i<fnv; i++) fh->vertex[i].coord->addShapeRef(fh);
        count++;
        if (fnv==2) linecount++;
        else
        {
            polycount++;
            if (fnv==4) quadcount++; else tricount++;
        }
    }
    else
    {
        delete *f;
        *f=NULL;
    }
}

//---FaceSet::print-----------------------------------------------------------
// Input : FILE *fd      : VRML output file descriptor
//         int indent    : indention of output
//         int printmode : print lines of polys - see FACE_PRINT_xxx constants
// Output: none
//writes the content of all defined fields to fd
//the fields are printed using printSet. printmode - see printSet

void FaceSet::print(FILE *fd, int indent, int printmode)
{
    printIndent(fd,indent);
    fprintf(fd,"%s [ ",fieldIdString[FID_COORDINDEX]);
    printSet(fd,indent+13,FID_COORDINDEX,printmode);
    fprintf(fd," ]\n");

    if (count==0) return;

    if (firstFace->vertex[0].material)
    {
        printIndent(fd,indent);
        fprintf(fd,"%s [ ",fieldIdString[FID_MATERIALINDEX]);
        printSet(fd,indent+16,FID_MATERIALINDEX,printmode);
        fprintf(fd," ]\n");
    }

    if (firstFace->vertex[0].normal)
    {
        printIndent(fd,indent);
        fprintf(fd,"%s [ ",fieldIdString[FID_NORMALINDEX]);
        printSet(fd,indent+14,FID_NORMALINDEX,printmode);
        fprintf(fd," ]\n");
    }

    if (firstFace->vertex[0].textureCoord)
    {
        printIndent(fd,indent);
        fprintf(fd,"%s [ ",fieldIdString[FID_TEXTURECOORDINDEX]);
        printSet(fd,indent+20,FID_TEXTURECOORDINDEX,printmode);
        fprintf(fd," ]\n");
    }
}

//---FaceSet::generateNextLOD-------------------------------------------------
// Input : none
// Output: none
//creates a new FaceSet (lod+1) which contains the next Level of Detail
//variable next points to this new FaceSet

void FaceSet::generateNextLOD()
{
    int newlod = lod+1;

    FaceSet *tsh = new FaceSet(newlod,coordSet,
                               materialBinding,normalBinding,
                               vertexOrdering);
    next = tsh;


    coordSet->deleteShapeRefs();

    //First all quadrangles will be inserted in the new FaceSet
    //then all triangles and finally all edges. This ensures that
    //no polygon which is already inserted must be removed from the set

    Face *newface;
    int i;
    for (i=4; i>1; i--)
    {
        Face *face = firstFace;
        while (face)
        {
            if (face->getVertexCountOfRepresent(newlod)==i)
            {
                newface = face->createRepresent(newlod);
                tsh->insert(&newface);
            }
            face=(Face *)face->next;
        }
    }
}

//---FaceSet::getCenter-------------------------------------------------------
// Input : none
// Output: float &cx, float &cy, float &cz : the calculated center
//get the center of all coordinates used in the set - result in cx,cy,cz

void FaceSet::getCenter(float &cx, float &cy, float &cz)
{
    Face *face = firstFace;
    if (face==NULL)
    {
        cx=cy=cz = 0;
        return;
    }

    Coord *coord = face->vertex[0].coord;
    Bounds bounds = Bounds(coord->x,coord->y,coord->z,
                           coord->x,coord->y,coord->z);

    while (face)
    {
        int vertices = face->vertices;
        for (int i=0; i<vertices; i++)
        { 
            coord = face->vertex[i].coord;
            bounds.enlargeIfOutside(coord->x,coord->y,coord->z);
        }
        face=(Face *)face->next;
    }

    cx = (bounds.xmax+bounds.xmin) /2;
    cy = (bounds.ymax+bounds.ymin) /2;
    cz = (bounds.zmax+bounds.zmin) /2;
}

//---FaceSet::getSphere-------------------------------------------------------
// Input : none
// Output: float : the radius of the bounding sphere
//get the radius of the bounding sphere of all faces

float FaceSet::getSphere()
{
    Face *face = firstFace;
    if (face==NULL) return 0;

    Coord *coord = face->vertex[0].coord;
    Bounds bounds = Bounds(coord->x,coord->y,coord->z,
                           coord->x,coord->y,coord->z);

    while (face)
    {
        int vertices = face->vertices;
        for (int i=0; i<vertices; i++)
        { 
            coord = face->vertex[i].coord;
            bounds.enlargeIfOutside(coord->x,coord->y,coord->z);
        }
        face=(Face *)face->next;
    }

    float dx = bounds.xmax - bounds.xmin;
    float dy = bounds.ymax - bounds.ymin;
    float dz = bounds.zmax - bounds.zmin;
    return sqrt(dx*dx + dy*dy + dz*dz);;
}

//---FaceSet::insertOk--------------------------------------------------------
// Input : Face *f : pointer to a face for checking
// Output: int     : TRUE or FALSE
//returns 0 if f is a point or if f is a line and FaceSet::nolines!=0 
//or if f is equal or part of a face already inserted in the set

int FaceSet::insertOk(Face *f)
{
    if (f->vertices<=1) return 0;
    if (nolines && f->vertices==2) return 0;

    int n = f->vertices-1;
    int i;
    for (i=0; i<n; i++)
    {
        ShapeRef *frh = f->vertex[i].coord->shapeRef; 
        while (frh)
        {
            if (f->isEquivOrSubsetOf((Face *)frh->shape,vertexOrdering)) return 0;
            frh=frh->next;
        }
    }

    return 1;
}

//---FaceSet::printSet--------------------------------------------------------
// Input : FILE *fd      : VRML output file descriptor
//         int indent    : indention of output
//         int fid       : field identifier (see FID_xxx in enum.h)
//         int printmode : print lines or polys (FACE_PRINT_XXX)
// Output: none
//print out the field (specified by fid) of the FaceSet containing the
//data of the inserted faces by calling their print-method. 
//the method prints either only the data of the polygons (printmode==
//FACE_PRINT_POLY) or only the data of the lines (vertices==2) (printmode==
//FACE_PRINT_LINE) - fid see enum.h

void FaceSet::printSet(FILE *fd, int indent, int fid, int printmode)
{
    int j;
    int first=1;
    int fnv;

    global_cpl=0;

    Face *face = firstFace;
    while (face)
    {
        fnv = face->vertices;

        if (printmode==FACE_PRINT_LINE && fnv==2 ||
            printmode==FACE_PRINT_POLY && fnv>2)
        {

            if (first) first=0;
            else
            {
                fprintf(fd,",");
                if (global_cpl>=global_longcpl)
                {
                    fprintf(fd,"\n");
                    printIndent(fd,indent);
                    global_cpl=0;
                }
            }

            switch (fid)
            {
                case FID_COORDINDEX:
                    for (j=0; j<fnv; j++)
                        fprintf(fd,"%d,",face->vertex[j].coord->index);
                    fprintf(fd,"-1");
                    global_cpl+=fnv+1;
                    break;

                case FID_MATERIALINDEX:
                    if (materialBinding==VRML_PER_PART ||
                        materialBinding==VRML_PER_FACE ||
                        materialBinding==VRML_PER_PART_INDEXED ||
                        materialBinding==VRML_PER_FACE_INDEXED)
                    {
                        fprintf(fd,"%d",face->vertex[0].material->index);
                        global_cpl++;
                    }

                    if (materialBinding==VRML_PER_VERTEX ||
                        materialBinding==VRML_PER_VERTEX_INDEXED)
                    {
                        for (j=0; j<fnv; j++)
                            fprintf(fd,"%d,",face->vertex[j].material->index);
                        fprintf(fd,"-1");
                        global_cpl+=fnv+1;
                    }
                    break;

                case FID_NORMALINDEX:
                    if (normalBinding==VRML_PER_PART ||
                        normalBinding==VRML_PER_FACE ||
                        normalBinding==VRML_PER_PART_INDEXED ||
                        normalBinding==VRML_PER_FACE_INDEXED)
                    {
                        fprintf(fd,"%d",face->vertex[0].normal->index);
                        global_cpl++;
                    }

                    if (normalBinding==VRML_PER_VERTEX ||
                        normalBinding==VRML_PER_VERTEX_INDEXED)
                    {
                        for (j=0; j<fnv; j++)
                            fprintf(fd,"%d,",face->vertex[j].normal->index);
                        fprintf(fd,"-1");
                        global_cpl+=fnv+1;
                    }
                    break;

                case FID_TEXTURECOORDINDEX:
                    for (j=0; j<fnv; j++)
                        fprintf(fd,"%d,",face->vertex[j].textureCoord->index);
                    fprintf(fd,"-1");
                    global_cpl+=fnv+1;
                    break;

                default: assert(0);
            }
        }   

        face=(Face *)face->next;
    }    

}

//----------------------------------------------------------------------------
