// ===========================================================================
//  Copyright 1996 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       face
//  TYPE:       c++ header
//  PROJECT:    lodestar
//  CONTENT:    class FaceSet, class Face
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rs      Reinhard Sainitzer
//		hb      Herbert Buchegger
// ===========================================================================
//  HISTORY:
//
// ===========================================================================
//  MANUAL:
//
// The class FaceSet is used to store faces defined in VRML-Nodes from type
// IndexedFaceSet. Faces are inserted using the method FaceSet::
// insert(Face **f). the insert-method removes faces which are equal or part
// of already inserted faces. The faces are inserted internally in a linear list.
// The FaceSet is also used to produce Levels of Detail. The method 
// generateNextLOD creates a new FaceSet and fills this FaceSet with the
// faces of the next lod (polygons of the current FaceSet are mapped onto
// the representants of the coordinates in the next lod) doublettes and 
// collapsed polygons will be discarded - the variable next points then
// to this new FaceSet (next lod).
//
// SEE ALSO:
//     Shape, Vertex
//
// ===========================================================================

#ifndef _FACE_H_
#define _FACE_H_

#include "global.h"
#include "coord.h"
#include "shape.h"

//----------------------------------------------------------------------------
//   constants for parameter printmode
//----------------------------------------------------------------------------

#define FACE_PRINT_LINE   1
#define FACE_PRINT_POLY   2

//----------------------------------------------------------------------------

class Face : public Shape
{
public:

    int vertices;          //count of vertices of this face (max 4)
    Vertex vertex[4];      //data of the vertices of this face
    float vertexArea[4];   //area - defined by the 2 edges running to this
                           //vertex - calculated in contructor
    Vec vertexNormal[4];   //normals of the vertices - calculated in constructor

    static float splitCosinus;  //cosinus of the maximum allowed angle of the 
                                //vertexNormals in opposite side of an Face
                                //with 4 vertices - used in function
                                //isIrregularQuadrangle 

    Face(Vertex v1, Vertex v2,   //a vertex is not defined if Vertex::coord==NULL
         Vertex v3, Vertex v4);  //the constructor closes this 'NULL-holes' by
                                 //shifting the vertices right from the 'hole' one
                                 //position to the left - also the constructor 
                                 //removes vertices with double defined coords
                                 //further the constructor calculates the values
                                 //of the variables vertexArea and vertexNormal

    void removeVertex(int index);     //remove vertex at index and shift the
                                      //vertices right from index one position
                                      //to the left

    Face *createRepresent(int targetlod); //creates a new Face whose vertices 
                                          //are the representatives of the
                                          //vertices of this Face in the next lod

    int getVertexCountOfRepresent(int targetlod); //gets the count of the Face
                                                  //mapped to the representatives
                                                  //of the coordinates
                                                  //(count of the Face the function
                                                  //createRepresent would produce)

    int isEquivOrSubsetOf(Face *f,              //compare this Face with f
                          int vertexOrdering);  //if (a subset of) the coordinates
                                                //of this Face are equivalent to f
                                                //then this function returns 1
                                                //e.g. if this Face is a 
                                                //triangle and f is a quadrangle
                                                //and the triangle is part of
                                                //the quadrangle the function 
                                                //returns 1 - if vertexOrdering
                                                //is VRML_UNKNOWN_ORDERING then
                                                //the orientation of the faces
                                                //will be ignored

    float calcErrorVolume(Coord *newrepresent,  //calcs the volume of the 
                          int targetlod);       //pyramid which base is defined
                                                //by the area of the face and the
                                                //top is newrepresent
                                                //uses calcVolume

    float calcVolume(int i1, int i2, int i3,      //calcs the volume of the 
                     Coord *cnew, int targetlod); //tetraeder which base is
                                                  //defined by the coordinates
                                                  //with indices i1,i2,i3 and
                                                  //the top is cnew
                                                  //NOTE:if cnew isn't the
                                                  //represent of at least one of
                                                  //i1,i2,i3 (in lod targetlod)
                                                  //then zero is returned

    float calcErrorArea(Coord *newrepresent,  //if newrepresent is the repre-
                        int targetlod);       //sentative of at least one of
                                              //the coords of this Face then:
                                              //The difference of the area of
                                              //the face and the area of the
                                              //face where all coords which 
                                              //have newrepresent as representative
                                              //in lod targetlod are replaced
                                              //by newrepresent

    int isIrregularQuadrangle(int &splitindex); //returns 1 if the face is a
                                                //quadrangle and the face isn't
                                                //convex or it is too much
                                                //distorted in 3D
                                                //in param splitangle the function
                                                //returns the the index where the 
                                                //quadrangle has to be split
                                                //into triangles (so that the
                                                //angle between the 2 new
                                                //triangles is minimal)
};

//----------------------------------------------------------------------------

class FaceSet
{
public:

    int lod;              //the number of the lod representated by this faceset
    int count;            //count of all faces in the set
    int polycount;        //count of the triangles and quadrangles in the set
    int linecount;        //count of the lines in the set
    int tricount;         //count of the triangles in the set
    int quadcount;        //count of the quadrangles in the set
    int materialBinding;  //(vrml) material binding used in this set 
                          //used in printSet for printing out the index
                          //references to Material-Nodes 
    int normalBinding;    //(vrml) normal binding used in this set
                          //used in printSet for printing out the index
                          //references to Normal-Nodes
    int vertexOrdering;   //(vrml) vertex ordering used in this set
                          //used in insertOk and Face::isEquivOrSubsetOf
                          //if value is UNKNOWN_ORDERING the orientation 
                          //of the faces will be ignored
    float range;          //the maximum distance to the camera for
                          //this faceset (lod) before switching to the next
    CoordSet *coordSet;   //pointer to the CoordSet where the coords used
                          //in this FaceSet come from
    FaceSet *next;        //pointer to the next FaceSet (lod)
                          //do not modify this variable! the variable is set
                          //by generateNextLOD

    static int nolines;   //throw away lines
    static int noquads;   //always split quadrangles into triangles

    FaceSet(int alod, CoordSet *cs, int amaterialBinding, int anormalBinding,
            int avertexOrdering);
    ~FaceSet();

    void insert(Face **f);  //insert f in the FaceSet.
                            //f is inserted internally in a list.
                            //if f is a point f is deleted
                            //if f is a line and FaceSet::nolines!=0 f is deleted
                            //if f is a quadrangle and f is not convex or
                            //   too much distorted in 3D (Face::splitCosinus)
                            //   the quadrangle will be splitted into
                            //   2 triangles which will be inserted recursivly
                            //if f is equal or part of an already inserted
                            //   face f is deleted
                            //NOTE: faces with more vertices must be inserted
                            //      before faces with less vertices - otherwise
                            //      the FaceSet is unable to remove all faces
                            //      which are part of others

    void print(FILE *fd, int indent ,int printmode);  //writes the content of
                                                      //all defined fields to fd
                                                      //the fields are printed
                                                      //using printSet
                                                      //printmode - see printSet

    void generateNextLOD();  //creates a new FaceSet (lod+1)
                             //which contains the next Level of Detail
                             //variable next points to this new FaceSet

    void getCenter(float &cx, float &cy, float &cz); //get the center of all
                                                     //coordinates used in the
                                                     //set - result in cx,cy,cz

    float getSphere();    //get the radius of the bounding sphere of all faces
     
protected:
    Face *firstFace; //faces are stored in a list - pointer to the first face
    Face *lastFace;  //faces are stored in a list - pointer to the last face

    int insertOk(Face *f);  //returns 0 if f is a point - f is a line and
                            //FaceSet::nolines!=0 - if f is equal or part
                            //of a face already inserted in the set

    void printSet(FILE *fd, int indent,    //print out the field (specified by 
                  int fid, int printmode); //fid) of the FaceSet containing
                                           //the data of the inserted faces by 
                                           //calling their print-method. 
                                           //the method prints either only the
                                           //data of the polygons (printmode==
                                           //FACE_PRINT_POLY) or only the data
                                           //of the lines (vertices==2) (printmode
                                           //==FACE_PRINT_LINE) 
                                           //fid - see enum.h
};

//----------------------------------------------------------------------------

#endif

