// ===========================================================================
//  Copyright 1996 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       fields.h
//  TYPE:       c++ header
//  PROJECT:    lodestar
//  CONTENT:    class Value and all classes derived from class Value
//              class Field
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rs      Reinhard Sainitzer
//		hb      Herbert Buchegger
// ===========================================================================
//  HISTORY:
//
// ===========================================================================
//  MANUAL:     The class Field has a pointer to a class Value and a pointer
//              to the next Field object (or NULL). Every Field has a id (see
//              enum.h). So it is possible to create a list for a Node object.
//              The classes derived from class Value are used to store the 
//              VRML file content in a object tree.
//              Possibly the Value object references to a next Value object,
//              so you can implement multible fields.
//
//
// ===========================================================================

#ifndef _FIELDS_H_
#define _FIELDS_H_

#include <stdio.h>

// Field type
enum { FTnothing,FTfamily,FTvertexOrdering,FTshapeType,FTfaceType,FTwrapS,
       FTwrapT,FTrenderCulling,FTvalue,FTparts,FTjustification,FTmap};

class Value                                      // superclass for F... classes
{
public:
    Value();                                     // set next to NULL
    virtual ~Value();                            // abstract function

    virtual void print(FILE * fd);               // abstract function
    virtual void printLn(FILE * fd,int indent);  // prints a carrige return

    Value * next;   // pointer to the next Value object of the same type
};

//-----------------------------------------------------------------------------

class FBitMask : public Value       
{
public:
    FBitMask(int anumber);
    virtual void print(FILE * fd);   // function which prints constants strings
                                     // depending on the bit mask type
    int number;        
    int  type;                       // see enum.h
};

//-----------------------------------------------------------------------------

class FBool : public Value
{
public:
    FBool(int anumber);
    virtual void print(FILE * fd);
    
    int number;                      // true or false
};

//-----------------------------------------------------------------------------

class FEnum : public Value           
{
public:
    FEnum(int anumber);
    virtual void print(FILE * fd);   // function which prints constants strings
                                     // depending on the enum type
    int number;
    int  type;
};

//-----------------------------------------------------------------------------

class FFloat : public Value
{
public:
    FFloat(float anumber);
    virtual void print(FILE * fd);
    virtual void printLn(FILE * fd,int indent);

    float number;
};

//-----------------------------------------------------------------------------

class FImage : public Value        // used to store a image (width, height and 
{                                  // number of components in the image
public:
    FImage(int anumber,Value * anext);
    virtual void print(FILE * fd);

    int number;
};

//-----------------------------------------------------------------------------

class FLong : public Value
{
public:
    FLong(int anumber);
    virtual void print(FILE * fd);
    virtual void printLn(FILE * fd,int indent);

    int number;
};

//-----------------------------------------------------------------------------

class FMatrix : public Value
{
public:
    FMatrix(float anumber11,float anumber12,float anumber13,float anumber14,
	    float anumber21,float anumber22,float anumber23,float anumber24,
	    float anumber31,float anumber32,float anumber33,float anumber34,
            float anumber41,float anumber42,float anumber43,float anumber44);
    virtual void print(FILE * fd);

    float number11;
    float number12;
    float number13;
    float number14;
    float number21;
    float number22;
    float number23;
    float number24;
    float number31;
    float number32;
    float number33;
    float number34;
    float number41;
    float number42;
    float number43;
    float number44;
};

//-----------------------------------------------------------------------------

class FRotation : public Value
{
public:
    FRotation(float ax,float ay,float az,float aangle);
    virtual void print(FILE * fd);

    float x;
    float y;
    float z;
    float angle;
};

//-----------------------------------------------------------------------------

class FString : public Value
{
public:
    FString(char * astring);
    ~FString();
    virtual void print(FILE * fd);

    char * string;
};

//-----------------------------------------------------------------------------

class FVec2f : public Value
{
public:
    FVec2f(float anumber1,float anumber2);
    virtual void print(FILE * fd);

    float x; 
    float y;
};

//-----------------------------------------------------------------------------

class FVec3f : public Value
{
public:
    FVec3f(float anumber1,float anumber2,float anumber3);
    virtual void print(FILE * fd);

    float x;
    float y;
    float z;
};

//-----------------------------------------------------------------------------

class Field
{
public:
    Field(int aid, Value * avalue);
    virtual ~Field();
    virtual void print(FILE * fd,int indent);

    int id;         // see enum.h
    Value * value;  // pointer to a Value object list
    Field * next;   // next Field object of same or other id
};

#endif


