// ===========================================================================
//  Copyright 1996 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       global
//  TYPE:       c++ code
//  PROJECT:    lodestar
//  CONTENT:    global variables and functions and some basic classes
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rs      Reinhard Sainitzer
//		hb      Herbert Buchegger
// ===========================================================================
//  HISTORY:
//
// ===========================================================================

#include "global.h"
#include <math.h>
#include <string.h>

//----------------------------------------------------------------------------
//   global exported variables
//----------------------------------------------------------------------------

int global_indent = 4;
int global_tab = 8;
int global_noindent = 0;
int global_floatcpl = 0;
int global_longcpl = 0;
int global_cpl = 0;

//----------------------------------------------------------------------------
//   class Vec 
//----------------------------------------------------------------------------

//---Vec::Vec-----------------------------------------------------------------
// Input :
// Output:

Vec::Vec()
{
    x=0;
    y=0;
    z=0;
}

//---Vec::Vec-----------------------------------------------------------------
// Input : float ax, ay, az    : init
// Output:

Vec::Vec(float ax, float ay, float az)
{
    x=ax;
    y=ay;
    z=az;
}

//---Vec::getLength-----------------------------------------------------------
// Input :
// Output: float               : vector length

float Vec::getLength()
{
    return sqrt(x*x + y*y + z*z);
}

//---Vec::isNullVector--------------------------------------------------------
// Input :
// Output: int                 : TRUE or FALSE

int Vec::isNullVector()
{
    return (x==0 && y==0 && z==0);
}

//---Vec operator /= ---------------------------------------------------------

Vec &Vec::operator /=(float f)
{
    x/=f;
    y/=f;
    z/=f;

    return *this;
}

//---Vec operator + ----------------------------------------------------------

Vec operator +(const Vec &v1, const Vec &v2)
{
    Vec v;
    v.x = v1.x+v2.x;
    v.y = v1.y+v2.y;
    v.z = v1.z+v2.z;
    return v;
}

//---Vec operator - ----------------------------------------------------------

Vec operator -(const Vec &v1, const Vec &v2)
{
    Vec v;
    v.x = v1.x-v2.x;
    v.y = v1.y-v2.y;
    v.z = v1.z-v2.z;
    return v;
}

//----------------------------------------------------------------------------
//   class Matrix 
//----------------------------------------------------------------------------

//---Matrix::Matrix-----------------------------------------------------------
// Input :
// Output:

Matrix::Matrix()
{
    int row,col;
    for (row=0; row<4; row++)
        for (col=0; col<4; col++)
            if (row==col) e[row][col]=1; 
                else e[row][col]=0;
}

//---Matrix::Matrix-----------------------------------------------------------
// Input : float e11 - e44      : init
// Output:

Matrix::Matrix(float e11,float e12,float e13,float e14,
               float e21,float e22,float e23,float e24,
               float e31,float e32,float e33,float e34,
               float e41,float e42,float e43,float e44)
{
    e[0][0]=e11; e[0][1]=e12; e[0][2]=e13; e[0][3]=e14;
    e[1][0]=e21; e[1][1]=e22; e[1][2]=e23; e[1][3]=e24;
    e[2][0]=e31; e[2][1]=e32; e[2][2]=e33; e[2][3]=e34;
    e[3][0]=e41; e[3][1]=e42; e[3][2]=e43; e[3][3]=e44;
}

//---Matrix operator *--------------------------------------------------------

Matrix operator *(Matrix const &m1, const Matrix &m2)
{
    Matrix m;
    int row,col,i;
    for (row=0; row<4; row++)
        for (col=0; col<4; col++)
        {
            m.e[row][col]=0;
            for (i=0; i<4; i++)
                m.e[row][col]+=m1.e[row][i]*m2.e[i][col];
        }
    return m; 
}

//---Matrix friend function mult3x3-------------------------------------------

Vec mult3x3(Vec const &v, const Matrix &m)
{
    Vec r;

    r.x = v.x*m.e[0][0] + v.y*m.e[1][0] + v.z*m.e[2][0];
    r.y = v.x*m.e[0][1] + v.y*m.e[1][1] + v.z*m.e[2][1];
    r.z = v.x*m.e[0][2] + v.y*m.e[1][2] + v.z*m.e[2][2];

    return r;
}

//----------------------------------------------------------------------------
//   class Bounds
//----------------------------------------------------------------------------

//---Bounds::Bounds-----------------------------------------------------------
// Input : float axmin - azmax     : init
// Output:

Bounds::Bounds(float axmin, float aymin, float azmin,
               float axmax, float aymax, float azmax)
{
    setTo(axmin,aymin,azmin, axmax,aymax,azmax);
}

//---Bounds::setTo------------------------------------------------------------
// Input : float axmin - azmax     : init
// Output:

void Bounds::setTo(float axmin, float aymin, float azmin,
	           float axmax, float aymax, float azmax)
{
    xmin = axmin;
    ymin = aymin;
    zmin = azmin;
    xmax = axmax;
    ymax = aymax;
    zmax = azmax; 
}

//---Bounds::enlargeIfOutside-------------------------------------------------
// Input : float x,y,z            : 3d vector
// Output:
// if (x,y,z) is outside the cube defined by (xmin,ymin,zmin) and 
// (xmax,ymax,zmax) the cube is enlarged appropriate

void Bounds::enlargeIfOutside(float x, float y, float z)
{
    if (x<xmin) xmin = x;
    if (y<ymin) ymin = y;
    if (z<zmin) zmin = z;
    if (x>xmax) xmax = x;
    if (y>ymax) ymax = y;
    if (z>zmax) zmax = z;
}

//---Bounds::makeCubic--------------------------------------------------------
// Input :
// Output:
// if the edges of the cube doesn't have the same lengths this function
// enlarges the cube so that all edges have the same lengths 
// (the new enlarged cube have the same midpoint)

void Bounds::makeCubic()
{
    float xmid = (xmin+xmax) / 2;
    float ymid = (ymin+ymax) / 2;
    float zmid = (zmin+zmax) / 2;

    float xd = xmax-xmid;
    float yd = ymax-ymid;
    float zd = zmax-zmid;

    float d = xd;
    if (yd>d) d=yd;
    if (zd>d) d=zd;

    if (xmid+d>xmax) xmax=xmid+d;
    if (xmid-d<xmin) xmin=xmid-d;

    if (ymid+d>ymax) ymax=ymid+d;
    if (ymid-d<ymin) ymin=ymid-d;

    if (zmid+d>zmax) zmax=zmid+d;
    if (zmid-d<zmin) zmin=zmid-d;
}

//---Bounds::getMaxElongation-------------------------------------------------
// Input :
// Output: float                  : 
// get the maximum elongation 
// (the length of the edge of the cube with the maximum length)

float Bounds::getMaxElongation()
{
    float xd = xmax-xmin;
    float yd = ymax-ymin;
    float zd = zmax-zmin;

    float d = xd;
    if (yd>d) d=yd;
    if (zd>d) d=zd;

    return d;
}

//---printIndent--------------------------------------------------------------
// Input : FILE *fd          : VRML output file descriptor
//         int indent        : count of blanks
// Output:

void printIndent(FILE *fd, int indent)
{
    if (global_noindent) return;

    int i;
    for (i=indent/global_tab; i>0; i--) fprintf(fd,"\t");
    for (i=indent%global_tab; i>0; i--) fprintf(fd," ");
}

//---printFloat---------------------------------------------------------------
// Input : FILE * fd         : VRML output file descriptor
//         double value      : variable to print
// Output:

void printFloat(FILE *fd,double value)
{
    char str[100];
    int i;
    sprintf(str,"%f",value);
    for (i=strlen(str)-1;i>=0 && str[i]=='0';i--) str[i]=0;
    if (str[i]=='.') str[i]=0;

    fprintf(fd,"%s",str);
}
