// ===========================================================================
//  Copyright 1996 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       gnodes
//  TYPE:       c++ code
//  PROJECT:    lodestar
//  CONTENT:    implementation of class State, class Node, class UseNode,
//              class GroupNode, class RootNode, class SeparatorNode,
//              class TransformSeparatorNode, class UnknownNode,
//              class PerspectiveCameraNode, class OrthographicCameraNode
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rs      Reinhard Sainitzer
//		hb      Herbert Buchegger
// ===========================================================================
//  HISTORY:
//
// ===========================================================================

#include "nodes.h"
#include "global.h"
#include "enum.h"
#include "tri.h"
#include <limits.h>
#include <assert.h>

//----------------------------------------------------------------------------
//   class State
//----------------------------------------------------------------------------
//This class is used to store the current state while traversing the 
//VRML node tree with method setupDepends 
//with the informations in this object e.g. the FaceSet is able to find
//the right Coordinate3,Normal,... node and the actual bindings .

//---State::State-------------------------------------------------------------
// Input :
// Output:

State::State()
{
    lastCoordNode = NULL;
    lastMaterialNode = NULL;
    lastNormalNode = NULL;
    lastTextureCoordNode = NULL;
    lastTextureNode = NULL;

    currMaterialBinding = VRML_DEFAULT;
    currNormalBinding = VRML_DEFAULT;
    currVertexOrdering = VRML_UNKNOWN_ORDERING;
    currFaceType = VRML_CONVEX;

    currTransMatrix = Matrix();
}

//---State::getMaxScale-------------------------------------------------------
// Input :
// Output:
//get the maximum of the current scale in 3D
//it calculates the lengthes of the 4 diagonals of an unit cube (1,1,1) 
//transformed by the current transformation (currTransMatrix)
//the max length divided by the length of an unit cube (sqrt(3)) is returned

float State::getMaxScale()
{
    Vec p000 = mult3x3(Vec(0,0,0), currTransMatrix);
    Vec p100 = mult3x3(Vec(1,0,0), currTransMatrix);
    Vec p010 = mult3x3(Vec(0,1,0), currTransMatrix);
    Vec p110 = mult3x3(Vec(1,1,0), currTransMatrix);
    Vec p001 = mult3x3(Vec(0,0,0), currTransMatrix);
    Vec p101 = mult3x3(Vec(1,0,1), currTransMatrix);
    Vec p011 = mult3x3(Vec(0,1,1), currTransMatrix);
    Vec p111 = mult3x3(Vec(1,1,1), currTransMatrix);
 
    Vec diagonal;
    float len;
    float maxLength = 0;
    
    diagonal = p111-p000;
    len = diagonal.getLength();
    if (len>maxLength) maxLength=len;

    diagonal = p101-p010;
    len = diagonal.getLength();
    if (len>maxLength) maxLength=len;

    diagonal = p110-p001;
    len = diagonal.getLength();
    if (len>maxLength) maxLength=len;

    diagonal = p011-p100;
    len = diagonal.getLength();
    if (len>maxLength) maxLength=len;

/*
    for (int row=0; row<4; row++)
    {
        printf("%f %f %f %f\n",
            currTransMatrix.e[row][0],
            currTransMatrix.e[row][1],
            currTransMatrix.e[row][2],
            currTransMatrix.e[row][3]);
    }
    printf("scalefactor = %f\n",maxLength/sqrt(3)); //!!! remove
*/
    return maxLength / sqrt(3);
}

//----------------------------------------------------------------------------
//   class Node
//----------------------------------------------------------------------------
//This class is used to store and handle VRML-nodes - this class is able
//to contain any type of VRML-nodes ecxept those derived from Group but
//this class can only store and print VRML-nodes therefore it is used
//directly only for VRML-nodes which are uninteresting for lod-generation
//(e.g. AsciiText, DirectionalLight, PointLight, WWWInline, ...)
//many specialized classes are derived from this class

int Node::nolods = 0;          //prevents nodes to produce lods
                               //(which are are able to this)
int Node::extractlod = -1;     //print out only the lod with number extractlod
                               //the lods are only generated up to this number
                               //-1 disables the mechanism
int Node::removedcount = 0;    //count of the nodes removed by the join alg.
int Node::joinmode = 0;        //see JM_... flags
int Node::verbosejoin = 0;     //print out information while doing the
                               //join algorithm if !=0

IdentSet *Node::identifiers = new IdentSet();  //this class stores all ident-
                                               //ifiers (=nodespec = DEF xxx)

//---Node::Node---------------------------------------------------------------
// Input : int aid           : init
//         char *afieldspec  : init
//         Field *afield     : init
//         int fline         : init
//         int lline         : init
// Output:

Node::Node(int aid, char *afieldspec, Field *afield, int fline, int lline)
{
    id = aid;
    fieldspec = afieldspec;
    field = afield;
    nodespec = NULL;
    next = NULL;
    last = NULL;
    firstline = fline;
    lastline = lline;
    used=0;
    joinflag=0;
}

//---Node::~Node--------------------------------------------------------------
// Input :
// Output:

Node::~Node()
{
    removedcount++;
    if (verbosejoin) 
    {
        message("join","deleting node %s ",nodeIdString[id]);
        if (nodespec) printf("with DEF >%s<",nodespec);
        printf("\n");
    }

    if (fieldspec) delete fieldspec;
    if (nodespec) delete nodespec;
    deleteFields();
}

//---Node::registerName-------------------------------------------------------
// Input :
// Output:
//if nodespec!=NULL this method creates an Indent-object and inserts
//it in the global IdentSet-object (Node::identifiers)

void Node::registerName()
{
    if (nodespec) identifiers->insert(new Ident(this));
}

//---Node::print--------------------------------------------------------------
// Input : FILE *fd        : VRML output file descriptor
//         int indent      : indent for output file
// Output:
//print out the node inclusive all fields to fd

void Node::print(FILE *fd, int indent)
{
    printHead(fd,indent);
    printFields(fd,indent);
    printTail(fd,indent);
}

//---Node::printHead----------------------------------------------------------
// Input : FILE *fd        : VRML output file descriptor
//         int &indent     : indention of output
// Output: int &indent     : new indention
//print out the head of the node this includes { and evt. DEF 
//indent will be incremented by global_indent

void Node::printHead(FILE *fd, int &indent)
{
    printIndent(fd,indent);
    if (nodespec) fprintf(fd,"DEF %s ",nodespec);
    fprintf(fd,"%s {\n",nodeIdString[id]);
    indent+=global_indent;

    if (fieldspec) 
    {    
        printIndent(fd,indent);
        fprintf(fd,"fields %s\n",fieldspec);
    }
}

//---Node::printTail----------------------------------------------------------
// Input : FILE *fd        : VRML output file descriptor
//         int &indent     : indention of output
// Output: int &indent     : new indention
//indent will be decremented by global_indent
//print out the tail (is '}') incl. indention to fd

void Node::printTail(FILE *fd, int &indent)
{
    indent-=global_indent;
    printIndent(fd,indent);
    fprintf(fd,"}\n");
}

//---Node::printFields--------------------------------------------------------
// Input : FILE *fd        : VRML output file descriptor
//         int indent      : indention of output
// Output:
//print out all fields stored in this node 

void Node::printFields(FILE *fd, int indent)
{
    Field *fh = field;
    while (fh!=NULL)
    {
        fh->print(fd,indent);
        fh=fh->next;
    }
}

//---Node::deleteFields-------------------------------------------------------
// Input :
// Output:
//delete all fields of this node

void Node::deleteFields()
{
    Field *fh;
    while (field!=NULL)
    {
        fh=field;
        field=field->next;
        delete fh;
    }
}

//---Node::getField-----------------------------------------------------------
// Input : int aid     : field-identifier (see enum.h)
// Output: Field *     : a pointer to the found Field-object
//get a pointer to a field object stored in this node specified by aid
//(see FID_xxx constants in enum.h)
//always use this function !
//fields may be defined more once - this function returns the last


Field *Node::getField(int aid)
{
    Field *fh = field;
    Field *last = NULL;
    while (fh)
    {
        if (aid==fh->id) last=fh;
        fh=fh->next;
    }
    return last;
}

//---Node::deleteValues-------------------------------------------------------
// Input : Value **list    : pointer to a list of object derived from Value
//         int count       : count of elements to delete
// Output: Value **list    : pointer to the remainding list
//delete fist count elements of a list of objects derived from
//class Value and skip the deleted values

void Node::deleteValues(Value **list, int count)
{
    Value *help;
    while (*list!=NULL && count>0)
    {
        help=(*list)->next;
        delete *list;
        *list=help;
        count--;
    }
}

//---Node::deleteAllValues----------------------------------------------------
// Input : Value **list    : pointer to a list of objects derived from Value
// Output: Value **list    : NULL
//delete all elements of a list of objects derived from class Value

void Node::deleteAllValues(Value **list)
{
    Value *help;
    while (*list!=NULL)
    {
        help=(*list)->next;
        delete *list;
        *list=help;
    }
}

//---Node::setupDepends-------------------------------------------------------
// Input : State *astate   : current state
// Output:

void Node::setupDepends(State *astate)
{
   //abstract
}

//---Node::setupCamera--------------------------------------------------------
// Input : Node **camera   : first found PerspectiveCamera
// Output:
// abstract - nothing to do - used for searching for the
// first PerspectiveCamera

void Node::setupCamera(Node **camera)
{
    //nothing to do
}

//---Node::markAsUsed---------------------------------------------------------
// Input :
// Output:
// mark this node as used - the method sets variable used to 1 - 
// this means that this node is used by a USE statement
// virtual - see also GroupNode::markAsUsed !

void Node::markAsUsed()
{
    //printf("def >%s< is used %s\n",nodespec,nodeIdString[id]);
    used=1;
}

//---Node::message------------------------------------------------------------
// Input : const char *section :string indicates the verbose section
//                              specified as parameter of the programm
//         const char *msg     :format string of the message
//         ...                 :arguments for format string
// Output:
// prints out a (verbose) message to stdout

void Node::message(const char *section, const char *msg, ...)
{
    va_list ap;
    va_start(ap,msg);
    if (section && section[0]) printf("%s: ",section);
    if (firstline || lastline) printf("%d-%d ",firstline,lastline);
    printf("%s: ",nodeIdString[id]);
    vprintf(msg,ap);
}

//---Node::warning------------------------------------------------------------
// Input : const char *msg     :format string of the warning message
//         ...                 :arguments for format string
// Output:
// prints out a warning message to stderr

void Node::warning(const char *msg, ...)
{
    va_list ap;
    va_start(ap,msg);
    fprintf(stderr,"warning: ");
    if (firstline || lastline) fprintf(stderr,"%d-%d ",firstline,lastline);
    fprintf(stderr,"%s: ",nodeIdString[id]);
    vfprintf(stderr,msg,ap);
}

//---Node::error--------------------------------------------------------------
// Input : const char *msg     :format string of the error message
//         ...                 :arguments for format string
// Output:
// prints out an error message to stderr
// and stops the program with exit code 1

void Node::error(const char *msg, ...)
{
    va_list ap;
    va_start(ap,msg);
    fprintf(stderr,"error: ");
    if (firstline || lastline) fprintf(stderr,"%d-%d ",firstline,lastline);
    fprintf(stderr,"%s: ",nodeIdString[id]);
    vfprintf(stderr,msg,ap);
    exit(1);
}

//---Node::join---------------------------------------------------------------
// Input :
// Output:
// join manage function
// checks if this node is joinable with the next nodes

void Node::join()
{
    joinmode=JM_NOTHING;

    Node * help = next;
    while (help!=NULL)
    {
        if (id!=help->id) help=NULL;
        else
        {

            if (!(help->joinflag&JF_REMOVEABLE) && joinable(help))
            {
                // join the 2 nodes
                joinflag|=JF_REMOVEABLE;
                prepareForJoin();
                help->prepareForJoin();
                makeJoin(help);
                help->firstline=help->lastline=0;
                help=NULL;
            }
            else help=help->next;
        }
    }
}

//---Node::prepareForJoin-----------------------------------------------------
// Input :
// Output:
// virtual abstract function

void Node::prepareForJoin()
{
}

//---Node::joinable-----------------------------------------------------------
// Input : Node *node      : second node
// Output: int             :
// used for the most nodes (UseNode,UnknownNode,... are not joinable)
// this function returns always FALSE

int Node::joinable(Node * node)
{
    return(0);
}

//---Node::makeJoin-----------------------------------------------------------
// Input : Node * node     : not used
// Output:
// virtual abstract function

void Node::makeJoin(Node * node)
{
}

//---Node::compareField-------------------------------------------------------
// Input : Field * f1      : first field for compare, see field->value
//         Field * f2      : second field for compare
// Output: int             :
// compare a list of FFloat objects

int Node::compareField(Field * f1, Field *f2)
{
    if (!f1 && !f2) return(1);
    else if (f1 && f2)
    {
        FFloat * v1 = ((FFloat *)f1->value);
        FFloat * v2 = ((FFloat *)f2->value);

        while (v1!=NULL && v2!=NULL)
        {
            if (v1->number!=v2->number) return(0);
            v1=(FFloat *)(v1->next);
            v2=(FFloat *)(v2->next);
            if (v1==NULL && v2==NULL) return(1);
            if (v1==NULL || v2==NULL) return(0);
        }
    }

    return(0);
}

//---Node::compare3DField-----------------------------------------------------
// Input : Field * f1      : first field for compare, see field->value
//         Field * f2      : second field for compare
// Output: int             :
// compare a list of FVec3f objects

int Node::compare3DField(Field * f1, Field *f2)
{
    if (!f1 && !f2) return(1);
    else if (f1 && f2)
    {
        FVec3f * v1=(FVec3f *)(f1->value);
        FVec3f * v2=(FVec3f *)(f2->value);

        while (v1!=NULL && v2!=NULL)
        {
            if (v1->x!=v2->x || v1->y!=v2->y || v1->z!=v2->z) return 0;
            v1=(FVec3f *)(v1->next);
            v2=(FVec3f *)(v2->next);
            if (v1==NULL && v2==NULL) return(1);
            if (v1==NULL || v2==NULL) return(0);
        }
    }
    return(0);
}

//---Node::joinFields---------------------------------------------------------
// Input : Field * f1   : Field->value list will be append to the second list
//         Field * f2   : field of the second list
// Output:
// joinFields appends the list from the second field to the list
// from the first list and sets this new list as list from the second list

void Node::joinFields(Field * f1,Field * f2)
{
    if (!f1) return;
    else if (f1 && !f2) return;
    else
    {
       if (f1->value && !f2->value)
       {
           f2->value=f1->value;
           f1->value=NULL;
       }
       else if (f1->value && f2->value)
       {
           FLong * help = (FLong *)(f1->value);
           while (help->next) help=(FLong *)(help->next);
           help->next=f2->value;
           f2->value=f1->value;
           f1->value=NULL;
       }
    }
}

//---Node::makeJoinField------------------------------------------------------
// Input : Node * node      : the second node for join
//         int fid          : the id's of the 2 nodes
// Output:
// makeJoinField appends the list of the second field to the list
// of the first field and uses this new list as list of the second
// field. If the field id is FID_SHININESS a default value will be used
// if there is no value present.

void Node::makeJoinField(Node * node,int fid)
{
    Field * f1 = getField(fid);
    Field * f2 = node->getField(fid);

    if (f1==NULL)
    {
        // default value of shininess is 0.2
        if (fid==FID_SHININESS) f1 = new Field(fid,new FFloat(0.2));
        else f1 = new Field(fid,new FFloat(0));
        f1->next=field;
        field=f1;
    }
    if (f2==NULL)
    {
        // default value of shininess is 0.2
        if (fid==FID_SHININESS) f2 = new Field(fid,new FFloat(0.2));
        else f2 = new Field(fid,new FFloat(0));
        f2->next=node->field;
        node->field=f2;
    }

    if (f2->value!=NULL)
    {
        Value * help = f1->value;
        while (help->next!=NULL) help=help->next;
        help->next=f2->value;
        f2->value=f1->value;
        f1->value=NULL;
    }
}

//---Node::makeJoinField3D----------------------------------------------------
// Input : Node * node      : the second node for join
//         int fid          : the id's of the 2 nodes
// Output:
// makeJoinField3D appends the list of the second field to the list
// of the first field and uses this new list as list of the second
// field. If the field id is FID_AMBIENTCOLOR or FID_DIFFUSECOLOR a 
// default value will be used if there is no value present.

void Node::makeJoinField3D(Node * node,int fid)
{
    Field * f1 = getField(fid);
    Field * f2 = node->getField(fid);

    if (f1==NULL)
    {
        if (fid==FID_AMBIENTCOLOR) f1 = new Field(fid,new FVec3f(0.2,0.2,0.2));
        else if (fid==FID_DIFFUSECOLOR) 
                             f1 = new Field(fid,new FVec3f(0.8,0.8,0.8));
        else f1 = new Field(fid,new FVec3f(0,0,0));
        f1->next=field;
        field=f1;
    }
    if (f2==NULL)
    {
        if (fid==FID_AMBIENTCOLOR) f2 = new Field(fid,new FVec3f(0.2,0.2,0.2));
        else if (fid==FID_DIFFUSECOLOR) 
                             f2 = new Field(fid,new FVec3f(0.8,0.8,0.8));
        else f2 = new Field(fid,new FVec3f(0,0,0));
        f2->next=node->field;
        node->field=f2;
    }

    if (f2->value!=NULL)
    {
        Value * help = f1->value;
        while (help->next!=NULL) help=help->next;
        help->next=f2->value;
        f2->value=f1->value;
        f1->value=NULL;
    }
}

//---Node::printTree----------------------------------------------------------
// Input : int indent      : indention of output file
// Output:

void Node::printTree(int indent)
{
    printf("\n");
    printf("%4d-%4d ",firstline,lastline);
    printIndent(stdout,indent);
    if (id!=0) printf("%s",nodeIdString[id]);
    else printf("other ...\n");
    if (nodespec) printf(" DEF %s",nodespec);
}

//----------------------------------------------------------------------------
//   class UseNode : Node
//----------------------------------------------------------------------------
//used for VRML-node USE 

//---UseNode::UseNode---------------------------------------------------------
// Input : const char *akeyword : init
//         int fline            : init
//         int lline            : init
// Output:

UseNode::UseNode(const char *akeyword, int fline, int lline) :
         Node(NID_UNDEFINED,NULL,NULL,fline,lline)
{
    keyword = akeyword;
}

//---UseNode::print-----------------------------------------------------------
// Input : FILE *fd    : VRML output file descriptor
//         int indent  : indention of output
// Output:
// prints out the USE-node to fd

void UseNode::print(FILE *fd, int indent)
{
    printIndent(fd,indent);
    fprintf(fd,"USE %s\n",keyword);
}

//---UseNode::setupDepends----------------------------------------------------
// Input : State *astate   : current state
// Output:
// checks if a DEF exists for this USE
// calls setupDepends of the DEF-node (tree)
// marks DEF-node (tree) as used

void UseNode::setupDepends(State *astate)
{
    Ident *ident = identifiers->search(keyword);
    if (ident==NULL) error("USE identifier >%s< not defined\n",keyword);

    ident->node->setupDepends(astate);
    ident->node->markAsUsed();
}

//----------------------------------------------------------------------------
//   class GroupNode : Node
//----------------------------------------------------------------------------
//used for VRML-node Group and is superclass for Separator,TransformSeparator
//is able to contain subnodes

//---GroupNode::GroupNode-----------------------------------------------------
// Input : int aid           : init
//         char *afieldspec  : init
//         Field *afield     : init
//         Node asub         : init 
//         int fline         : init
//         int lline         : init
// Output:

GroupNode::GroupNode(int aid, char *afieldspec, Field *afield, Node *asub,
                     int fline, int lline) :
           Node(aid,afieldspec,afield,fline,lline)
{
    sub = asub;
}

//---GroupNode::~GroupNode----------------------------------------------------
// Input :
// Output:

GroupNode::~GroupNode()
{
    Node *nh;
    while (sub!=NULL)
    {
        nh=sub;
        sub=sub->next;
        delete nh;
    }
}

//---GroupNode::print---------------------------------------------------------
// Input : FILE *fd       : VRML output file descriptor
//         int indent     : indention of output
// Output:
// prints out the GroupNode and all subnodes - uses printSubNodes

void GroupNode::print(FILE *fd, int indent)
{
    printHead(fd,indent);
    printFields(fd,indent);
    printSubNodes(fd,indent);
    printTail(fd,indent);
}

//---GroupNode::printSubNodes-------------------------------------------------
// Input : FILE *fd       : VRML output file descriptor
//         int indent     : indention of output
// Output:
// prints out all subnodes of this GroupNode

void GroupNode::printSubNodes(FILE *fd, int indent)
{
    Node *nh = sub;
    while (nh!=NULL)
    {
        nh->print(fd,indent);
        nh->deleteFields();
        nh=nh->next;
    }
}

//---GroupNode::setupDepends--------------------------------------------------
// Input : State *astate  : current state
// Output: State *astate  : new state
// calls setupDepends for all subnodes

void GroupNode::setupDepends(State *astate)
{
    Node *nh = sub;
    while(nh!=NULL)
    {
        nh->setupDepends(astate);
        nh=nh->next;
    }
}

//---GroupNode::setupCamera---------------------------------------------------
// Input : Node **camera  : first found PerspectiveCamera
// Output: Node **camera  : first found PerspectiveCamera
// used for searching for the first PerspectiveCamera
// searches in all subnodes of this GroupNode

void GroupNode::setupCamera(Node **camera)
{
    Node *nh = sub;
    while (nh!=NULL)
    {
        nh->setupCamera(camera);
        nh=nh->next;
    }
}

//---GroupNode::markAsUsed----------------------------------------------------
// Input : 
// Output:
// marks this node and all subnodes as used (variable used=1)
// this means this node-tree is used by a USE-node

void GroupNode::markAsUsed()
{
    Node::markAsUsed();
    Node *nh = sub;
    while(nh!=NULL)
    {
        nh->markAsUsed();
        nh=nh->next;
    }
}

//---GroupNode::join----------------------------------------------------------
// Input :
// Output:
// this function makes recursive calls for all for sub lists. So it is 
// possible, that every node can be compared with the next nodes. If there are
// 2 sub lists, it is also necessary to compare this lists. 
// If 2 nodes or 2 lists (with sub lists) are joinable (there are the same
// nodes - only Separator,Normal,Texture,Coord,IndexedFaceSet,Group,Binding,
// IndexedLineSet nodes are allowed - in the same order) then join this two
// lists to one list.
// If there is a Separator node on the end of a list, then this node will be
// removed (the sub list will be append tho the current next list).
//
//----------------------------------------------------------------------------
// example:
//
// Separator                                  original tree
//     Separator                     <-       this Separator join 
//         Separator                 <=       this 2 lists
//             Material
//             IndexedFaceSet
//         Separator                 <=
//             Material
//             IndexedFaceSet
//     Separator
//         Separator
//             Material
//             IndexedFaceSet
//----------------------------------------------------------------------------
// Separator
//     Separator                     <-       this Separator removes
//         Separator                 <=       this not necessary Separator
//             Material              
//             IndexedFaceSet
//     Separator
//         Separator
//             Material
//             IndexedFaceSet
//----------------------------------------------------------------------------
// Separator                                  
//     Separator
//         Material    
//         IndexedFaceSet
//     Separator                     <-       this Separator removes
//         Separator                 <=       this not necessary Separator
//             Material                    
//             IndexedFaceSet
//----------------------------------------------------------------------------
// Separator                         <-       this Separator join   
//     Separator                     <=       this 2 lists
//         Material    
//         IndexedFaceSet
//     Separator                     <=
//         Material   
//         IndexedFaceSet
//----------------------------------------------------------------------------
// Separator                         <-       this Separator removes
//     Separator                     <=       this not necessary Separator
//         Material
//         IndexedFaceSet
//----------------------------------------------------------------------------
// Separator
//     Material    
//     IndexedFaceSet
//----------------------------------------------------------------------------
//
// The GroupNode::join function removes nodes from the current sub list when 
// the joinflag&JF_REMOVEABLE is TRUE.
// (this flag will be set in makeJoin, when a node will be joined with a
// second one, the first node can be removed)
  
void GroupNode::join()
{
    if (id==NID_LOD || id==NID_LEVELOFDETAIL || id==NID_SWITCH) return;

    if (joinflag&JF_RUNNING) Node::join();
    else if (!(joinflag&JF_REMOVEABLE))
    {
        joinflag|=JF_RUNNING;

        Node *nh = sub;
        while(nh!=NULL)
        {
            if (nh->id==NID_SEPARATOR || nh->id==NID_GROUP) nh->join();
            nh=nh->next;
        }

        nh = sub;
        while(nh!=NULL)
        {
            nh->join();
            nh=nh->next;
        }

        // if the last node is a Separator, then it is removeable
        // in this case the sub list must append to the current next list
        nh = sub;
        if (nh!=NULL)
        {
            while (nh->next!=NULL) nh=nh->next;

            if ((nh->id==NID_SEPARATOR || nh->id==NID_GROUP) &&
                !(nh->joinflag&JF_NOT_REMOVEABLE) &&
                !(nh->used))
            {
                nh->next=((GroupNode *)nh)->sub;
                ((GroupNode *)nh)->sub=NULL;
                    nh->joinflag|=JF_REMOVEABLE;
            }
        }

        // remove removeable nodes
        while (sub && sub->joinflag&JF_REMOVEABLE)
        {
            // delete node
            Node * help = sub;
            sub=sub->next;
            delete(help);
        }
        nh=sub;
        while(nh!=NULL && nh->next!=NULL)
        {
            if (nh->next->joinflag&JF_REMOVEABLE)
            {
                // delete node
                Node * help = nh->next;
                nh->next=nh->next->next;
                delete(help);
            }
            else nh=nh->next;
        }

    }
}

//---GroupNode::prepareForJoin------------------------------------------------
// Input : int flag      : joinflag
//         int mode      : joinmode
// Output:
// It is necessary to prepare a node for joining:
// Node : Material
//            Insert a material binding node with binding per face indexed for
//                every indexed face set node
//            Create a material index for every indexed face set node
//            Do it also in the sub nodes
//        Texture, Normal, Coord
//            Set pointcount in coord, normal or texturecorrd node
//            Do it also in the sub nodes

void GroupNode::prepareForJoin(int flag,int mode)
{
    // material
    if (flag==JF_PREPARED_MATERIAL && !(joinflag&JF_PREPARED_MATERIAL) && 
        joinmode&mode)
    {
        joinflag|=flag;

        if (sub!=NULL)
        {
            // Group and Separator node recursive call
            if (sub->id==NID_SEPARATOR || sub->id==NID_GROUP)
                ((GroupNode *)sub)->prepareForJoin(flag,mode);

            if (sub->id==NID_INDEXEDFACESET && 
                sub->getField(FID_MATERIALINDEX)==NULL)
            {
                // insert Binding Node with binding per face indexed
                FEnum * fenum = new FEnum(VRML_PER_FACE_INDEXED);
                fenum->type=FTvalue;

                //insert binding node
                MaterialBindingNode * mbn = new MaterialBindingNode(
                    NID_MATERIALBINDING,
                    NULL,
                    new Field(FID_VALUE,fenum),0,0);
                mbn->joinflag|=JF_PREPARED_MATERIAL;
                mbn->next=sub;
                sub=mbn;
                ((ShapeNode *)sub)->state.lastMaterialNode->createMaterialIndex(sub,0);
            }

            Node * help=sub;
            while (help->next!=NULL)
            {
	        // Group and Separator node recursive call
                if (help->next->id==NID_SEPARATOR || help->next->id==NID_GROUP)
                    ((GroupNode *)help->next)->prepareForJoin(flag,mode);

                if (help->next->id==NID_INDEXEDFACESET &&
                    help->next->getField(FID_MATERIALINDEX)==NULL)
                {
                    FEnum * fenum = new FEnum(VRML_PER_FACE_INDEXED);
                    fenum->type=FTvalue;
                    // insert Binding Node with binding per face indexed
                    MaterialBindingNode * mbn = new MaterialBindingNode(
                        NID_MATERIALBINDING,
                        NULL,
                        new Field(FID_VALUE,fenum),0,0);
                    mbn->joinflag|=JF_PREPARED_MATERIAL;
                    mbn->next=help->next;
                    help->next=mbn;
                    help=help->next;

                    ((ShapeNode *)help->next)->state.lastMaterialNode->createMaterialIndex(help->next,0);
                }
                help=help->next;
            }
        }
    }
    // normal, coord, texture
    else if ( ( (flag==JF_PREPARED_NORMAL && !(joinflag&JF_PREPARED_NORMAL) ) ||
                (flag==JF_PREPARED_COORD && !(joinflag&JF_PREPARED_COORD) ) || 
                (flag==JF_PREPARED_TEXTURE && !(joinflag&JF_PREPARED_TEXTURE) ) )  && 
              joinmode&mode)
    {
        joinflag|=flag;

        // set pointcount in coord, normal or texturecorrd node
        Node * help = sub;
        while (help!=NULL)
        {
            if (mode&JM_TEXTURE && help->id==NID_TEXTURE2)
            {
                Field * f = help->getField(FID_VECTOR);
                ((TextureCoordNode *)help)->pointcount=0;
                if (f!=NULL)
                {
                    Value * v = f->value;
                    while (v!=NULL)
                    {
                        ((TextureCoordNode *)help)->pointcount++;
                        v=v->next;
                    }
                }
            }
            if (mode&JM_NORMAL && help->id==NID_NORMAL)
            {
                Field * f = help->getField(FID_VECTOR);
                ((NormalNode *)help)->pointcount=0;
                if (f!=NULL)
                {
                    Value * v = f->value;
                    while (v!=NULL)
                    {
                        ((NormalNode *)help)->pointcount++;
                        v=v->next;
                    }
                }
            }
            if (mode&JM_COORD && help->id==NID_COORDINATE3)
            {
                Field * f = help->getField(FID_POINT);
                ((CoordNode *)help)->pointcount=0;
                if (f!=NULL)
                {
                    Value * v = f->value;
                    while (v!=NULL)
                    {
                        ((CoordNode *)help)->pointcount++;
                        v=v->next;
                    }
                }
            }
            if (help->id==NID_SEPARATOR || help->id==NID_GROUP)
                ((GroupNode *)help)->prepareForJoin(flag,mode);
            help=help->next;
        }
    }
}

//---GroupNode::prepareForJoin------------------------------------------------
// Input :
// Output:
// prepare Material, Normal, Coord and TextureCoord nodes

void GroupNode::prepareForJoin()
{
    if (joinmode&JM_MATERIAL) prepareForJoin(JF_PREPARED_MATERIAL,JM_MATERIAL);
    if (joinmode&JM_NORMAL)   prepareForJoin(JF_PREPARED_NORMAL  ,JM_NORMAL);
    if (joinmode&JM_COORD)    prepareForJoin(JF_PREPARED_COORD   ,JM_COORD);
    if (joinmode&JM_TEXTURE)  prepareForJoin(JF_PREPARED_TEXTURE ,JM_TEXTURE);
}

//---GroupNode::joinable------------------------------------------------------
// Input : Node * node    :
// Output: int            :
// checks if the sub lists are joinable
// Trick: If the list was modified by a prepareForJoin function, a MaterialBinding
// node is maybe too much. In this case this node must be skipped.

int GroupNode::joinable(Node * node)
{
    if (id!=node->id) return(0);
    if (used || node->used) return(0);

    Node *n1 = sub;
    Node *n2 = ((GroupNode *)node)->sub;
    while(n1!=NULL && n2!=NULL)
    {
        while (n2!=NULL &&
               (n2->id==NID_MATERIALBINDING) &&
               (n2->joinflag&JF_PREPARED_MATERIAL))
            n2=n2->next;
        while (n1!=NULL &&
               (n1->id==NID_MATERIALBINDING) &&
               (n1->joinflag&JF_PREPARED_MATERIAL))
            n1=n1->next;
        if (n1==NULL && n2!=NULL) return(0);
        if (n1!=NULL && n2==NULL) return(0);
        if (n1==NULL && n2==NULL) return(1);

        //printf("used %d %d\n",n1->used,n2->used);

        if (!(n1->joinable(n2))) return(0);
        n1=n1->next;
        n2=n2->next;
    }
    if (n1 || n2) return(0);

    return(1);
}

//---GroupNode::makeJoin------------------------------------------------------
// Input : Node *node     :
// Output:
// call function recursive

void GroupNode::makeJoin(Node * node)
{
    Node::makeJoin(node);

    Node *n1 = sub;
    Node *n2 = ((GroupNode *)node)->sub;
    while(n1!=NULL && n2!=NULL)
    {
        n1->makeJoin(n2);
        n2->firstline=n2->lastline=0;  // delete the line information
        n1=n1->next;
        n2=n2->next;
    }
}

//---GroupNode::printTree-----------------------------------------------------
// Input : int indent      : indention of output
// Output:

void GroupNode::printTree(int indent)
{
    Node::printTree(indent);

    Node *nh = sub;
    while (nh!=NULL)
    {
        nh->printTree(indent+4);
        nh=nh->next;
    }
}

//----------------------------------------------------------------------------
//   class RootNode : GroupNode : Node
//----------------------------------------------------------------------------
//has no VRML dendant - used to store the VRML-header and is root of all
//VRML-nodes in a file

//---RootNode::RootNode-------------------------------------------------------
// Input : const char *aidentification  : VRML-header (e.g. "#VRML 1.0 ascii")
//         Node *asub                   : the VRML node tree
// Output:

RootNode::RootNode(const char *aidentification, Node *asub) :
          GroupNode(NID_UNDEFINED,NULL,NULL,asub,0,0)
{
    identification = aidentification;
}

//---RootNode::print----------------------------------------------------------
// Input : FILE *fd       : VRML output 
//         int indent     :
// Output:
// prints out a VRML-header and the node tree

void RootNode::print(FILE *fd, int indent)
{
    fprintf(fd,"%s\n",identification);
    printSubNodes(fd,indent);
}

//---RootNode::join-----------------------------------------------------------
// Input :
// Output:
// Set the first Separator as not removeable

void RootNode::join()
{
    if (sub!=NULL && sub->id==NID_SEPARATOR && sub->next==NULL)
        sub->joinflag=JF_NOT_REMOVEABLE;
    
    GroupNode::join();
}

//----------------------------------------------------------------------------
//   class SeparatorNode : GroupNode : Node
//----------------------------------------------------------------------------
//used for VRML-node Separator - setupDepends restores State 
//after traversing the subnodes

//---SeparatorNode::SeparatorNode---------------------------------------------
// Input : int aid           : init
//         char *afieldspec  : init
//         Field *afield     : init
//         Node *asub        : init
//         int fline         : init
//         int lline         : init
// Output:

SeparatorNode::SeparatorNode(int aid, char *afieldspec, Field *afield,
                             Node *asub, int fline, int lline) :
               GroupNode(aid,afieldspec,afield,asub,fline,lline)
{
}

//---SeparatorNode::setupDepends----------------------------------------------
// Input : State *astate     : current state
// Output:
// saves the current state
// calls setupDepends for all subnodes
// restores the state

void SeparatorNode::setupDepends(State *astate)
{
    State oldState = *astate; 
    GroupNode::setupDepends(astate);
    *astate = oldState;
}

//----------------------------------------------------------------------------
//   class TransformSeparatorNode : GroupNode : Node
//----------------------------------------------------------------------------
//used for VRML-node TransformSeparator - setupDepends restores
//transformation matrix (State::currTransMatrix) after traversing the subnodes

//---TransformSeparatorNode::TransformSeparatorNode---------------------------
// Input : int aid          : init
//         char *afieldspec : init
//         Field *afield    : init
//         Node *asub       : init
//         int fline        : init
//         int lline        : init
// Output:

TransformSeparatorNode::TransformSeparatorNode(int aid, char *afieldspec,
                                               Field *afield, Node *asub,
                                               int fline, int lline) :
                        GroupNode(aid,afieldspec,afield,asub,fline,lline)
{
}

//---TransformSeparatorNode::setupDepends-------------------------------------
// Input : State *astate    : current state
// Output:
// saves State::currTransMatrix
// calls setupDepends for all subnodes
// restores State::currTransMatrix

void TransformSeparatorNode::setupDepends(State *astate)
{
    Matrix oldTransMatrix = astate->currTransMatrix; 
    GroupNode::setupDepends(astate);
    astate->currTransMatrix = oldTransMatrix;

    /*
    for (int row=0; row<4; row++)
    {
        message("","%f %f %f %f\n",
            astate->currTransMatrix.e[row][0],
            astate->currTransMatrix.e[row][1],
            astate->currTransMatrix.e[row][2],
            astate->currTransMatrix.e[row][3]);
    }
    */
}

//----------------------------------------------------------------------------
//   class UnknownNode : Node
//----------------------------------------------------------------------------
//used to store nodes which are not known by this program
//alist stores the whole body of the node (evt. incl. subnodes)

//---UnknownNode::UnknownNode-------------------------------------------------
// Input : const char *aidstr  : string containing the type of the node
//         char *alist         : string containing the whole node body
//         int fline           : init
//         int lline           : init
// Output:

UnknownNode::UnknownNode(const char *aidstr, char *alist, int fline, int lline) :
             Node(NID_UNDEFINED,NULL,NULL,fline,lline)
{
    idstr = aidstr;
    field  = new Field(FID_UNDEFINED,new FString(alist));
}

//---UnknownNode::print-------------------------------------------------------
// Input : FILE *fd    : VRML output file descriptor
//         int indent  : indention of output
// Output:
// prints out the unknown node (not formated)

void UnknownNode::print(FILE *fd, int indent)
{
    printIndent(fd,indent);
    if (nodespec) fprintf(fd,"DEF %s ",nodespec);
    fprintf(fd,"%s %s\n",idstr,((FString *)field->value)->string);

    warning("unknown node type >%s< (this may cause problems)\n",idstr); 
}

//----------------------------------------------------------------------------
//   class PerspectiveCameraNode : Node
//----------------------------------------------------------------------------
//used for VRML-node PerspectiveCamera - only necessary for finding this
//type of camera via setupCamera

//---PrespectiveCameraNode----------------------------------------------------
// Input : int aid               : init
//         char *afieldspec      : init
//         Field *afield         : init
//         int fline             : init
//         int lline             : init
// Output:

PerspectiveCameraNode::PerspectiveCameraNode(int aid, char *afieldspec, 
                                             Field *afield, int fline, int lline) :
                       Node(aid,afieldspec,afield,fline,lline)
{
}

//---PerspectiveCameraNode::setupCamera---------------------------------------
// Input : Node **camera  : first found camera (or NULL)
// Output: Node **camera  : this camera if this was the first
// if no PerspectiveCamera was found so far camera is set to this 
// otherwise if this camera and the first have different height angles
// a warning will be printed

void PerspectiveCameraNode::setupCamera(Node **camera)
{
    if ((*camera)==NULL) *camera=this;
    else
    {
        Field *heightAngle = getField(FID_HEIGHTANGLE);
        Field *heightAngle2 = (*camera)->getField(FID_HEIGHTANGLE);
        float ha = 0.785398;
        float ha2 = ha;
        if (heightAngle)  ha =  ((FLong *)heightAngle->value)->number;
        if (heightAngle2) ha2 = ((FLong *)heightAngle2->value)->number;

        if (ha!=ha2) warning("multiple cameras with different heightAngles (using first)\n");
    }
}

//----------------------------------------------------------------------------
//   class OrthographicCameraNode : Node
//----------------------------------------------------------------------------
//used for VRML-node OrthographicCamera - the only difference to 
//class Node is that setupCamera simply produces a warning

//---OrthographicCameraNode::OrthographicCameraNode---------------------------
// Input : int aid               : init
//         char *afieldspec      : init
//         Field *afield         : init
//         int fline             : init
//         int lline             : init
// Output:

OrthographicCameraNode::OrthographicCameraNode(int aid, char *afieldspec, 
                                               Field *afield, int fline, int lline) :
                        Node(aid,afieldspec,afield,fline,lline)
{
}

//---OrthographicCameraNode::setupCamera--------------------------------------
// Input : Node **camera : not used
// Output:
//this method simply produces a warning if lods are generated because
//lods makes no sense with orthographic cameras

void OrthographicCameraNode::setupCamera(Node **camera)
{
    if (nolods==0) warning("lods makes no sense with this type of camera\n");
}

//----------------------------------------------------------------------------

