// ===========================================================================
//  Copyright 1996 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       line
//  TYPE:       c++ code
//  PROJECT:    lodestar
//  CONTENT:    implementation of class LineSet, class Line
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rs      Reinhard Sainitzer
//		hb      Herbert Buchegger
// ===========================================================================
//  HISTORY:
//
// ===========================================================================

#include "line.h"
#include "enum.h"
#include <assert.h>

//----------------------------------------------------------------------------
//   class Line : Shape 
//----------------------------------------------------------------------------

//---Line::Line---------------------------------------------------------------
// Input : Vertex v1 : init
//         Vertex v2 : init
// Output: none

Line::Line(Vertex v1, Vertex v2) : Shape()
{
    vertex1=v1;
    vertex2=v2;
}

//---Line::getRepresentOfVertex-----------------------------------------------
// Input : int index     : index of a vertex (1 or 2)
//         int targetlod : number of the lod of the representative
// Output: Vertex        : representative vertex
//returns the representant in the next lod (targetlod)
//of the vertex determined by index (1 or 2)

Vertex Line::getRepresentOfVertex(int index, int targetlod)
{
    Vertex old,rep,other;

    assert(index==1 || index==2);
    if (index==1) 
    { 
        old=vertex1;
        other=vertex2;
    } 
    else
    {
        old=vertex2;
        other=vertex1;
    }

    Coord *represent = old.coord->getRepresent(targetlod);
    Coord *selected = represent->getSelected(old.coord->lod);


    if (selected==old.coord)
    {
        //point has no representative or point is selected for the 
        //representative - this means this point is unchanged

        rep = old;
        rep.coord = represent;
    }
    else if (selected!=other.coord)
    {
        //the selected point is not a part of this line 

        rep = old;
        rep.coord = represent;

    }   //else nullvertex   

    return rep;
}

//---Line::isEquivTo----------------------------------------------------------
// Input : Line *l : compare this Line with l
// Output: int     : TRUE or FALSE
//returns true if the coordinates of this line are equal to those of l

int Line::isEquivTo(Line *l)
{
    if (vertex1.coord==l->vertex1.coord && 
        vertex2.coord==l->vertex2.coord) return 1;

    if (vertex1.coord==l->vertex2.coord && 
        vertex2.coord==l->vertex1.coord) return 1;

    return 0;   
}

//----------------------------------------------------------------------------
//   class LineSet
//----------------------------------------------------------------------------

//---LineSet::LineSet---------------------------------------------------------
// Input : int alod             : init
//         CoordSet *cs         : init
//         int amaterialBinding : init
//         int anormalBinding   : init
// Output: none

LineSet::LineSet(int alod, CoordSet *cs, int amaterialBinding,
                 int anormalBinding)
{
    lod = alod;
    coordSet = cs;
    materialBinding = amaterialBinding;
    normalBinding = anormalBinding;

    range = 0;
    count = 0;

    firstLine = NULL;
    lastLine = NULL;

    next = NULL;
}

//---LineSet::~LineSet--------------------------------------------------------
// Input : none
// Output: none

LineSet::~LineSet()
{
    Line *lh;
    while (firstLine)
    {
        lh=firstLine;
        firstLine=(Line *)firstLine->next; 
        delete lh;
    }
}

//---LineSet::insert----------------------------------------------------------
// Input : Line **l : p to p to the line to insert
// Output: Line **l : *l = NULL if l is deleted 
//insert l in the LineSet. l is inserted internally in a linear list.
//if LineSet::insertOk returns 0 then l is deleted and
//the pointer to l is set to NULL

void LineSet::insert(Line **l)
{
    if (insertOk(*l))
    {
        if (firstLine)
        {
            lastLine->next=*l;
            lastLine=*l;
        }
        else firstLine=lastLine=*l;

        (*l)->vertex1.coord->addShapeRef(*l);
        (*l)->vertex2.coord->addShapeRef(*l);
        count++;
    }
    else
    {
        delete *l;
        *l=NULL;
    }
}

//---LineSet::print-----------------------------------------------------------
// Input : FILE *fd   : VRML output file descriptor
// Output: int indent : indention of output
//writes the content of all defined fields to fd
//the fields are printed using printSet

void LineSet::print(FILE *fd, int indent)
{
    printIndent(fd,indent);
    fprintf(fd,"%s [ ",fieldIdString[FID_COORDINDEX]);
    printSet(fd,indent+13,FID_COORDINDEX);
    fprintf(fd," ]\n");

    if (count==0) return;

    if (firstLine->vertex1.material)
    {
        printIndent(fd,indent);
        fprintf(fd,"%s [ ",fieldIdString[FID_MATERIALINDEX]);
        printSet(fd,indent+16,FID_MATERIALINDEX);
        fprintf(fd," ]\n");
    }

    if (firstLine->vertex1.normal)
    {
        printIndent(fd,indent);
        fprintf(fd,"%s [ ",fieldIdString[FID_NORMALINDEX]);
        printSet(fd,indent+14,FID_NORMALINDEX);
        fprintf(fd," ]\n");
    }

    if (firstLine->vertex1.textureCoord)
    {
        printIndent(fd,indent);
        fprintf(fd,"%s [ ",fieldIdString[FID_TEXTURECOORDINDEX]);
        printSet(fd,indent+20,FID_TEXTURECOORDINDEX);
        fprintf(fd," ]\n");
    }
}

//---LineSet::generateNextLOD-------------------------------------------------
// Input : none
// Output: none
//creates a new LineSet (lod+1) which contains the next Level of
//Detail - variable next points to this new LineSet

void LineSet::generateNextLOD()
{
    int newlod = lod+1;

    LineSet *lsh = new LineSet(newlod,coordSet,materialBinding,normalBinding);
    next = lsh;

    coordSet->deleteShapeRefs();


    Line *newline;
    Line *line = firstLine;
    while (line)
    {
        newline = new Line(line->getRepresentOfVertex(1,newlod),
                           line->getRepresentOfVertex(2,newlod));
        lsh->insert(&newline);
        line=(Line *)line->next;
    }
}

//---LineSet::getCenter-------------------------------------------------------
// Input : none
// Output: float &cx, float &cy, float &cz : the calculated center
//get the center of all coordinates used in the LineSet - result in cx,cy,cz

void LineSet::getCenter(float &cx, float &cy, float &cz)
{
    Line *line = firstLine;
    if (line==NULL)
    {
        cx=cy=cz = 0;
        return;
    }

    Coord *coord = line->vertex1.coord;
    Bounds bounds = Bounds(coord->x,coord->y,coord->z,
                           coord->x,coord->y,coord->z);

    while(line)
    {
        coord = line->vertex1.coord;
        bounds.enlargeIfOutside(coord->x,coord->y,coord->z); 
        coord = line->vertex2.coord;
        bounds.enlargeIfOutside(coord->x,coord->y,coord->z); 
        line=(Line *)line->next;
    }

    cx = (bounds.xmax+bounds.xmin) /2;
    cy = (bounds.ymax+bounds.ymin) /2;
    cz = (bounds.zmax+bounds.zmin) /2;
}

//---LineSet::getSphere-------------------------------------------------------
// Input : none
// Output: float : the radius of the bounding sphere
//get the radius of the bounding sphere of all lines

float LineSet::getSphere()
{
    Line *line = firstLine;
    if (line==NULL) return 0;

    Coord *coord = line->vertex1.coord;
    Bounds bounds = Bounds(coord->x,coord->y,coord->z,
                           coord->x,coord->y,coord->z);

    while(line)
    {
        coord = line->vertex1.coord;
        bounds.enlargeIfOutside(coord->x,coord->y,coord->z);
        coord = line->vertex2.coord;
        bounds.enlargeIfOutside(coord->x,coord->y,coord->z);
        line=(Line *)line->next;
    }

    float dx = bounds.xmax - bounds.xmin;
    float dy = bounds.ymax - bounds.ymin;
    float dz = bounds.zmax - bounds.zmin;
    return sqrt(dx*dx + dy*dy + dz*dz);
}

//---LineSet::insertOk--------------------------------------------------------
// Input : Line *l : pointer to a line for checking
// Output: int     : TRUE or FALSE
//returns 0 if:
//   if l->vertex1.coord==NULL
//   if l->vertex2.coord==NULL
//   l is a point (coords are equal)
//   a line with same coords was already inserted
//else 1 is returned

int LineSet::insertOk(Line *l)
{
    if (l->vertex1.coord==NULL) return 0;
    if (l->vertex2.coord==NULL) return 0;
    if (l->vertex1.coord==l->vertex2.coord) return 0;

    ShapeRef *shapeRef = l->vertex1.coord->shapeRef; 
    while (shapeRef)
    {
        if (l->isEquivTo((Line *)shapeRef->shape)) return 0;
        shapeRef=shapeRef->next;
    }

    return 1;
}

//---LineSet::printSet--------------------------------------------------------
// Input : FILE *fd   : VRML output file descriptor
//         int indent : indention of output
//         int fid    : field identifier (see FID_xxx constants in enum.h)
// Output: none
//print out the field (specified by fid) of the LineSet, containing
//the data of the inserted lines, by calling their print-method. 
//fid - see enum.h

void LineSet::printSet(FILE *fd, int indent, int fid)
{
    int first=1;

    global_cpl=0;

    Line *line = firstLine;
    while (line)
    {
        if (first) first=0;
        else
        {
            fprintf(fd,",");
            if (global_cpl>=global_longcpl)
            {
                fprintf(fd,"\n");
                printIndent(fd,indent);
                global_cpl=0;
            }
        }

        switch (fid)
        {
            case FID_COORDINDEX:
                fprintf(fd,"%d,%d,-1",line->vertex1.coord->index,
                                      line->vertex2.coord->index);
                global_cpl+=2;
                break;

            case FID_MATERIALINDEX:
                if (materialBinding==VRML_PER_PART ||
                    materialBinding==VRML_PER_PART_INDEXED ||
                    materialBinding==VRML_PER_FACE ||
                    materialBinding==VRML_PER_FACE_INDEXED)
                {
                    fprintf(fd,"%d",line->vertex1.material->index);
                    global_cpl++;
                    break;
                }
                if (materialBinding==VRML_PER_VERTEX ||
                    materialBinding==VRML_PER_VERTEX_INDEXED)
                {
                    fprintf(fd,"%d,%d,-1",line->vertex1.material->index,
                                          line->vertex2.material->index);
                    global_cpl+=2;
                }
                break;

            case FID_NORMALINDEX:
                if (normalBinding==VRML_PER_PART ||
                    normalBinding==VRML_PER_PART_INDEXED ||
                    normalBinding==VRML_PER_FACE ||
                    normalBinding==VRML_PER_FACE_INDEXED)
                {
                    fprintf(fd,"%d",line->vertex1.normal->index);
                    global_cpl++;
                    break;
                }
                if (normalBinding==VRML_PER_VERTEX ||
                    normalBinding==VRML_PER_VERTEX_INDEXED)
                {
                    fprintf(fd,"%d,%d,-1",line->vertex1.normal->index,
                                          line->vertex2.normal->index);
                    global_cpl+=2;
                }
                break;

            case FID_TEXTURECOORDINDEX:
                fprintf(fd,"%d,%d,-1",line->vertex1.textureCoord->index,
                                      line->vertex2.textureCoord->index);
                global_cpl+=2;
                break;

            default: assert(0);
        }

        line=(Line *)line->next;
    }    

}

//----------------------------------------------------------------------------
