// ===========================================================================
//  Copyright 1996 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       line
//  TYPE:       c++ header
//  PROJECT:    lodestar
//  CONTENT:    class LineSet, class Line
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rs      Reinhard Sainitzer
//		hb      Herbert Buchegger
// ===========================================================================
//  HISTORY:
//
// ===========================================================================
//  MANUAL:
//
// The class LineSet is used to store lines defined in VRML-Nodes of type
// IndexedLineSet (the class don't supports polylines so it is necessary to
// split these polylines into normal lines). Lines are inserted using the
// method LineSet::insert(Line **l). The class removes double defined
// lines (with same coordiantes). The lines are inserted internally into a
// linear list.
// The LineSet is also used to produce Levels of Detail. The method 
// generateNextLOD creates a new LineSet and fills this LineSet with the
// lines of the next lod (lines of the current LineSet are mapped onto
// the representants of the coordinates in the next lod) doublettes and 
// collapsed lines (points)  will be discarded - the variable next points
// then to this new LineSet (next lod).
//
// SEE ALSO:
//     Shape, Vertex
//
// ===========================================================================

#ifndef _LINE_H_
#define _LINE_H_

#include "coord.h"
#include "shape.h"

//----------------------------------------------------------------------------

class Line : public Shape
{
public:

    Vertex vertex1;  //pointers to coords, materials, normals,texturecoords
    Vertex vertex2;  //pointers to coords, materials, normals,texturecoords

    Line(Vertex v1, Vertex v2);

    Vertex getRepresentOfVertex(int index,      //returns the representant
                                int targetlod); //in the next lod (targetlod)
                                                //of the vertex determined by
                                                //index (1 or 2)
    int isEquivTo(Line *l);      //returns true if the coordinates of this
                                 //line are equal to those of l
};

//----------------------------------------------------------------------------

class LineSet
{
public:

    int lod;             //the number of the lod representated by this lineset
    int count;           //count of all lines in the set
    int materialBinding; //(vrml) material binding used in this set 
                         //used in printSet for printing out the index
                         //references to Material-Nodes 
    int normalBinding;   //(vrml) normal binding used in this set
                         //used in printSet for printing out the index
                         //references to Normal-Nodes
    float range;         //the maximum distance to the camera for
                         //this lineset (lod)
    CoordSet *coordSet;  //pointer to the CoordSet where the coords used
                         //in this LineSet come from
    LineSet *next;       //pointer to the next LineSet (lod)
                         //do not modify this variable! the variable is set
                         //by generateNextLOD

    LineSet(int alod, CoordSet *cs, int amaterialBinding, 
            int anormalBinding);
    ~LineSet();

    void insert(Line **l);  //insert l in the LineSet. l is inserted
                            //internally in a linear list.
                            //if LineSet::insertOk returns 0 then l is 
                            //deleted and the pointer to l is set to NULL 

    void print(FILE *fd, int indent);  //writes the content of all defined
                                       //fields to fd the fields are
                                       //printed using printSet

    void generateNextLOD();   //creates a new LineSet (lod+1)
                              //which contains the next Level of
                              //Detail - variable next points to
                              //this new LineSet

    void getCenter(float &cx, float &cy, float &cz); //get the center of all
                                                     //coordinates used in the
                                                     //set - result in cx,cy,cz

    float getSphere();    //get the radius of the bounding sphere of all lines

protected:
    Line *firstLine; //lines are stored in a list - pointer to the first line
    Line *lastLine;  //lines are stored in a list - pointer to the last line

    int insertOk(Line *l);   //returns 0 if:
                             //   if l->vertex1.coord==NULL
                             //   if l->vertex2.coord==NULL
                             //   l is a point (coords are equal)
                             //   a line with same coords was already inserted
                             //else 1 is returned

    void printSet(FILE *fd, int indent,   //print out the field (specified by 
                  int fid);               //fid) of the LineSet containing
                                          //the data of the inserted lines by 
                                          //calling their print-method. 
                                          //fid - see enum.h
};

//----------------------------------------------------------------------------

#endif

