// ===========================================================================
//  Copyright 1996 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       main
//  TYPE:       c++ code
//  PROJECT:    lodestar
//  CONTENT:
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rs      Reinhard Sainitzer
//              hb      Herbert Buchegger
// ===========================================================================
//  HISTORY:
//
// ===========================================================================

#include "nodes.h"
#include "tri.h"
#include "enum.h"
#include "option.h"
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <dirent.h>
#include <limits.h>
#include <new.h>

//----------------------------------------------------------------------------

#define VERSION   "1.01"
#define URL       "http://www.cg.tuwien.ac.at/research/vr/lodestar/"
#define COPYRIGHT "(c) 1996 Technical University of Vienna  Department of Computer Graphics\n\
written by Reinhard Sainitzer and Herbert Buchegger"

//----------------------------------------------------------------------------

#ifndef PATH_MAX
#ifdef MAXPATHLEN
#define PATH_MAX MAXPATHLEN
#endif
#endif

//----------------------------------------------------------------------------

void outOfMemory()
{
    fprintf(stderr,"error: out of memory\n");
    exit(1);
}

//----------------------------------------------------------------------------

extern int yyparse(void);
extern Node *yacc_tree;
extern int yacc_nolods;
extern lex_setInput(FILE *f);

//---usage--------------------------------------------------------------------
// Input : char *app              : program name string
//         const char *errormsg   : error message string
// Output:

void usage(char *app, const char *errormsg)
{
    fprintf(stderr,"%s\n\n",errormsg);
    fprintf(stderr,"usage: %s [options] inputfile\n\n",app);
    fprintf(stderr,"   general-options:\n");
    fprintf(stderr,"      (-o | --output) outputfile          (-j | --join)\n");
    fprintf(stderr,"      (-n | --nolods)                     (-C | --copyright)\n");
    fprintf(stderr,"      (-v | --verbose) section['+'section]*\n");
    fprintf(stderr,"   lod-options:\n");
    fprintf(stderr,"      (-1 | --errorarea) weight           (-m | --mindiff) percentage\n");
    fprintf(stderr,"      (-2 | --errorvolume) weight         (-x | --extract) lod\n");
    fprintf(stderr,"      (-3 | --weightedmean) weight        (-l | --nolines)\n");
    fprintf(stderr,"      (-c | --cameraheightangle) radian   (-q | --noquads)\n");
    fprintf(stderr,"      (-e | --errorrange) percentage      (-s | --quadsplitangle) radian\n");
    fprintf(stderr,"   format-options:\n");
    fprintf(stderr,"      (-I | --indent) count               (-D | --decimalcountperline) count\n");
    fprintf(stderr,"      (-T | --tab) count                  (-F | --floatcountperline) count\n");
    fprintf(stderr,"      (-N | --noindent)\n");
    fprintf(stderr,"\n");
    exit(1);
}

//---getSection---------------------------------------------------------------
// Input : const char **optstr   : string with all verbose options
// Output: char *                : returns the first option string
// The function returns the characters before the first '+' or
// the hole string if there is no '+' - this characters will be deleted
// in optstr. So it is possible to use a extended
// option string: --verbose join+coord...

char *getSection(const char **optstr)
{
    char *h = strchr((char *)(*optstr),'+');
    int len;
    if (h) len = h-*optstr; else len = strlen(*optstr);

    char *optarg = strncpy(new char[len+1], *optstr, len);
    optarg[len]='\0';

    if (h) *optstr=h+1; else *optstr=NULL;
    return optarg;
}

//---main---------------------------------------------------------------------
// Input : int argc          : ...
//         char **argv       : ...
// Output:

int main(int argc, char **argv)
{
    printf("LODESTAR Level of Detail Generator for VRML 1.0 Release %s\n\n",VERSION);

    set_new_handler(&outOfMemory);

    char outfilename[PATH_MAX+1];
    int printtree=0;

    //---work on command line options & arguments-----------------------------


    OptionString osVerbose("verbose","v");
    OptionString osOutput("output","o");
    OptionFlag   ofNoLODs("nolods","n");
    OptionFlag   ofJoin("join","j");
    OptionFlag   ofCopyright("copyright","C");

    OptionDouble odWeightAlgErrorArea("errorarea","1",8);
    OptionDouble odWeightAlgErrorVolume("errorvolume","2",1);
    OptionDouble odWeightAlgWeightedMean("weightedmean","3",0);
    OptionDouble odCameraHeightAngle("cameraheightangle","c",0.785398);
    OptionDouble odErrorRange("errorrange","e",2.0);
    OptionDouble odMinDiff("mindiff","m",20.0);
    OptionLong   olExtractLOD("extract","x",-1);
    OptionFlag   ofNoLines("nolines","l");
    OptionFlag   ofNoQuads("noquads","q");
    OptionDouble odQuadSplitAngle("quadsplitangle","s",0.174533);

    OptionLong   olIndent("indent","I",4);
    OptionLong   olTab("tab","T",8);
    OptionFlag   ofNoIndent("noindent","N");
    OptionLong   olDecimalCountPerLine("decimalcountperline","D",8);
    OptionLong   olFloatCountPerLine("floatcountperline","F",6);

    if (!Option::parse(argc,argv)) usage(argv[0],"invalid option");

    if (ofCopyright.isSpecified())
    {
        printf("%s\n",COPYRIGHT);
        printf("URL: %s\n\n",URL);
        exit(0);
    }
 
    if (argc<2) usage(argv[0],"no input file specified");
    if (argc>2) usage(argv[0],"too much parameters");

    if (osVerbose.isSpecified())
    {
        char *optarg;
        const char *optstr = osVerbose.getValue();

        while (optstr)
        {
            optarg = getSection(&optstr);
           
            if (optarg[0]=='\0') usage(argv[0],"invalid verbose parameter");
            
            if (strcmp(optarg,"all") ==0)
            {
    	        Triangulator::verbose = 1;
	        CoordNode::verbose = 1;
	        MaterialNode::verbose = 1;
	        NormalNode::verbose = 1;
	        TextureCoordNode::verbose = 1;
	        FaceNode::verbose = 1;
            }
            else if (strcmp(optarg,"default")==0)
            {
                CoordNode::verbose=1;
                MaterialNode::verbose=1;
                NormalNode::verbose=1;
                TextureCoordNode::verbose=1;
                FaceNode::verbose=1;
                LineNode::verbose=1;
            }
            else if (strcmp(optarg,"triangulator") ==0) Triangulator::verbose = 1;
            else if (strcmp(optarg,"coord")   ==0) CoordNode::verbose = 1;
            else if (strcmp(optarg,"material")==0) MaterialNode::verbose = 1;
            else if (strcmp(optarg,"normal")  ==0) NormalNode::verbose = 1;
            else if (strcmp(optarg,"texcoord")==0) TextureCoordNode::verbose = 1;
            else if (strcmp(optarg,"face")    ==0) FaceNode::verbose = 1;
            else if (strcmp(optarg,"line")    ==0) LineNode::verbose = 1;
            else if (strcmp(optarg,"join")    ==0) Node::verbosejoin = 1;
            else if (strcmp(optarg,"tree")    ==0) printtree = 1;
            else usage(argv[0],"invalid verbose parameter");

            delete optarg;
        }
                
    }

    if (osOutput.isSpecified())
    {
        if (strlen(osOutput.getValue())>PATH_MAX) usage(argv[0],"outfile path too long");
        strcpy(outfilename,osOutput.getValue());
    }
    else
    {
        #ifdef OUTPUT_PARAM_REQUIRED
        usage(argv[0],"in this version the parameter -o or --output must be specified !!!");
        #endif

        char *all = strdup(argv[1]);
        char *name;
        char *ext;
        char ins[12];
        char *h;

        h = strrchr(all,'/');
        #ifdef BACKSLASH
        char *h2 = strrchr(all,'\\');
        if (h2>h) h=h2;
        #endif
        if (h) name=h+1; else name=all;

        h = strrchr(all,'.');
        if (h) {
            ext=h+1;
            *h='\0'; 
        } 
        else ext=name+strlen(name);

        strcpy(ins,"lod");
        if (ofNoLODs.isSpecified()) strcpy(ins,"opt");
        if (olExtractLOD.isSpecified()) sprintf(ins,"%d",(int)olExtractLOD.getValue());
        if (strlen(name)+strlen(ins)+strlen(ext)>PATH_MAX) 
            usage(argv[0],"outfile path too long");

        if (*ext=='\0') sprintf(outfilename,"%s.%s",name,ins);
                   else sprintf(outfilename,"%s.%s.%s",name,ins,ext);
        delete all;
    }

    if (ofNoLODs.isSpecified())
    {
        Node::nolods = 1;
        yacc_nolods = 1;
    }

    global_indent = olIndent.getValue();
    global_tab = olTab.getValue();
    global_noindent = ofNoIndent.isSpecified(); 
    global_longcpl = olDecimalCountPerLine.getValue();
    global_floatcpl = olFloatCountPerLine.getValue();

    if (ofNoLines.isSpecified()) FaceSet::nolines = 1;
    if (ofNoQuads.isSpecified()) FaceSet::noquads = 1;

    CoordSet::minDiff = odMinDiff.getValue() / 100;

    OctreeCoordSet::weightAlgErrorArea = odWeightAlgErrorArea.getValue();
    OctreeCoordSet::weightAlgErrorVolume = odWeightAlgErrorVolume.getValue();
    OctreeCoordSet::weightAlgWeightedMean = odWeightAlgWeightedMean.getValue();

    Face::splitCosinus = cos(odQuadSplitAngle.getValue());

    Node::extractlod = olExtractLOD.getValue();
    if (olExtractLOD.isSpecified() && olExtractLOD.getValue()<0) 
        usage(argv[0],"option --extract can't be less than zero");

    if (ofNoLODs.isSpecified() && olExtractLOD.isSpecified())
        usage(argv[0],"options --nolods and --extract can't be used simultaneous");

    if (odErrorRange.getValue()<=0) usage(argv[0],"illegal errorrange");

    //---open input & output file---------------------------------------------

    FILE *in;
    if ((in=fopen(argv[1],"r")) == NULL)
    {
        fprintf(stderr,"error: cannot open %s\n",argv[1]);
        perror("error");
        exit(1);
    }
    lex_setInput(in);

    FILE *fd;
    if ((fd=fopen(outfilename,"w")) == NULL)
    {
        fprintf(stderr,"error: cannot create %s\n",outfilename);
        perror("error");
        exit(1);
    }

    //---wake up the yacc :-) ------------------------------------------------

    printf("parsing ...\n");
    if (yyparse() != 0)
    {
        printf("failed.\n");
        return 1;
    }

    //---add header and Info-Node---------------------------------------------

    time_t currtime = time(NULL);
    char *infostring = new char[500];
    sprintf(infostring,"\
Created by LODESTAR - Level of Detail Generator Release %s\n\
\n\
authors: Reinhard Sainitzer & Herbert Buchegger\n\
contact: Technical University of Vienna Austria\n\
         URL: %s\n\
date:    %s",VERSION,URL,ctime(&currtime));

    Node *infonode = new Node(NID_INFO,NULL,
                         new Field(FID_STRING,
                             new FString(infostring)),0,0); 
    infonode->last=infonode;

    if (yacc_tree && yacc_tree->next==NULL &&
           (yacc_tree->id==NID_GROUP ||
            yacc_tree->id==NID_SEPARATOR ||
            yacc_tree->id==NID_TRANSFORMSEPARATOR))
    {
        if (((GroupNode *)yacc_tree)->sub)
        {
            infonode->next=((GroupNode *)yacc_tree)->sub;
            infonode->last=((GroupNode *)yacc_tree)->sub->last;
            ((GroupNode *)yacc_tree)->sub=infonode;
        }
        else ((GroupNode *)yacc_tree)->sub=infonode;
    }
    else
    {
        SeparatorNode *separatornode = 
            new SeparatorNode(NID_SEPARATOR,NULL,NULL,infonode,0,0);
        infonode->next=yacc_tree;
        infonode->last=yacc_tree->last;
        yacc_tree=separatornode;
    }

    RootNode *root = new RootNode("#VRML V1.0 ascii",yacc_tree);

    //---find out dependencies between nodes----------------------------------

    printf("dependencies ...\n");

    State *state = new State();

    state->lastCoordNode = new CoordNode(NID_COORDINATE3,NULL,NULL,-1,-1);
    state->lastMaterialNode = new MaterialNode(NID_MATERIAL,NULL,NULL,-1,-1);
    state->lastNormalNode = new NormalNode(NID_NORMAL,NULL,NULL,-1,-1);
    state->lastTextureCoordNode = new TextureCoordNode(NID_TEXTURECOORDINATE2,NULL,NULL,-1,-1);
    state->lastTextureNode = new TextureNode(NID_TEXTURE2,NULL,NULL,-1,-1);

    root->setupDepends(state);

    //---cameraheightangle----------------------------------------------------

    CoordSet::errorRange = odErrorRange.getValue();
    CoordSet::cameraHeightAngle = odCameraHeightAngle.getValue();

    if (!odCameraHeightAngle.isSpecified() && !ofNoLODs.isSpecified())
    {
        printf("searching for a camera ...\n");
        Node *camera = NULL;
        root->setupCamera(&camera);
        if (camera)
        {
            Field *heightAngle = camera->getField(FID_HEIGHTANGLE);
            if (heightAngle) 
                CoordSet::cameraHeightAngle=((FFloat *)heightAngle->value)->number;
        }
    }

    //---trying to join nodes-------------------------------------------------

    if (ofJoin.isSpecified())
    {
        printf("joining ...");
        fflush(stdout);
        root->join();
        printf(" %d Nodes removed.\n",Node::removedcount);
    }
    Node::verbosejoin = 0;

    //---print VRML tree------------------------------------------------------

    if (printtree)
    {
        root->printTree(0);
        printf("\n");
    }

    //---do all the other stuff-----------------------------------------------

    printf("writing ...\n");
    root->print(fd,0);

    //---clean up-------------------------------------------------------------

    fclose(fd);
    fclose(in);
    printf("success. (%s)\n",outfilename);
    return 0;
}


