// ===========================================================================
//  Copyright 1996 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       material
//  TYPE:       c++ code
//  PROJECT:    lodestar
//  CONTENT:    implementation of class MaterialSet, class Material, class RGB
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rs      Reinhard Sainitzer
//		hb      Herbert Buchegger
// ===========================================================================
//  HISTORY:
//
// ===========================================================================

#include "material.h"
#include "enum.h"
#include <assert.h>

//----------------------------------------------------------------------------
//   class RGB 
//----------------------------------------------------------------------------

//---RGB::RGB-----------------------------------------------------------------
// Input : none
// Output: none

RGB::RGB()
{
    r=0;
    g=0;
    b=0;
}

//---RGB::RGB-----------------------------------------------------------------
// Input : float ar,ag,ab : init
// Output: none

RGB::RGB(float ar, float ag, float ab)
{
    r=ar;
    g=ag;
    b=ab;
}

//---RGB::compareWith---------------------------------------------------------
// Input : RGB c  : compare this RGB-object with c
// Output: int    : -1, 0 or 1
//compare the RGB-values with RGB-object c
//used for inserting in a binary tree

int RGB::compareWith(RGB c)
{
    if (r < c.r) return -1;
    if (r > c.r) return 1;
    if (g < c.g) return -1;
    if (g > c.g) return 1;
    if (b < c.b) return -1;
    if (b > c.b) return 1;
    return 0;
}

//---RGB::setTo---------------------------------------------------------------
// Input : FVec3f *v  : pointer to an vector from class FVec3f
// Output: none
// set rgb to xyz of class FVec3f

void RGB::setTo(FVec3f *v)
{
    r=v->x;
    g=v->y;
    b=v->z;
}

//---RGB::deviationTo---------------------------------------------------------
// Input : RGB c : value for compare
// Output: float : deviation

float RGB::deviationTo(RGB c)
{
    float dr = r - c.r;
    float dg = g - c.g;
    float db = b - c.b;

    return (dr*dr + dg*dg + db*db) / 3;
}

//---RGB::meanValue-----------------------------------------------------------
// Input : RGB c       : calc the mean value of this RGB-ojbect and c
//         RGB *result : pointer to a RGB-object for the result
// Output: RGB *result : result is written in this RGB-object

void RGB::meanValue(RGB c, RGB *result)
{
    result->r = (r + c.r) /2;
    result->g = (g + c.g) /2;
    result->b = (b + c.b) /2;
}

//---RGB::weightedMeanValue---------------------------------------------------
// Input : RGB c       : calc the weighted mean value of this RGB-obj and c
//         float w1    : weight for this RGB-object
//         float w2    : weight for RGB-object c
//         RGB *result : pointer to a RGB-object for the result
// Output: RGB *result : result is written in this RGB-object

void RGB::weightedMeanValue(RGB c, float w1, float w2, RGB *result)
{
    float w = w1+w2;
    result->r = (w1*r + w2*c.r) /w;
    result->g = (w1*g + w2*c.g) /w;
    result->b = (w1*b + w2*c.b) /w;
}

//---RGB operator-------------------------------------------------------------

RGB operator +(RGB c1, RGB c2)
{
    return RGB(c1.r+c2.r, c1.g+c2.g, c1.b+c2.b);
}

//---RGB operator-------------------------------------------------------------

RGB operator -(RGB c1, RGB c2)
{
    return RGB(c1.r-c2.r, c1.g-c2.g, c1.b-c2.b);
}

//----------------------------------------------------------------------------
//   class Material : Data
//----------------------------------------------------------------------------

//---Material::Material-------------------------------------------------------
// Input : none
// Output: none

Material::Material() : Data()
{
    ambientColor=RGB();
    diffuseColor=RGB();
    specularColor=RGB();
    emissiveColor=RGB();
    shininess=0;
    transparency=0;

    left=NULL;
    right=NULL;
}

//---Materia::Material--------------------------------------------------------
// Input : RGB amb, diff,spec,emis     : init
//         float shin,trans            : init
//         int astatus                 : init
// Output: none

Material::Material(RGB amb, RGB diff, RGB spec, RGB emis,
                   float shin, float trans, int astatus) : Data()
{
    ambientColor=amb;
    diffuseColor=diff;
    specularColor=spec;
    emissiveColor=emis;
    shininess=shin;
    transparency=trans;

    status=astatus;
    left=NULL;
    right=NULL;
}

//---Material::print----------------------------------------------------------
// Input : FILE *fd : VRML output file descriptor
//         int fid  : field id (see FID_xxx constants in enum.h)
// Output:
//write the content of field fid (see FID_xxx constants for material in enum.h)
//to file fd - if field is a color write the rgb-values separated by blanks
//and increase global variable global_cpl at 3 else the float-values if
//written to file and global_cpl is increased at 1

void Material::print(FILE *fd, int fid)
{
    RGB c;
    float f=0;

    switch (fid)
    {
        case FID_AMBIENTCOLOR:  c=ambientColor;  break;
        case FID_DIFFUSECOLOR:  c=diffuseColor;  break;
        case FID_SPECULARCOLOR: c=specularColor; break;
        case FID_EMISSIVECOLOR: c=emissiveColor; break;
        case FID_SHININESS:     f=shininess;     break;
        case FID_TRANSPARENCY:  f=transparency;  break;
        default: assert(0);
    }
 
    if (fid!=FID_SHININESS && fid!=FID_TRANSPARENCY)
    {
        printFloat(fd,c.r); fprintf(fd," ");
        printFloat(fd,c.g); fprintf(fd," ");
        printFloat(fd,c.b);
        global_cpl+=3;
    }
    else 
    {
        printFloat(fd,f);
        global_cpl++;
    }
}

//---Material::compareWith----------------------------------------------------
// Input : Material *m : compare this Material with m
// Output: int         : -1,0 or 1 (0 if equal)
//compare all defined fields (see status) with the fields of Material m
//status must be the same in both objects !!
//function returns -1,0 or 1 
//used for inserting in a binary tree

int Material::compareWith(Material *m)
{
    int result;

    if (status & MAT_STAT_AMB)
    {    
        result = ambientColor.compareWith(m->ambientColor);
        if (result!=0) return result;
    }

    if (status & MAT_STAT_DIFF)
    {    
        result = diffuseColor.compareWith(m->diffuseColor);
        if (result!=0) return result;
    }

    if (status & MAT_STAT_SPEC)
    {    
        result = specularColor.compareWith(m->specularColor);
        if (result!=0) return result;
    }

    if (status & MAT_STAT_EMIS)
    {
        result = emissiveColor.compareWith(m->emissiveColor);
        if (result!=0) return result;
    }

    if (status & MAT_STAT_SHIN)
    {
        if (shininess < m->shininess) return -1;
        if (shininess > m->shininess) return 1;
    }

    if (status & MAT_STAT_TRANS)
    {
        if (transparency < m->transparency) return -1;
        if (transparency > m->transparency) return 1;
    }

    return 0;
}

//---Material::deviationTo----------------------------------------------------
// Input : Material *m        : calc the deviation of this Material to m
// Output: float              : deviation

float Material::deviationTo(Material *m) 
{
    int num = 0;
    float dev = 0;
    float d;

    if (status & MAT_STAT_AMB)
    {    
        dev += ambientColor.deviationTo(m->ambientColor);
        num++;
    }

    if (status & MAT_STAT_DIFF)
    {    
        dev += diffuseColor.deviationTo(m->diffuseColor);
        num++;
    }

    if (status & MAT_STAT_SPEC)
    {    
        dev += specularColor.deviationTo(m->specularColor);
        num++;
    }

    if (status & MAT_STAT_EMIS)
    {
        dev += emissiveColor.deviationTo(m->emissiveColor);
        num++;
    }

    if (status & MAT_STAT_SHIN)
    {
        d = shininess - m->shininess;
        dev += d*d;
        num++;
    }

    if (status & MAT_STAT_TRANS)
    {
        d = transparency - m->transparency;
        dev += d*d;
        num++;
    }

    if (num==0) return 0;
    return dev/num;
}

//---Material::meanValue------------------------------------------------------
// Input : Material *m      : calc the mean value of this Material and m
//         Material *result : a pointer to a Material for the result
// Output: Material *result : the result is written in Material result

void Material::meanValue(Material *m, Material *result) 
{
    ambientColor.meanValue(m->ambientColor,&result->ambientColor);
    diffuseColor.meanValue(m->diffuseColor,&result->diffuseColor);
    specularColor.meanValue(m->specularColor,&result->specularColor);
    emissiveColor.meanValue(m->emissiveColor,&result->emissiveColor);

    result->shininess = (shininess + m->shininess) /2;
    result->transparency = (transparency + m->transparency) /2;
    result->status = status & m->status;
}

//---Material::weightedMeanValue----------------------------------------------
// Input : Material *m      : calc the mean weighted mean value of this and m
//         float w1         : weight for this Material
//         float w2         : weight for Material m
//         Material *result : a pointer to a Material for the result
// Output: Material *result : the result is written in Material result

void Material::weightedMeanValue(Material *m,float w1,float w2,Material *result) 
{
    ambientColor.weightedMeanValue(m->ambientColor,w1,w2,&result->ambientColor);
    diffuseColor.weightedMeanValue(m->diffuseColor,w1,w2,&result->diffuseColor);
    specularColor.weightedMeanValue(m->specularColor,w1,w2,&result->specularColor);
    emissiveColor.weightedMeanValue(m->emissiveColor,w1,w2,&result->emissiveColor);

    float w = w1+w2;
    result->shininess = (w1*shininess + w2*m->shininess) /w;
    result->transparency = (w1*transparency + w2*m->transparency) /w;
    result->status = status & m->status;
}

//---Material operator--------------------------------------------------------

Material operator +(Material m1, Material m2)
{
    return Material(m1.ambientColor  + m2.ambientColor,
                    m1.diffuseColor  + m2.diffuseColor,
                    m1.specularColor + m2.specularColor,
                    m1.emissiveColor + m2.emissiveColor,
                    m1.shininess     + m2.shininess,
                    m1.transparency  + m2.transparency,
                    m1.status & m2.status);
}

//---Material operator--------------------------------------------------------

Material operator -(Material m1, Material m2)
{
    return Material(m1.ambientColor  - m2.ambientColor,
                    m1.diffuseColor  - m2.diffuseColor,
                    m1.specularColor - m2.specularColor,
                    m1.emissiveColor - m2.emissiveColor,
                    m1.shininess     - m2.shininess,
                    m1.transparency  - m2.transparency,
                    m1.status & m2.status);
}

//----------------------------------------------------------------------------
//   class MaterialSet : DataSet
//----------------------------------------------------------------------------

//---MaterialSet::MaterialSet-------------------------------------------------
// Input : none
// Output: none

MaterialSet::MaterialSet() : DataSet()
{
    materialRoot=NULL;
}

//---MaterialSet::~MaterialSet------------------------------------------------
// Input : none
// Output: none

MaterialSet::~MaterialSet()
{
    deleteTree(materialRoot);
}

//---MaterialSet::insert------------------------------------------------------
// Input : Material **m : p to p to the material to insert
// Output: Material **m : *m=NULL if material is deleted (equal material exists)
//first m is inserted in a binary tree using insertTree
//if an equal Material already exists m is deleted and the pointer to m
//is redirected to the existing Material
//then a reference to m is appended to DataSet::refArray using insertArray

void MaterialSet::insert(Material **m)
{
    insertTree(m,&materialRoot);
    insertArray(*m);
}

//---MaterialSet::print-------------------------------------------------------
// Input : FILE *fd         : VRML output file descriptor
//         int indent       : indention of output
// Output:
//writes the content of all valid(defined) fields (see status) to fd
//The fields are printed using calls to MaterialSet::printSet
//additionally DataSet::setIndices is called

void MaterialSet::print(FILE *fd, int indent)
{
    setIndices();

    if (materialRoot==NULL) return;

    if (materialRoot->status & MAT_STAT_AMB)
    {
        printIndent(fd,indent);
        fprintf(fd,"%s [ ",fieldIdString[FID_AMBIENTCOLOR]);
        printSet(fd,indent+15,FID_AMBIENTCOLOR);
        fprintf(fd," ]\n");
    }

    if (materialRoot->status & MAT_STAT_DIFF)
    {
        printIndent(fd,indent);
        fprintf(fd,"%s [ ",fieldIdString[FID_DIFFUSECOLOR]);
        printSet(fd,indent+15,FID_DIFFUSECOLOR);
        fprintf(fd," ]\n");
    }

    if (materialRoot->status & MAT_STAT_SPEC)
    {
        printIndent(fd,indent);
        fprintf(fd,"%s [ ",fieldIdString[FID_SPECULARCOLOR]);
        printSet(fd,indent+16,FID_SPECULARCOLOR);
        fprintf(fd," ]\n");
    }

    if (materialRoot->status & MAT_STAT_EMIS)
    {
        printIndent(fd,indent);
        fprintf(fd,"%s [ ",fieldIdString[FID_EMISSIVECOLOR]);
        printSet(fd,indent+16,FID_EMISSIVECOLOR);
        fprintf(fd," ]\n");
    }

    if (materialRoot->status & MAT_STAT_SHIN)
    {
        printIndent(fd,indent);
        fprintf(fd,"%s [ ",fieldIdString[FID_SHININESS]);
        printSet(fd,indent+12,FID_SHININESS);
        fprintf(fd," ]\n");
    }

    if (materialRoot->status & MAT_STAT_TRANS)
    {
        printIndent(fd,indent);
        fprintf(fd,"%s [ ",fieldIdString[FID_TRANSPARENCY]);
        printSet(fd,indent+15,FID_TRANSPARENCY);
        fprintf(fd," ]\n");
    }
}

//---MaterialSet::getIndex----------------------------------------------------
// Input : int index  : array index
// Output: Data *     : Data object with index index
//get Material by index (insertion order !!)
//same as DataSet::getIndex BUT index is taken modulo (count of inserted
//Materials with doublets)

Data *MaterialSet::getIndex(int index)
{
    return (Data *)((*refArray)[index%refArray->getSize()].ref);
}

//---MaterialSet::insertTree--------------------------------------------------
// Input : Material **m    : p to p to the material to insert
//         Material **root : p to p to the root of tree
// Output: Material **m    : *m=NULL if material is deleted (equal mat exists)
//         Material **root : *root is modified eventually
//insert m in the binary tree specified by root. If an equal Material
//already exists m is deleted and the pointer to m is redirected to
//the existing Material

void MaterialSet::insertTree(Material **m, Material **root)
{
    if ((*root)==NULL) 
    {
        //found a place to insert

        *root = *m;
        return;
    }

    int result = (*m)->compareWith(*root);
    if (result<0) insertTree(m,&((*root)->left));
    else if (result>0) insertTree(m,&((*root)->right));

    else
    {
        //material with same specifications already exists

        delete *m;
        *m = *root;
    }
}

//---MaterialSet::deleteTree--------------------------------------------------
// Input : Material * root : root of the tree to delete
// Output: none
//delete the binary tree specified by root

void MaterialSet::deleteTree(Material *root)
{
    if (root == NULL) return;
    deleteTree(root->left);
    deleteTree(root->right);
    delete root;
}

//---MaterialSet::printSet----------------------------------------------------
// Input : FILE * fd             : VRML output file descriptor
//         int indent            : indention of output
//         int fid               : field id (see FID_xxx constants in enum.h)
// Output:
//same as DataSet::printSet
//print out the data of field fid - but doublettes at the end of the list
//are not printed. e.g if the list is abaaccdddd this method prints abaaccd
//but if countUncompressed=8 it prints abaaccdd

void MaterialSet::printSet(FILE *fd, int indent, int fid)
{
    int i;
    int co = refArray->getSize();
    Material *mat;
    int first=1;
    global_cpl=0;

    if (countUncompressed>getMaxIndex()+1) countUncompressed=getMaxIndex()+1;

    //---remove doublettes at the end of the list

    int equal=1;
    while(co>1 && co>countUncompressed && equal)
    {
        Material *mat1=(Material *)getIndex(co-1);
        Material *mat2=(Material *)getIndex(co-2);
        int r=0;
        switch (fid)
        {
            case FID_AMBIENTCOLOR:
                r=mat1->ambientColor.compareWith(mat2->ambientColor);
                break;
            case FID_DIFFUSECOLOR:
                r=mat1->diffuseColor.compareWith(mat2->diffuseColor);
                break;
            case FID_SPECULARCOLOR:
                r=mat1->specularColor.compareWith(mat2->specularColor);
                break;
            case FID_EMISSIVECOLOR:
                r=mat1->emissiveColor.compareWith(mat2->emissiveColor);
                break;
            case FID_SHININESS:
                if (mat1->shininess==mat2->shininess) r=0; else r=1;
                break;
            case FID_TRANSPARENCY:
                if (mat1->transparency==mat2->transparency) r=0; else r=1;
                break;
            default: assert(0);
        }
        if (r!=0) equal=0; else co--;
    }

    //---reset done-flags

    for (i=0; i<co; i++) 
    {
        mat=(Material *)getIndex(i);
        mat->status &= ~DATASET_STAT_DONE;
    }

    //---print the list
 
    for (i=0; i<co; i++)
    {
        mat=(Material *)getIndex(i); 

        if ((mat->status & DATASET_STAT_DONE)==0 || (i<countUncompressed))
        {
            if (first) first=0;
            else
            {
                if (global_cpl>=global_floatcpl)
                {
                    fprintf(fd,",\n");
                    printIndent(fd,indent);
                    global_cpl=0;
                }
                else fprintf(fd,",");
            }

            mat->print(fd,fid);
            mat->status |= DATASET_STAT_DONE;
        }
    }
}

//----------------------------------------------------------------------------

