// ===========================================================================
//  Copyright 1996 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       material
//  TYPE:       c++ header
//  PROJECT:    lodestar
//  CONTENT:    class MaterialSet, class Material, class RGB
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rs      Reinhard Sainitzer
//		hb      Herbert Buchegger
// ===========================================================================
//  HISTORY:
//
// ===========================================================================
//  MANUAL:
//
//  The class MaterialSet is used to store materials defined in VRML-Nodes
//  from type Material. It removes double defined Materials (see also DataSet).
//  MaterialSet is derived from DataSet and Material is derived from Data.
//  Materials which will be inserted in this class must be from type Material
//  The materials are inserted using the method MaterialSet::insert(Material **m)
//  The Material are inserted internally in a binary tree. Additionally
//  References to the Materials are appended to the ReferenceArray (see also
//  DataSet) to get the Materials by index in insertion order 
//
//  In VRML a Material contains max six fields - four colors ambientColor,
//  diffuseColor, specularColor and emissiveColor and the fields shininess and
//  transparency. While shininess and transparency are simple float-values in the 
//  class Material the four colors are stored in subobjects from type RGB.
//  Because not all fields must be defined it is necessary to define which variables
//  contains valid data. This will be done by setting the corresponding bit
//  in DataSet::status (see also MAT_STAT_xxxxx constants)
//  All Material which will be inserted in a MaterialSet must have the same status! 
//
//  The Fields in the VRML-Node Material are Multiple-valued fields which can
//  have different lengths. If a field have less data than the other fields
//  you should fill this field with the last defined value. The method
//  MaterialSet::printSet ignores this filled values (it doesn't print
//  doublettes at the end of the list)
//
//  SEE ALSO
//      DataSet
//
// ===========================================================================


#ifndef _MATERIAL_H_
#define _MATERIAL_H_

#include "dataset.h"
#include "global.h"
#include "fields.h"
#include <stdio.h>

//----------------------------------------------------------------------------
//   constants for DataSet::status
//----------------------------------------------------------------------------

#define MAT_STAT_AMB     0x01
#define MAT_STAT_DIFF    0x02
#define MAT_STAT_SPEC    0x04
#define MAT_STAT_EMIS    0x08
#define MAT_STAT_SHIN    0x10
#define MAT_STAT_TRANS   0x20

//----------------------------------------------------------------------------
//used for storing and comparing colors in a Material

class RGB
{
public:

    float r;
    float g;
    float b;

    RGB();
    RGB(float ar, float ag, float ab);

    int compareWith(RGB c);             //compare this with c - returns
                                        //-1, 0 or 1 - used for inserting in
                                        //a binary tree
    void setTo(FVec3f *v);              //set r,g,b  to v->x, v->y, v->z

    //these methods are not used currently
    float deviationTo(RGB c);
    void meanValue(RGB c, RGB *result);
    void weightedMeanValue(RGB c, float w1, float w2, RGB *result);
    friend RGB operator +(RGB c1, RGB c2);
    friend RGB operator -(RGB c1, RGB c2);
};

//----------------------------------------------------------------------------

class Material : public Data
{
public:

    RGB ambientColor;
    RGB diffuseColor;
    RGB specularColor;
    RGB emissiveColor;
    float shininess;
    float transparency;

    Material *left;     //pointer to left successor for binary tree
    Material *right;    //pointer to right successor for binary tree

    Material();
    Material(RGB amb, RGB diff, RGB spec, RGB emis,
             float shin, float trans, int astatus);

    virtual void print(FILE *fd, int fid);  //writes the data spcified by fid
                                            //to file fd - fid see enum.h
    int compareWith(Material *m);           //compare this with m - returns
                                            //-1, 0 or 1 - used for inserting
                                            //in a binary tree

    //these methods are not used currently
    float deviationTo(Material *m);
    void meanValue(Material *m, Material *result);
    void weightedMeanValue(Material *m, float w1, float w2, Material *result);
    friend Material operator +(Material m1, Material m2);
    friend Material operator -(Material m1, Material m2);
};

//----------------------------------------------------------------------------

class MaterialSet : public DataSet
{
public:

    MaterialSet();
    ~MaterialSet();

    void insert(Material **m);  //insert m in the MaterialSet. m is inserted
                                //internally in a binary tree using insertTree.
                                //if an equal Material already exists m is
                                //deleted and the pointer to m is redirected
                                //to the existing Material
                                //additionally a reference to m is appended to
                                //DataSet::refArray via DataSet::insertArray
                                      
    void print(FILE *fd, int indent);  //writes the content of all valid fields
                                       //(see status) to fd. the fields are
                                       //printed using calls to MaterialSet::
                                       //printSet additionally
                                       //DataSet::setIndices is called

    Data *getIndex(int index);  //get Material by index (insertion order !!)
                                //same as DataSet::getIndex BUT index is taken
                                //modulo (count of inserted Materials
                                //with doublets)
     
protected:
    Material *materialRoot;     //root of the binary tree

    void insertTree(Material **m,      //insert m in the binary tree
                    Material **root);  //specified by root. if an equal
                                       //Material already exists m is deleted
                                       //and the pointer to m is redirected
                                       //to the existing Material
                                                     
    void deleteTree(Material *root);   //delete the binary tree
                                       //specified by root

    void printSet(FILE *fd, int indent, //same as DataSet::printSet
                  int fid);             //print out the data of field fid
                                        //but the function doesn't prints
                                        //doublettes at the end of the list
                                        //fid - see enum.h
};

//----------------------------------------------------------------------------

#endif

