// ===========================================================================
//  Copyright 1996 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       nodes
//  TYPE:       c++ header
//  PROJECT:    lodestar
//  CONTENT:    class Node and all classes derived from Node,
//              class Ident, class IdentSet, class State
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rs      Reinhard Sainitzer
//		hb      Herbert Buchegger
// ===========================================================================
//  HISTORY:
//
// ===========================================================================
//  MANUAL:
//
//  The class Ident is used to create a tree with all identifers (= nodespec).
//  To manage this tree you can use the class IdentSet. There are functions
//  to insert, search and delete Ident objects in the tree. Every Ident object
//  points to a Nodes object and this one points to the nodespec string.
//  If there are more identical identifers search returns the last inserted
//  identifer. Nodes can be inserted in this IdentSet by using registerName()
//
//  The class State is used to store the current state while traversing the 
//  VRML node tree with method setupDepends 
//  with the informations in this object e.g. the FaceSet is able to find
//  the right Coordinate3,Normal,... node and the actual bindings .
//
//  The class Node is used to store and handle VRML-nodes - this class is able
//  to contain any type of VRML-nodes ecxept those derived from Group but
//  this class can only store and print VRML-nodes therefore it is used
//  directly only for VRML-nodes which are uninteresting for lod-generation
//  (e.g. AsciiText, DirectionalLight, PointLight, WWWInline, ...)
//  many specialized classes are derived from this class
//
// SEE ALSO:
//     Field, CoordSet, MaterialSet, NormalSet, TextureCoordSet,
//     FaceSet, LineSet
//
// ===========================================================================

#ifndef _NODES_H_
#define _NODES_H_

#include "fields.h"
#include "coord.h"
#include "material.h"
#include "normal.h"
#include "texcoord.h"
#include "face.h"
#include "line.h"
#include <stdarg.h>

//---forward definitions------------------------------------------------------

class IdentSet;
class Node;
class CoordNode;
class MaterialNode;
class NormalNode;
class TextureCoordNode;
class TextureNode;

//----------------------------------------------------------------------------
// This is a simple class for storing all identifers in a tree.
// The left side contains the object with the smaller and the right side the 
// object with the greater string.

class Ident 
{
public:

    Node *node;                // Node::nodespec is the identifer
    Ident *left;               // the smaller string
    Ident *right;              // the greater string

    Ident(Node *anode);
};

//----------------------------------------------------------------------------
// With this class you can manage an Ident object tree (insert,search,delete).
// If there are more identical identifers search returns the last inserted
// identifier.

class IdentSet
{
public:

    IdentSet();
    ~IdentSet();
    void insert(Ident *ident);       // insert new Ident object on the place
    Ident *search(const char *name); // find the last identifer with the name

protected:

    Ident *root;

    void insertTree(Ident *ident, Ident **root);       // used from insert
    Ident *searchTree(const char *name, Ident *root);  // used from search
    void deleteTree(Ident *root);
};

//----------------------------------------------------------------------------
//This class is used to store the current state while traversing the 
//VRML node tree with method setupDepends 
//with the informations in this object e.g. the FaceSet is able to find
//the right Coordinate3,Normal,... node and the actual bindings .

class State
{
public:

    CoordNode *lastCoordNode;
    MaterialNode *lastMaterialNode;
    NormalNode *lastNormalNode;
    TextureCoordNode *lastTextureCoordNode;
    TextureNode *lastTextureNode;

    int currMaterialBinding;
    int currNormalBinding;
    int currVertexOrdering;
    int currFaceType;

    Matrix currTransMatrix;

    State();

    float getMaxScale();  //get the maximum of the current scale in 3D
                          //it calculates the lengthes of the 4 diagonals of
                          //an unit cube (1,1,1) transformed by the current
                          //transformation (currTransMatrix). The max length
                          //divided by the length of an unit cube (sqrt(3))
                          //is returned
};

//----------------------------------------------------------------------------
//This class is used to store and handle VRML-nodes - this class is able
//to contain any type of VRML-nodes ecxept those derived from Group but
//this class can only store and print VRML-nodes therefore it is used
//directly only for VRML-nodes which are uninteresting for lod-generation
//(e.g. AsciiText, DirectionalLight, PointLight, WWWInline, ...)
//many specialized classes are derived from this class

#define JF_REMOVEABLE          1  //join flag for variable joinflag
                                  //JF_REMOVEABLE will be set from a node which
                                  //will be joined with a second one. So the
                                  //separator of this node knows if he can
                                  //delete them or not.

#define JF_NOT_REMOVEABLE      2  //this node is not removeable
                                  //(example root->sub)

#define JF_RUNNING             4  //prevents that a GroupNode will be joined 2 times

//join two Indices (material,coord,texture,normal) which accesses the same set
#define JF_PREPARED_MATERIAL   8
#define JF_PREPARED_COORD     16
#define JF_PREPARED_TEXTURE   32  // same set and same bitmap!
#define JF_PREPARED_NORMAL    64 

//join mode for variable static int Node::joinmode
//join two Indices (material,coord,texture,normal) which accesses the same set
#define JM_NOTHING         0
#define JM_MATERIAL        1
#define JM_COORD           2
#define JM_NORMAL          4
#define JM_TEXTURE         8      // same set and same bitmap!

class Node
{
public:

    int id;                       //node-identifier see NID_xxx in enum.h
    char *fieldspec;              //this string contains optionally the 
                                  //fields specification (used in VRML for
                                  //extensibility) 
    Field *field;                 //pointer to a list of fields
    char *nodespec;               //the name of this node (if the node has one)
                                  //this is the identifier after DEF in VRML
    int used;                     //this node is reused by a USE-statement if ==1
    Node *next;                   //pointer to the next node
    Node *last;                   //pointer to the last node in the list
                                  //(needed by the yacc-parser)
    int firstline;                //node begins at line firstline (inputfile)
    int lastline;                 //node ends at line lastline (inputfile)
    int joinflag;                 //see JF_... flags
    static int joinmode;          //see JM_... flags
    static int nolods;            //no lods will be generated and printed
                                  //used by some subclasses (FaceNode,LineNode)
    static int extractlod;        //used for extracting a specified lod
                                  //lod are only generated til this value
                                  //and only the lods with this number will be
                                  //printed. -1 disables the mechanism
    static int removedcount;      //count of removed nodes by join
    static IdentSet *identifiers; //pointer to a global IdentSet-object
                                  //if nodespec!=NULL the method registerName()
                                  //creates an Indent-object and inserts it in the
                                  //global IdentSet-object
    static int verbosejoin;       //print out information while doing the
                                  //join algorithm if !=0

    Node(int aid, char *afieldspec, Field *afield, int fline, int lline);  
         
    virtual ~Node();

    void registerName();            //if nodespec!=NULL this method creates an
                                    //Indent-object and inserts it in the global
                                    //IdentSet-object (Node::identifiers)

    virtual void print(FILE *fd,    //print out the node inclusive
                       int indent); //all fields to fd
                                      
    void printHead(FILE *fd,        //print out the head of the node this
                   int &indent);    //includes { and evt. DEF. indent will   
                                    //be incremented by global_indent

    void printTail(FILE *fd,        //indent will be decremented by
                   int &indent);    //global_indent. Prints out the tail
                                    //(is '}') incl. indention to fd

    void printFields(FILE *fd,      //print out all fields stored in this node
                     int indent);   //incl. indention to fd

    void deleteFields();            //delete all fields of this node

    Field *getField(int aid);       //get a pointer to a field object stored
                                    //in this node specified by aid
                                    //(see FID_xxx constants in enum.h)
                                    //always use this function ! -
                                    //fields may be defined more once
                                    //this function returns the last

    void deleteValues(Value **list, //delete fist count elements of a list
                      int count);   //of objects derived from class Value
                                    //and skip the deleted values
                      
    void deleteAllValues(Value **list); //delete all elements of a list of
                                        //objects derived from class Value
                                        //*list=NULL after calling

    virtual void setupDepends(State *astate); //abstract - used for
                                              //determining the dependencies
                                              //between the nodes

    virtual void setupCamera(Node **camera); //abstract - used for searching
                                             //for the first PerspectiveCamera

    virtual void markAsUsed();  //mark this node as used - the method sets
                                //variable Node::used to 1 - this means that
                                //this node is used by a USE-statement
                                //virtual - see also GroupNode::markAsUsed !

    void message(const char *section,   //prints out a (verbose) message to
                 const char *msg, ...); //stdout - section indicates the
                                        //verbose section specified as 
                                        //parameter of the program

    void warning(const char *msg, ...); //prints out a warning message to stderr

    void error(const char *msg, ...);  //prints out an error message to stderr
                                       //and stops the program with exit code 1

    virtual void join();               //join manage function
                                       //checks if this node is joinable
                                       //with the next nodes

    virtual int joinable(Node * node); //virtual function for comparing 2 nodes 
                                       // used for the most nodes (UseNode,
                                       //UnknownNode,... are not joinable)
                                       //this function returns always FALSE

    virtual void prepareForJoin();        // abstract function
    virtual void makeJoin(Node * node);   // abstract function

    int compareField(Field * f1, Field *f2);   //compare a list of FFloat objects
    int compare3DField(Field * f1, Field *f2); //compare a list of FVec3f objects

    void joinFields(Field * f1,Field * f2);  //joinFields appends the list
                                             //of the second field to the list
                                             //of the first field and sets this
                                             //new list as list of the second
                                             //list

    virtual void makeJoinField(Node * node,  //makeJoinField appends the list
                               int fid);     //of the second field to the list
                                             //of the first field and uses this
                                             //new list as list of the second
                                             //field. If the field id is
                                             //FID_SHININESS a default value will
                                             //be used if there is no value present.


    virtual void makeJoinField3D(Node * node, //makeJoinField3D appends the list
                                 int fid);    //of the second field to the list
                                              //of the first field and uses this
                                              //new list as list of the second
                                              //field. If the field id is
                                              //FID_AMBIENTCOLOR or FID_DIFFUSECOLOR
                                              //a default value will be used
                                              //if there is no value present.

    virtual void printTree(int indent);       //print out a schema of the
                                              //node-tree to stdout
};

//----------------------------------------------------------------------------
//used for VRML-node USE

class UseNode : public Node
{
public:

    const char *keyword;   //the name of the node to use

    UseNode(const char *akeyword, int fline, int lline);

    virtual void print(FILE *fd,int indent);  //print out the USE-node to fd

    virtual void setupDepends(State *astate); //checks if a DEF exists for this
                                              //USE calls setupDepends of the
                                              //DEF-node (tree) 
                                              //marks DEF-node (tree) as used
};
    
//----------------------------------------------------------------------------
//used for VRML-node Group and is superclass for Separator,TransformSeparator
//is able to contain subnodes

class GroupNode : public Node
{
public:

    Node *sub;      //pointer to a list of subnodes

    GroupNode(int aid, char *afieldspec, Field *afield,
              Node *asub, int fline, int lline);

    virtual ~GroupNode();

    virtual void print(FILE *fd,         //prints out the GroupNode and all
                       int indent);      //subnodes - uses printSubNodes
                       
    void printSubNodes(FILE *fd,              //prints out all subnodes of
                       int indent);           //this GroupNode

    virtual void setupDepends(State *astate); //calls setupDepends for all subnodes

    virtual void setupCamera(Node **camera);  //used for searching for the first
                                              //PerspectiveCamera searches in all
                                              // subnodes of this GroupNode

    virtual void markAsUsed();     //marks this node and all subnodes as used
                                   //(variable used=1) - this means this node-
                                   //tree is used by a USE-node

    virtual void join();    //calls join functions for subnodes
                            //removes nodes with joinflag==JF_REMOVEABLE
                            //this flag is set in makeJoin() (in the first node )

    virtual int joinable(Node * node);  //checks if the sub lists are joinable

    
    virtual void prepareForJoin();          //prepare sub nodes for joining
    virtual void prepareForJoin(int flag,   //!!! 
                                int mode);

    virtual void makeJoin(Node * node);     //call makeJoin for the sub list

    virtual void printTree(int indent);       //print out a schema of the
                                              //node-tree to stdout
};

//----------------------------------------------------------------------------
//has no VRML dendant - used to store the VRML-header and is root of all
//VRML-nodes in a file

class RootNode : public GroupNode
{
public:

    const char *identification;              //VRML-header

    RootNode(const char *aidentification, Node *asub);

    virtual void print(FILE *fd,int indent); // prints out a VRML-header and
                                             // the node tree
    virtual void join();    //calls join functions for subnodes
                            //set first sub node as not removeable
};

//----------------------------------------------------------------------------
//used for VRML-node Separator

class SeparatorNode : public GroupNode
{
public:

    SeparatorNode(int aid,char *afieldspec,Field *afield,
                  Node *asub, int fline, int lline);

    virtual void setupDepends(State *astate);  //restores State after 
                                               //traversing the subnodes
};

//----------------------------------------------------------------------------
//used for VRML-node TransformSeparator

class TransformSeparatorNode : public GroupNode
{
public:

    TransformSeparatorNode(int aid, char *afieldspec, Field *afield,
                           Node *asub, int fline, int lline);

    virtual void setupDepends(State *astate); //restores transformation matrix
                                              //(State::currTransMatrix) after
                                              //traversing the subnodes
};

//----------------------------------------------------------------------------
//used for VRML-node Rotation

class RotationNode : public Node
{
public:

    RotationNode(int aid, char *afieldspec, Field *afield,
                 int fline, int lline);

    virtual void setupDepends(     //The Field with field-id FID_ROTATION
                   State *astate); //defines a 3D rotation about an arbitrary
                                   //axis through the origin. The rotation is
                                   //accumulated into the current transformation
                                   //(astate->currTransMatrix)
};

//----------------------------------------------------------------------------
//used for VRML-node Scale

class ScaleNode : public Node
{
public:

    ScaleNode(int aid, char *afieldspec, Field *afield,
              int fline, int lline);

    virtual void setupDepends(     //The Field with field-id FID_SCALEFACTOR
                   State *astate); //defines a 3D scaling about the origin.
                                   //The scale-operation is accumulated into
                                   //the current transformation
                                   //(astate->currTransMatrix)
};

//----------------------------------------------------------------------------
//used for VRML-node Transform

class TransformNode : public Node
{
public:

    TransformNode(int aid, char *afieldspec, Field *afield,
                  int fline, int lline);

    virtual void setupDepends(     //The fields of this node defines a geometric
                   State *astate); //3D transformation - the transformation is
                                   //accumulated into the current transformation
                                   //(astate->currTransMatrix)
                                   //see tnodes.C for a detailed description
};

//----------------------------------------------------------------------------
//used for VRML-node MatrixTransform

class MatrixTransformNode : public Node
{
public:

    MatrixTransformNode(int aid, char *afieldspec, Field *afield,
                        int fline, int lline);

    virtual void setupDepends(     //The field with field-id FID_MATRIX defines
                   State *astate); //a geometric 3D transformation with a 4 by 4
                                   //matrix. The matrix is accumulated into the
                                   //current transformation
                                   //(astate->currentTransMatrix)
};

//----------------------------------------------------------------------------
//used for VRML-node Translation

class TranslationNode : public Node
{
public:

    TranslationNode(int aid, char *afieldspec,Field *afield,
                    int fline, int lline);

    virtual void setupDepends(     //The Field with field-id FID_TRANSLATION
                   State *astate); //defines a translation by a3D vector. The
                                   //translation is accumulated into the
                                   //current transformation
                                   //(astate->currentTransMatrix)
};

//----------------------------------------------------------------------------
//used for VRML-node Coordinate3
//The print-method inserts all coordinates defined in this node (Value-list
//of Field with field-id FID_POINT) in an OctreeCoordSet-object (created
//in constructor) and deletes the list. The CoordSet removes doublettes and
//is also able to print the coordinates - the CoordSet has capabilities
//required for index-searching and lod-generation (see CoordSet-OctreeCoordSet)

class CoordNode : public Node
{
public:

    CoordSet *coordSet;     //pointer to a OctreeCoordSet-object
                            //created in constructor - filled with data
                            //in the print-method
    Node *lastShapeUser;    //pointer to the last FaceNode or LineNode
                            //which uses coordinates from this node
    Node *lastOtherUser;    //pointer to the last (ConeNode,CubeNode,...)
                            //which uses coordinates from this node
    static int verbose;     //CoordNode prints out information to stdout if !=0
    int pointcount;         //necessary for join

    CoordNode(int aid, char *afieldspec, Field *afield, int fline, int lline);
    virtual ~CoordNode();

    virtual void print(FILE *fd,    //prints out the node to fd
                       int indent); //the method inserts all coordinates defined
                                    //in this node (Value-list of Field with
                                    //field-id FID_POINT) in an OctreeCoordSet-
                                    //object (created in constructor) and deletes
                                    //the list - then the CoordSet is used to
                                    //print the coordinates

    virtual void setupDepends(State *astate); //sets astate->lastCoordNode
                                              //to this node
    
    virtual int joinable(Node * node);   //checks if two nodes are joinable
                                         //(compares id and Field objects)
    
    virtual void makeJoin(Node * node);  //join 2 CoordNodes
};

//----------------------------------------------------------------------------
//used for VRML-node Material
//The print-method inserts all materials defined in this node (Value-lists
//of Fields with field-id FID_AMBIENTCOLOR,FID_DIFFUSECOLOR,FID_SPECULARCOLOR,
//FID_EMISSIVECOLOR,FID_SHININESS,FID_TRANSPARENCY) in an MaterialSet-object
//(created in constructor) and deletes the lists. The MaterialSet removes
//doublettes and is also able to print the materials - the MaterialSet has
//capabilities required for index-searching (see MaterialSet)

class MaterialNode : public Node
{
public:
   
    MaterialSet *materialSet; //pointer to a MaterialSet-object
                              //created in constructor - filled with data
                              //in the print-method
    Node *lastShapeUser;      //pointer to the last FaceNode or LineNode
                              //which uses materials of this node
    Node *lastOtherUser;      //pointer to the last (ConeNode,CubeNode,...)
                              //which uses materials of this node
    static int verbose;       //node prints out information to stdout if !=0

    MaterialNode(int aid, char *afieldspec, Field *afield,int fline, int lline);
    virtual ~MaterialNode();

    virtual void print(FILE *fd,     //prints out the node to fd
                       int indent);  //the method inserts all materials defined
                                     //in this node (Value-lists of Fields)
                                     //in an MaterialSet-object (created in
                                     //constructor) and deletes the lists -
                                     //then the MaterialSet is used to print
                                     //the materials

    virtual void setupDepends(State *astate); //sets astate->lastMaterialNode
                                              //to this node
    
    virtual int joinable(Node * node);    //checks if two nodes are joinable
                                          //(compares id and Field objects)
    
    virtual void makeJoin(Node * node);   //join 2 MaterialNodes

    void createMaterialIndex(Node * node, //create a materialIndex for a
                             int index);  //ShapeNode, so it is possible to join
                                          //2 ShapeNodes with different
                                          //MaterialNodes

    virtual void printTree(int indent);   //print out a schema of the
                                          //node-tree to stdout
};

//----------------------------------------------------------------------------
//used for VRML-node MaterialBinding

class MaterialBindingNode : public Node
{
public:

    MaterialBindingNode(int aid, char *afieldspec, Field *afield,
                        int fline, int lline);

    virtual void setupDepends(State *astate); //set materialbinding in astate to
                                              //the binding defined in this node

    virtual int joinable(Node * node); //check if 2 nodes are MaterialBindingNodes
                                       //and use the same binding

    virtual void printTree(int indent);   //print out a schema of the
                                          //node-tree to stdout
};

//----------------------------------------------------------------------------
//used for VRML-node Normal
//The print-method inserts all normals defined in this node (Value-lists
//of Fields with field-id FID_VECTOR) in an NormalSet-object
//(created in constructor) and deletes the lists. The NormalSet removes
//doublettes and is also able to print the normals - the NormalSet has
//capabilities required for index-searching (see NormalSet)

class NormalNode : public Node
{
public:

    NormalSet *normalSet; //pointer to a NormalSet-object
                          //created in constructor - filled with data
                          //in the print-method
    Node *lastShapeUser;  //pointer to the last FaceNode or LineNode
                          //which uses normals of this node
    Node *lastOtherUser;  //pointer to the last (ConeNode,CubeNode,...)
                          //which uses normals of this node
    static int verbose;   //node prints out information to stdout if !=0
    int pointcount;       //necessary for join

    NormalNode(int aid, char *afieldspec, Field *afield, int fline, int lline);
    virtual ~NormalNode();

    virtual void print(FILE *fd,    //prints out the node to fd
                       int indent); //The print-method inserts all normals
                                    //defined in this node (Value-lists of 
                                    //Fields with field-id FID_VECTOR) in an
                                    //NormalSet-object (created in constructor)
                                    //and deletes the lists - then the NormalSet
                                    //is used to print the normals

    virtual void setupDepends(State *astate); //sets astate->lastNormalNode
                                              //to this node

    virtual int joinable(Node * node);   //checks if two nodes are joinable
                                         //(compares id and Field objects)
    virtual void makeJoin(Node * node);  //join 2 nodes
};

//----------------------------------------------------------------------------
//used for VRML-node NormalBinding

class NormalBindingNode : public Node
{
public:

    NormalBindingNode(int aid, char *afieldspec, Field *afield,
                      int fline, int lline);

    virtual void setupDepends(State *astate); //sets normalbinding in astate to
                                              //the binding defined in this node

    virtual int joinable(Node * node); //check if 2 nodes are NormalBindingNodes
                                       //and use the same binding
};

//----------------------------------------------------------------------------
//used for VRML-node TextureCoordinate2
//The print-method inserts all texturecoords defined in this node (Value-list
//of Field with field-id FID_POINT) in an TextureCoordSet-object (created in
//constructor) and deletes the list. The TextureCoordSet removes doublettes and 
//is also able to print the normals - the textureCoordSet has capabilities 
//required for index-searching (see TextureCoordSet)

class TextureCoordNode : public Node
{
public:

    TextureCoordSet *textureCoordSet;  //pointer to a TextureCoordSet-object
                                       //created in constructor - filled
                                       //with data in the print-method

    State state;          //state of this node - is set in setupDepends
    Node *lastShapeUser;  //pointer to the last FaceNode or LineNode
                          //which uses texturecoords of this node
    Node *lastOtherUser;  //pointer to the last (ConeNode,CubeNode,...)
                          //which uses texturecoords of this node
    static int verbose;   //node prints out information to stdout if !=0
    int pointcount;       //necessary for join

    TextureCoordNode(int aid, char *afieldspec, Field *afield,
                     int fline, int lline);
    virtual ~TextureCoordNode();

    virtual void print(FILE *fd,    //prints out the node to fd
                       int indent); //The print-method inserts all texturecoords
                                    //defined in this node (Value-list of Field
                                    //with field-id FID_POINT) in an
                                    //TextureCoordSet-object (created in 
                                    //constructor) and deletes the list - then
                                    //the TextureCoordSet is used to
                                    //print texturecoords

    virtual void setupDepends(      //sets astate->lastTextureCoordNode to this
                   State *astate);  //node - sets TextureCoordNode::state to
                                    //astate - checks for unsupported 
                                    //combinations

    virtual int joinable(Node * node);  //checks if two nodes are joinable
                                        //(compares id and Field objects)
                                        //it is also necessary that every node
                                        //uses the same TextureNode
};

//----------------------------------------------------------------------------
//used for VRML-node TextureNode

class TextureNode : public Node
{
public:

    TextureNode(int aid, char *afieldspec, Field *afield, int fline, int lline);

    virtual void setupDepends(State *astate); //sets astate->lastTextureNodei
                                              //to this node
};

//----------------------------------------------------------------------------
//This class is a (abstract) superclass of class FaceNode and class LineNode.
//It contains methods which are common to the two subclasses

class ShapeNode : public Node
{
public:
    State state;             //state of this node - is set in setupDepends
    int newMaterialBinding;  //some bindings are not acceptable by this node
    int newNormalBinding;    //these bindings will be transformed - this
                             //variables contains the new used bindings -
                             //the variables are set in setupDepends

    ShapeNode(int aid, char *afieldspec, Field *afield, int fline, int lline);

    virtual void setupDepends(      //sets state, newMaterialBinding,
                   State *astate);  //newNormalBinding, sets var lastShapeUser 
                                    //of CoordNode,MaterialNode,... in the
                                    //current scope to this and other things
                                    //see snode.C for a detailed description

    void cleanUpLast();      //checks if this node is the last user of the
                             //CoordNode, MaterialNode, NormalNode and
                             //TextureCoordNode in the current scope - if this
                             //is true the data of these nodes will be deleted

    Coord *getCoordIndex(CoordSet *cs,              //these functions return
                         int index);                //a pointer to an element
    Material *getMaterialIndex(MaterialSet *ms,     //of the specified Set
                               int index);          //by index (insertion 
    Normal *getNormalIndex(NormalSet *ns,           //order !) 
                           int index);              //an error occures if
    TextureCoord *getTextureCoordIndex(             //the index is out of range
                    TextureCoordSet *ts,int index); //-> program exits
    
    virtual int joinable(Node * node);  //checks if 2 ShapeNodes are joinable
    
    virtual void makeJoin(Node * node); //join 2 ShapeNodes (Line or Face)
                                        //join the fields

    virtual void printTree(int indent); //print out a schema of the
                                        //node-tree to stdout
};

//----------------------------------------------------------------------------
//used for VRML-node IndexedFaceSet
//It is able to transform the index-lists to a FaceSet and produce lods
//and print out the lods - these actions are performed in the print-method.

class FaceNode : public ShapeNode
{
public:
    FaceSet *faceSet;     //pointer to a FaceSet-object - created and filled
                          //with data in the method generateFaceSet
    static int verbose;   //node prints out information to stdout if !=0

    FaceNode(int aid, char *afieldspec, Field *afield, int fline, int lline);
    virtual ~FaceNode();

    virtual void print(FILE *fd,     //The print-method calls the method
                       int indent);  //generateFaceSet wich reads the index-
                                     //lists and generates a FaceSet. Then the
                                     //method generateLODs will be called.
                                     //Then the print-method prints out the lods
                                     //- additionally some nodes are created
                                     //temporary and printed out if required
                                     // (e.g. MaterialBinding,Separator, ...)

    void generateFaceSet();  //The method creates a FaceSet - reads the index-
                             //lists of the fields with field-id FID_COORDINDEX,
                             //FID_MATERIALINDEX, FID_NORMALINDEX,FID_TEXTURE-
                             //COORDINDEX creates Faces using the current bindings
                             //(this a very complicated task because there are
                             //sooo many bindings). Polygons with more than 4
                             //vertices will be triangulated with an object of
                             //class Triangulator. for more information see
                             //snodes.C

    void generateLODs();     //generate the lods for this FaceNode !
                             //The lods are stored in separate FaceSets building
                             //a linear list - for more information see snode.C

    void deleteFaceSets();   //delete all FaceSets (==lods) of this node
};

//----------------------------------------------------------------------------
//used for VRML-node IndexedLineSet
//It is able to transform the index-lists to a LineSet and produce lods
//and print out the lods - these actions are performed in the print-method.

//The print-method calls the method generateLineSet wich reads the index-lists
//and generates a LineSet. Then the method generateLODs will be called.
//Then the print-method prints out the lods - additionally some nodes are
//created temporary and printed out if required (e.g. MaterialBinding,
//ShapeHints,Separator, ...)

class LineNode : public ShapeNode
{
public:
    LineSet *lineSet;

    static int verbose;

    LineNode(int aid, char *afieldspec, Field *afield, int fline, int lline);
    virtual ~LineNode();

    virtual void print(FILE *fd,     //The print-method calls the method
                       int indent);  //generateLineSet wich reads the index-
                                     //lists and generates a LineSet. Then the
                                     //method generateLODs will be called.
                                     //Then the print-method prints out the lods
                                     //- additionally some nodes are created
                                     //temporary and printed out if required
                                     //(e.g. MaterialBinding,Separator, ...)

    void generateLineSet();  //The method creates a LineSet - reads the index-
                             //lists of the fields with field-id FID_COORDINDEX,
                             //FID_MATERIALINDEX, FID_NORMALINDEX, FID_TEXTURE-
                             //COORDINDEX creates Lines using the current bindings
                             //(this a very complicated task because there are
                             //sooo many bindings). Polylines will be split into
                             //single lines. This Lines will be inserted into the
                             //LineSet - the index-lists will be deleted !

    void generateLODs();     //generate the lods for this LineNode !
                             //The lods are stored in separate LineSets building
                             //a linear list - for more information see snode.C

    void deleteLineSets();   //delete all LineSets (==lods) of this node
};

//----------------------------------------------------------------------------
//used for VRML-node PointSet - overrides setupDepends

class PointNode : public Node
{
public:

    PointNode(int aid, char *afieldspec, Field *afield, int fline, int lline);

    virtual void setupDepends(      //this method evt. increases the variable
                   State *astate);  //countUncompressed of the DataSets in the
                                    //nodes CoordNode, MaterialNode, NormalNode,
                                    //TextureCoordNode in the current scope
                                    //for more information see snodes.C 
};

//----------------------------------------------------------------------------
//used for VRML-node Cone - overrides setupDepends

class ConeNode : public Node
{
public:

    ConeNode(int aid, char *afieldspec, Field *afield, int fline, int lline);

    virtual void setupDepends(      //if the current material binding is PER_PART
                   State *astate);  //or PER_PART_INDEXED the method sets the
                                    //variable MaterialSet::countUncompressed
                                    //of the MaterialNode in the current scope
                                    //to the maximum of 2 and the former value 
                                    //additionally the method sets the variable
                                    //lastOtherUser of the MaterialNode to this
                                    //for more information see snodes.C
};

//----------------------------------------------------------------------------
//used for VRML-node Cube - overrides setupDepends

class CubeNode : public Node
{
public:

    CubeNode(int aid, char *afieldspec, Field *afield, int fline, int lline);

    virtual void setupDepends(      //if the current material binding is PER_PART
                   State *astate);  //(_INDEXED) or PER_FACE(_INDEXED) the method
                                    //sets the variable MaterialSet::countUncompressed
                                    //of the MaterialNode in the current scope
                                    //to the maximum of 6 and the former value
                                    //additionally the method sets the variable
                                    //lastOtherUser of the MaterialNode to this
                                    //for more information see snodes.C
};

//----------------------------------------------------------------------------
//used for VRML-node Cylinder - overrides setupDepends

class CylinderNode : public Node
{
public:

    CylinderNode(int aid, char *afieldspec, Field *afield, int fline, int lline);

    virtual void setupDepends(      //if the current material binding is PER_PART
                   State *astate);  //or PER_PART_INDEXED the method sets the
                                    //variable MaterialSet::countUncompressed
                                    //of the MaterialNode in the current scope
                                    //to the maximum of 3 and the former value
                                    //additionally the method sets the variable
                                    //lastOtherUser of the MaterialNode to this
                                    //for more information see snodes.C
};

//----------------------------------------------------------------------------
//used for VRML-node Cylinder - overrides setupDepends

class SphereNode : public Node
{
public:

    SphereNode(int aid, char *afieldspec, Field *afield, int fline, int lline);

    virtual void setupDepends(      //the method sets the variable lastOtherUser
                   State *astate);  //of the MaterialNode in the current scope
                                    //to this
};

//----------------------------------------------------------------------------
//used for VRML-node ShapeHints

class ShapeHintsNode : public Node
{
public:

    ShapeHintsNode(int aid, char *afieldspec, Field *afield,
                   int fline, int lline);

    virtual void setupDepends(State *astate); //sets vertexOrdering and
                                              //faceType in astate to the
                                              //values defined in this node
};

//----------------------------------------------------------------------------
//used to store nodes which are not known by this program
//alist stores the whole body of the node (evt. incl. subnodes)

class UnknownNode : public Node
{
public:
    const char *idstr;  //string containing the (unknown) type of the node

    UnknownNode(const char *aidstr, char *alist, int fline, int lline);

    virtual void print(FILE *fd,     //prints out the unknown node 
                       int indent);  //to fd (not formated)
};

//----------------------------------------------------------------------------
//used for VRML-node PerspectiveCamera - only necessary for finding this
//type of camera via setupCamera

class PerspectiveCameraNode : public Node
{
public:

    PerspectiveCameraNode(int aid, char *afieldspec, Field *afield,
                          int fline, int lline);

    virtual void setupCamera(       //if no PerspectiveCamera was found so far
                   Node **camera);  //camera is set to this otherwise if this
                                    //camera and the first have different height
                                    //angles a warning will be printed
};

//----------------------------------------------------------------------------
//used for VRML-node OrthographicCamera - the only difference to 
//class Node is that setupCamera simply produces a warning

class OrthographicCameraNode : public Node
{
public:

    OrthographicCameraNode(int aid, char *afieldspec, Field *afield,
                           int fline, int lline);

    virtual void setupCamera(       //this method simply produces a warning
                   Node **camera);  //if lods are generated because lods makes
                                    //no sense with orthographic cameras
};

//----------------------------------------------------------------------------

#endif
