// ===========================================================================
//  Copyright 1996 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       normal
//  TYPE:       c++ code
//  PROJECT:    lodestar
//  CONTENT:    implementation of class NormalSet and class Normal
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rs      Reinhard Sainitzer
//		hb      Herbert Buchegger
// ===========================================================================
//  HISTORY:
//
// ===========================================================================

#include "normal.h"
#include "enum.h"

//----------------------------------------------------------------------------
//   class Normal : Data
//----------------------------------------------------------------------------

//---Normal::Normal-----------------------------------------------------------
// Input : none
// Output: none

Normal::Normal() : Data()
{
    x=0;
    y=0;
    z=0;

    left=NULL;
    right=NULL;
}

//---Normal::Normal-----------------------------------------------------------
// Input : float ax,ay,az          : init
// Output: none

Normal::Normal(float ax, float ay, float az) : Data()
{
    x=ax;
    y=ay;
    z=az;

    left=NULL;
    right=NULL;
}

//---Normal::print------------------------------------------------------------
// Input : FILE *fd                : VRML output file descriptor
//         int fid                 : field id
// Output: none
//write x z y separated by blanks to fd
//increase global variable global_cpl at 3. This enables the method
//DataSet::printSet making carriage returns if global_cpl >= global_floatcpl

void Normal::print(FILE *fd, int fid)
{
    printFloat(fd,x); fprintf(fd," ");
    printFloat(fd,y); fprintf(fd," ");
    printFloat(fd,z);
    global_cpl+=3;
}

//---Normal::compareWith------------------------------------------------------
// Input : Normal *n : compare this Normal with n
// Output: int       : -1,0 or 1
//compare the Normal with Normal n
//if x<n->x return -1 if x>n->x return 1
//else same with y else same with z else return 0
//used for inserting in a binary tree

int Normal::compareWith(Normal *n)
{
    if (x<n->x) return -1;
    if (x>n->x) return 1;
    if (y<n->y) return -1;
    if (y>n->y) return 1;
    if (z<n->z) return -1;
    if (z>n->z) return 1;
    return 0;
}

//---Normal::deviationTo------------------------------------------------------
// Input : Normal *n : calculate the deviation of this Normal to n
// Output: float     : the deviation

float Normal::deviationTo(Normal *n)
{
    if (n==this) return 0;

    float dx = x - n->x;
    float dy = y - n->y;
    float dz = z - n->z;

    return dx*dx + dy*dy + dz*dz;
}

//---Normal::meanValue--------------------------------------------------------
// Input : Normal *n      : calc the mean value of this Normal and n
//         Normal *result : a pointer to a Normal for the result
// Output: Normal *result : result is written in this Normal

void Normal::meanValue(Normal *n, Normal *result)
{
    result->x = (x + n->x) /2;
    result->y = (y + n->y) /2;
    result->z = (z + n->z) /2;
}

//---Normal::weightedMeanValue------------------------------------------------
// Input : Normal *n      : calc the weighted mean value of this Normal and n
//         float w1       : weight for this Normal
//         float w2       : weight fot Normal n
//         Normal *result : a pointer to a Normal for the result
// Output: Normal *result : result is written in Normal result

void Normal::weightedMeanValue(Normal *n, float w1, float w2, Normal *result)
{
    float w = w1+w2;
    result->x = (w1*x + w2*n->x) /w;
    result->y = (w1*y + w2*n->y) /w;
    result->z = (w1*z + w2*n->z) /w;
}

//---Normal operator----------------------------------------------------------

Normal operator +(Normal n1, Normal n2)
{
    return Normal(n1.x+n2.x, n1.y+n2.y, n1.z+n2.z);
}

//---Normal operator----------------------------------------------------------

Normal operator -(Normal n1, Normal n2)
{
    return Normal(n1.x-n2.x, n1.y-n2.y, n1.z-n2.z);
}

//----------------------------------------------------------------------------
//   class NormalSet : DataSet
//----------------------------------------------------------------------------

//---NormalSet::NormalSet-------------------------------------------------------------------------
// Input : none
// Output: none

NormalSet::NormalSet() : DataSet()
{
    normalRoot=NULL;
}

//---NormalSet::~NormalSet-------------------------------------------------------------------------
// Input : none
// Output: none

NormalSet::~NormalSet()
{
    deleteTree(normalRoot);
}

//---NormalSet::insert-------------------------------------------------------------------------
// Input : Normal **n : p to p to the normal to insert
// Output: Normal **n : *n=NULL if normal is deleted (equal normal exists)
//first n is inserted in a binary tree using insertTree
//if an equal Normal already exists n is deleted and the pointer to n
//is redirected to the existing Normal
//then a reference to n is appended to DataSet::refArray using insertArray

void NormalSet::insert(Normal **n)
{
    insertTree(n,&normalRoot);
    insertArray(*n);
}

//---NormalSet::print-------------------------------------------------------------------------
// Input : FILE *fd   : VRML output file descriptor
//         int indent : indention of output
// Output:
//writes "vector [ ... ]" to fd where ... are the normals separated by commas
//these normals are printed using DataSet::printSet
//additionally DataSet::setindices is called 

void NormalSet::print(FILE *fd, int indent)
{
    setIndices();

    if (normalRoot==NULL) return;

    printIndent(fd,indent);
    fprintf(fd,"%s [ ",fieldIdString[FID_VECTOR]);
    printSet(fd,indent+9,0);
    fprintf(fd," ]\n");
}

//---NormalSet::insertTree-------------------------------------------------------------------------
// Input : Normal **n    : p to p to the normal to insert
//         Normal **root : p to p to the root of the binary tree to insert
// Input : Normal **n    : *n=NULL if normal is deleted (equal normal exists)
//         Normal **root : *root is modified eventually
//recursive method for inserting Normals in the binary tree
//the method uses Normal::compareWith for the decision at which
//subtree n has to be inserted
//if a normal with the same specifications already exists then n will be
//deleted and the pointer to n is redirected to the existing Normal

void NormalSet::insertTree(Normal **n, Normal **root)
{
    if ((*root)==NULL) 
    {
        //found a place to insert

        *root = *n;
        count++;
        return;
    }

    int result = (*n)->compareWith(*root);
    if (result<0) insertTree(n,&((*root)->left));
    else if (result>0) insertTree(n,&((*root)->right));

    else
    {
        //normal with same specifications already exists

        delete *n;
        *n = *root;
    }
}

//---NormlaSet::deleteTree-------------------------------------------------------------------------
// Input : Normal *root : root of the tree to delete
// Output: none

void NormalSet::deleteTree(Normal *root)
{
    if (root == NULL) return;
    deleteTree(root->left);
    deleteTree(root->right);
    delete root;
}

//----------------------------------------------------------------------------

