// ===========================================================================
//  Copyright 1996 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       normal
//  TYPE:       c++ header
//  PROJECT:    lodestar
//  CONTENT:    class Normal, class NormalSet
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rs      Reinhard Sainitzer
//		hb      Herbert Buchegger
// ===========================================================================
//  HISTORY:
//
// ===========================================================================
//  MANUAL:
//
//  The class NormalSet is used to store Normals defined in VRML-Nodes
//  from type Normal. It removes double defined Normals (see also DataSet).
//  NormalSet is derived from DataSet and Normal is derived from Data.
//  Normals which will be inserted in this class must be from type Normal
//  The normals are inserted using the method NormalSet::insert(Normal **n)
//  The Normals are inserted internally in a binary tree. Additionally
//  References to the Normals are appended to the ReferenceArray (see also
//  DataSet) to get the Normals by index in insertion order 
//
// SEE ALSO
//     DataSet
//
// ===========================================================================

#ifndef _NORMAL_H_
#define _NORMAL_H_

#include "dataset.h"
#include "global.h"
#include "fields.h"
#include <stdio.h>

//----------------------------------------------------------------------------

class Normal : public Data
{
public:

    float x;
    float y;
    float z;

    Normal *left;    //pointer to left successor for binary tree
    Normal *right;   //pointer to right successor for binary tree

    Normal();
    Normal(float ax, float ay, float az);

    virtual void print(FILE *fd, int fid);  //writes x y z to fd
                                            //fid is not used
    int compareWith(Normal *n);             //compare this with n - returns
                                            //-1, 0 or 1 used for inserting in
                                            //a binary tree

    //these methods are not used currently
    float deviationTo(Normal *n);
    void meanValue(Normal *n, Normal *result);
    void weightedMeanValue(Normal *n, float w1, float w2, Normal *result);
    friend Normal operator +(Normal n1, Normal n2);
    friend Normal operator -(Normal n1, Normal n2);
};

//----------------------------------------------------------------------------

class NormalSet : public DataSet
{
public:

    NormalSet();
    ~NormalSet();
    void insert(Normal **n);  //insert n in the NormalSet. n is inserted
                              //internally in a binary tree using insertTree.
                              //if an equal Normal already exists n is deleted
                              //and the pointer to n is redirected to the 
                              //existing Normal
                              //additionally a reference to n is appended to
                              //DataSet::refArray 

    void print(FILE *fd, int indent); //writes "vector [ ... ]" to fd where
                                      //... are the normals separated by commas
                                      //these normals are printed using 
                                      //DataSet::printSet.
                                      //additionally DataSet:setIndices is called
     
protected:
    Normal *normalRoot;      //root of the binary tree

    void insertTree(Normal **n, Normal **root); //insert n in the binary tree
                                                //specified by root. if an
                                                //equal Normal already exists
                                                //n is deleted and the pointer
                                                //to n is redirected to the
                                                //existing Normal

    void deleteTree(Normal *root);              //delete the binary tree
                                                //specified by root
};

//----------------------------------------------------------------------------

#endif

