// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       Option
//  TYPE:       c++ code
//  PROJECT:    VEGA Base Library
//  CONTENT:    commandline option parser
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 13:40:13  rft     last modification
//  23-mar-94 19:49:58  rft     created
// ===========================================================================

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "option.h"

Option * Option::optionList = 0;

Option::Option(
        const char * newLongName,
        const char * newShortName,
        int newSpecified
        )
:
    specified(newSpecified),
    longName(newLongName),
    shortName(newShortName),
    next(optionList)
{
    optionList = this;
}

int Option::isSpecified() const
{
    return specified;
}

Option * Option::findByLongName(const char * name)
{
    for (Option * optionPtr = optionList;
         optionPtr;
         optionPtr = optionPtr->next)
    {
        if (strcmp(name,optionPtr->longName) == 0)
            return optionPtr;
    }
    return (0);
}

Option * Option::findByShortName(const char * name)
{
    for (Option * optionPtr = optionList;
         optionPtr;
         optionPtr = optionPtr->next)
    {
        if (strcmp(name,optionPtr->shortName) == 0)
            return optionPtr;
    }
    return (0);
}

void Option::setSpecified()
{
    specified = 1;
}

int Option::missingParameter()
{
    return 0;
}

int Option::takesParameter(const char *)
{
    return 0;
}

int Option::parse(
        int & argc,
        char **  argv
        )
{
    int i;
    Option * optionPtr = 0;
    int numberOfOptionsAndParameters = 0;

    for (i = 1; i < argc; i++)
    {
        if (numberOfOptionsAndParameters > 0)
            argv[i - numberOfOptionsAndParameters] = argv[i];

        if (argv[i][0] == '-')
        {
            if (optionPtr && optionPtr->missingParameter())
                return 0;

            if (argv[i][1] == '-')
            {
                optionPtr = findByLongName(argv[i] + 2);
            }
            else
            {
                optionPtr = findByShortName(argv[i] + 1);
            }
            if (! optionPtr)
            {
                fprintf(stderr,"%s: unknown option '%s'\n",argv[0],argv[i]);
                return 0;
            }
            if (optionPtr->specified)
            {
                fprintf(stderr,"%s: option '%s' specified twice\n",argv[0],argv[i]);
                return 0;
            }
            optionPtr->setSpecified();
            numberOfOptionsAndParameters++;
        }
        else
        {
            if (optionPtr &&
                optionPtr->takesParameter(argv[i]) )
                numberOfOptionsAndParameters++;
        }
    }

    if (optionPtr && optionPtr->missingParameter())
        return 1;

    argc -= numberOfOptionsAndParameters;

    return 1;
}

OptionFlag::OptionFlag(
        const char * newLongName,
        const char * newShortName
        )
:
    Option(newLongName,newShortName,0)
{
  //    TR_FUN2(Option_tracer,OptionFlag,newLongName,newShortName);
}


OptionParameter::OptionParameter(
        const char * newLongName,
        const char * newShortName,
        int hasDefault
        )
:
    Option(newLongName,newShortName,0),
    parameter(hasDefault)
{}

void OptionParameter::setSpecified()
{
    Option::setSpecified();
    parameter = 0;
}

int OptionParameter::takesParameter(const char * par)
{
    if (!parameter)
    {
        setParameter(par);      
        parameter = 1;
        return 1;
    }
    return 0;
}

void OptionParameter::setParameter(const char *)
{
  //    EH_FATAL(Option_errors,"should be pure virtual");
}

int OptionParameter::missingParameter()
{
    if (!parameter)
    {
        fprintf(stderr,"missing option parameter\n");
        return 1;
    }
    return 0;
}

OptionString::OptionString(
        const char * newLongName,
        const char * newShortName
        )
:
    OptionParameter(newLongName,newShortName, 0), value(0)
{
}

OptionString::OptionString(
        const char * newLongName,
        const char * newShortName,
        const char * defaultValue
        )
:
    OptionParameter(newLongName,newShortName, 1), value(defaultValue)
{
}

const char * OptionString::getValue() const { return value; }

void OptionString::setParameter(const char * par) { value = par; }

OptionLong::OptionLong(
        const char * newLongName,
        const char * newShortName
        )
:
    OptionParameter(newLongName,newShortName, 0), value(0)
{
  //    TR_FUN2(Option_tracer,OptionLong,newLongName,newShortName);
}

OptionLong::OptionLong(
        const char * newLongName,
        const char * newShortName,
        int defaultValue
        )
:
    OptionParameter(newLongName,newShortName, 1), value(defaultValue)
{
}

long OptionLong::getValue() const { return value; }

void OptionLong::setParameter(const char * par) { value = atol(par); }

OptionDouble::OptionDouble(
        const char * newLongName,
        const char * newShortName
        )
:
    OptionParameter(newLongName,newShortName, 0), value(0)
{
}

OptionDouble::OptionDouble(
        const char * newLongName,
        const char * newShortName,
        double defaultValue
        )
:
    OptionParameter(newLongName,newShortName, 1), value(defaultValue)
{
}

double OptionDouble::getValue() const { return value; }

void OptionDouble::setParameter(const char * par) { value = atof(par); }

// ===========================================================================

