// ===========================================================================
//  Copyright 1996 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       pnodes
//  TYPE:       c++ code
//  PROJECT:    lodestar
//  CONTENT:    implementation of class CoordNode,
//              class MaterialNode, class MaterialBindingNode,
//              class NormalNode, class NormalBindingNode,
//              class TextureCoordNode, class TextureNode
//              class ShapeHintNode
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rs      Reinhard Sainitzer
//		hb      Herbert Buchegger
// ===========================================================================
//  HISTORY:
//
// ===========================================================================

#include "nodes.h"
#include "global.h"
#include "enum.h"
#include "tri.h"
#include <limits.h>
#include <assert.h>

//----------------------------------------------------------------------------
//   class CoordNode : Node
//----------------------------------------------------------------------------
//used for VRML-node Coordinate3
//The print-method inserts all coordinates defined in this node (Value-list
//of Field with field-id FID_POINT) in an OctreeCoordSet-object (created
//in constructor) and deletes the list. The CoordSet removes doublettes and
//is also able to print the coordinates - the CoordSet has capabilities
//required for index-searching and lod-generation (see CoordSet-OctreeCoordSet)

int CoordNode::verbose = 0;  //CoordNode prints out information
                             //to stdout if !=0

//---CoordNode::Node----------------------------------------------------------
// Input : int aid          : init
//         char *afieldspec : init
//         Field *afield    : init
//         int fline        : init
//         int lline        : init
// Output:

CoordNode::CoordNode(int aid, char *afieldspec, Field *afield, int fline, int lline) :
           Node(aid,afieldspec,afield,fline,lline)
{
    coordSet = new OctreeCoordSet();
    lastShapeUser = NULL;
    lastOtherUser = NULL;
}

//---CoordNode::~CoordNode----------------------------------------------------
// Input :
// Output:

CoordNode::~CoordNode()
{
    if (coordSet) delete coordSet;
}

//---CoordNode::print---------------------------------------------------------
// Input : FILE *fd    : VRML output file descriptor
//         int indent  : indention of output
// Output:
//The print-method inserts all coordinates defined in this node (Value-list
//of Field with field-id FID_POINT) in an OctreeCoordSet-object (created
//in constructor) and deletes the list. The CoordSet removes doublettes and
//is also able to print the coordinates - the CoordSet has capabilities
//required for index-searching and lod-generation (see CoordSet-OctreeCoordSet)

void CoordNode::print(FILE *fd, int indent)
{
    if (lastShapeUser==NULL && lastOtherUser==NULL && used==0)
    {
        warning("node is useless (no references)\n");
        delete coordSet;
        coordSet=NULL;
        return;
    }

    printHead(fd,indent);

    Field *point = getField(FID_POINT);

    if (point) 
    {
        //---get bounds of this node

        Bounds bounds;
        FVec3f *vh = (FVec3f *)point->value;
        bounds.setTo(vh->x,vh->y,vh->z, vh->x,vh->y,vh->z);
        while (vh!=NULL)
        {
            bounds.enlargeIfOutside(vh->x,vh->y,vh->z);        
            vh=(FVec3f *)(vh->next);
        }
        coordSet->setBounds(bounds);
        if (verbose) 
        {
            message("coord","min= %f %f %f\n",
                    bounds.xmin, bounds.ymin, bounds.zmin);
            message("coord","max= %f %f %f\n",
                    bounds.xmax, bounds.ymax, bounds.zmax);
        }

        //---insert all points in the new CoordSet and delete List

        int lcount = 0;
        Coord *coord;
        vh = (FVec3f *)point->value;
        while (vh!=NULL)
        {
            coord = new OctreeCoord(vh->x,vh->y,vh->z,0);
            coordSet->insert(&coord);
            deleteValues((Value **)(&vh),1);
            lcount++;
        }
        point->value = NULL;

        //---print the set

        coordSet->print(fd,indent);
        if (verbose) message("coord","list=%d set=%d uncompressed=%d\n",
                         lcount,coordSet->count,coordSet->countUncompressed);
    }
    else warning("no points\n");

    printTail(fd,indent);

    if (lastShapeUser==NULL)    //if no IndexedFaceSet or IndexedLineSet
    {                           //uses this coordinates they can be deleted
        delete coordSet;
        coordSet=NULL;
    }
}

//---CoordNode::setupDepends--------------------------------------------------
// Input : State *astate    : current state
// Output: State *astate    : new state
// sets astate->lastCoordNode to this node

void CoordNode::setupDepends(State *astate)
{
    astate->lastCoordNode = this;
}

//---CoordNode::joinable------------------------------------------------------
// Input : Node *node       : second node for compare
// Output: int              : TRUE or FALSE
// compare 2 Nodes

int CoordNode::joinable(Node * node)
{
    if (id!=node->id) return(0);
    if (used || node->used) return(0);

    if (!compare3DField(getField(FID_POINT),node->getField(FID_POINT)))
        joinmode|=JM_COORD;

    return(1);
}

//---CoordNode::makeJoin------------------------------------------------------
// Input : Node *node       : join of 2 CoordNodes
// Output:
// join 2 nodes

void CoordNode::makeJoin(Node * node)
{
    if (joinmode&JM_COORD)
    {
        makeJoinField3D(node,FID_POINT);
        ((CoordNode *)node)->pointcount+=pointcount;
    }
}

//----------------------------------------------------------------------------
//   class MaterialNode : Node
//----------------------------------------------------------------------------
//used for VRML-node Material
//The print-method inserts all materials defined in this node (Value-lists
//of Fields with field-id FID_AMBIENTCOLOR,FID_DIFFUSECOLOR,FID_SPECULARCOLOR,
//FID_EMISSIVECOLOR,FID_SHININESS,FID_TRANSPARENCY) in an MaterialSet-object
//(created in constructor) and deletes the lists. The MaterialSet removes
//doublettes and is also able to print the materials - the MaterialSet has
//capabilities required for index-searching (see MaterialSet)

int MaterialNode::verbose = 0;  //MaterialNode prints out information
                                //to stdout if !=0

//---MaterialNode::MaterialNode-----------------------------------------------
// Input : int aid          : init
//         char *afieldspec : init
//         Field *afield    : init
//         int fline        : init
//         int lline        : init
// Output:

MaterialNode::MaterialNode(int aid, char *afieldspec, Field *afield,
                           int fline, int lline) :
              Node(aid,afieldspec,afield,fline,lline)
{
    materialSet=new MaterialSet();
    lastShapeUser = NULL;
    lastOtherUser = NULL;
}

//---MaterialNode::~MaterialNode----------------------------------------------
// Input :
// Output:

MaterialNode::~MaterialNode()
{
    if (materialSet) delete materialSet;
}

//---MaterialNode::print------------------------------------------------------
// Input : FILE *fd         : VRML output file descriptor
//         int indent       : indention of output
// Output:
//The print-method inserts all material defined in this node (Value-lists
//of Fields with field-id FID_AMBIENTCOLOR,FID_DIFFUSECOLOR,FID_SPECULARCOLOR,
//FID_EMISSIVECOLOR,FID_SHININESS,FID_TRANSPARENCY) in an MaterialSet-object
//(created in constructor) and deletes the lists. The MaterialSet removes
//doublettes and is also able to print the materials - the MaterialSet has
//capabilities required for index-searching (see MaterialSet)

void MaterialNode::print(FILE *fd, int indent)
{
    if (lastShapeUser==NULL && lastOtherUser==NULL && used==0)
    {
        warning("node is useless (no references)\n");
        delete materialSet;
        materialSet=NULL;
        return;
    }

    printHead(fd,indent);

    if (field==NULL) warning("empty node\n");

    //--- find out which fields are defined

    FVec3f *ambV   = NULL;
    FVec3f *diffV  = NULL;
    FVec3f *specV  = NULL;
    FVec3f *emisV  = NULL;
    FFloat *shinV  = NULL;
    FFloat *transV = NULL;
    Field *ambientColor = getField(FID_AMBIENTCOLOR);
    Field *diffuseColor = getField(FID_DIFFUSECOLOR);
    Field *specularColor = getField(FID_SPECULARCOLOR);
    Field *emissiveColor = getField(FID_EMISSIVECOLOR);
    Field *shininess = getField(FID_SHININESS);
    Field *transparency = getField(FID_TRANSPARENCY);
    int status = 0;
   
    if (ambientColor)  
    {
        ambV=(FVec3f *)ambientColor->value;
        ambientColor->value = NULL;
        status |= MAT_STAT_AMB;
    }
    if (diffuseColor)
    {
        diffV=(FVec3f *)diffuseColor->value;
        diffuseColor->value = NULL;
        status |= MAT_STAT_DIFF;
    }
    if (specularColor)
    {
        specV=(FVec3f *)specularColor->value;
        specularColor->value = NULL;
        status |= MAT_STAT_SPEC;
    }
    if (emissiveColor)
    {
        emisV=(FVec3f *)emissiveColor->value;
        emissiveColor->value = NULL;
        status |= MAT_STAT_EMIS;
    }
    if (shininess)
    {
        shinV=(FFloat *)shininess->value;
        shininess->value = NULL;
        status |= MAT_STAT_SHIN;
    }
    if (transparency)
    {
        transV=(FFloat *)transparency->value;
        transparency->value = NULL;
        status |= MAT_STAT_TRANS;
    }

    if (verbose)
    {
        message("material","fields:");
        if (ambientColor)  printf(" amb");
        if (diffuseColor)  printf(" diff");
        if (specularColor) printf(" spec");
        if (emissiveColor) printf(" emis");
        if (shininess)     printf(" shin");
        if (transparency)  printf(" trans");
        printf("\n");
    }

    //---insert materials in MaterialSet and delete lists

    int lcount = 0;
    Material *mat;
    RGB ambH, diffH, specH, emisH; 
    float shinH, transH;

    diffH=specH=emisH=ambH= RGB(-1,-1,-1);
    shinH=transH = -1;

    while (ambV || diffV || specV || emisV || shinV || transV)
    {
        if (ambientColor)  if (ambV) ambH.setTo(ambV);
        if (diffuseColor)  if (diffV) diffH.setTo(diffV);
        if (specularColor) if (specV) specH.setTo(specV);
        if (emissiveColor) if (emisV) emisH.setTo(emisV);
        if (shininess)     if (shinV) shinH=shinV->number;
        if (transparency)  if (transV) transH=transV->number;

        mat = new Material(ambH,diffH,specH,emisH,shinH,transH,status);
        materialSet->insert(&mat);

        deleteValues((Value **)(&ambV),1);
        deleteValues((Value **)(&diffV),1);
        deleteValues((Value **)(&specV),1);
        deleteValues((Value **)(&emisV),1);
        deleteValues((Value **)(&shinV),1);
        deleteValues((Value **)(&transV),1);
        lcount++;
    }

    materialSet->print(fd,indent);

    if (verbose) message("material","list=%d set=%d uncompressed=%d\n",
                     lcount,materialSet->count,materialSet->countUncompressed);

    printTail(fd,indent);

    if (lastShapeUser==NULL)    //if no IndexedFaceSet or IndexedLineSet
    {                           //uses this materials they can be deleted
        delete materialSet;
        materialSet=NULL;
    }
}

//---MaterialNode::setupDepends-----------------------------------------------
// Input : State *astate     : current state
// Output: State *astate     : new state
// sets astate->lastMaterialNode to this node

void MaterialNode::setupDepends(State *astate)
{
    astate->lastMaterialNode = this;
}

//---MaterialNode::createMaterialIndex----------------------------------------
// Input : Node *node        :
//         int index         :
// Output:
// create a material index, necessary when two MaterialNodes are different.

void MaterialNode::createMaterialIndex(Node * node,int index)
{
        FLong * list = NULL;
        FLong * last = NULL;
        FLong * help = (FLong *)(node->getField(FID_COORDINDEX))->value;
        assert(help!=NULL);

        while (help!=NULL)
        {
            while (help!=NULL && help->number!=-1)
                help=(FLong *)(help->next);

            if (list==NULL)
            {
                list = new FLong(index);
                last = list;
            }
            else
            {
                last->next = new FLong(index);
                last=(FLong *)(last->next);
            }
            if (help!=NULL)
                help=(FLong *)(help->next);
        }

        Field * f = new Field(FID_MATERIALINDEX,list);
        f->next=node->field;
        node->field=f;
        ((FaceNode *)node)->state.currMaterialBinding=VRML_PER_FACE_INDEXED;
        ((FaceNode *)node)->newMaterialBinding=VRML_PER_FACE_INDEXED;
}

//---MaterialNode::joinable---------------------------------------------------
// Input : Node *node        : second node
// Output: int               : TRUE or FALSE
// join 2 MaterialNodes

int MaterialNode::joinable(Node * node)
{
    if (id!=node->id) return(0);
    if (used || node->used) return(0);

    if (!compare3DField(getField(FID_AMBIENTCOLOR),
                        node->getField(FID_AMBIENTCOLOR))) joinmode|=JM_MATERIAL;
    if (!compare3DField(getField(FID_DIFFUSECOLOR),
                        node->getField(FID_DIFFUSECOLOR))) joinmode|=JM_MATERIAL;
    if (!compare3DField(getField(FID_SPECULARCOLOR),
                        node->getField(FID_SPECULARCOLOR))) joinmode|=JM_MATERIAL;
    if (!compare3DField(getField(FID_EMISSIVECOLOR),
                        node->getField(FID_EMISSIVECOLOR))) joinmode|=JM_MATERIAL;
    if (!compareField(getField(FID_SHININESS),
                        node->getField(FID_SHININESS))) joinmode|=JM_MATERIAL;
    if (!compareField(getField(FID_TRANSPARENCY),
                        node->getField(FID_TRANSPARENCY))) joinmode|=JM_MATERIAL;

    return(1);
}

//---MaterialNode::makeJoin---------------------------------------------------
// Input : Node * node       : second node
// Output:
// join 2 material nodes

void MaterialNode::makeJoin(Node * node)
{
    if (joinmode&JM_MATERIAL)
    {
        makeJoinField3D(node,FID_AMBIENTCOLOR);
        makeJoinField3D(node,FID_DIFFUSECOLOR);
        makeJoinField3D(node,FID_SPECULARCOLOR);
        makeJoinField3D(node,FID_EMISSIVECOLOR);
        makeJoinField(node,FID_SHININESS);
        makeJoinField(node,FID_TRANSPARENCY);
    }
}

//---MaterialNode::printTree--------------------------------------------------
// Input : int indent     :
// Output:

void MaterialNode::printTree(int indent)
{
    Node::printTree(indent);

    printf(" (");
    Field * help = getField(FID_AMBIENTCOLOR);
    if (help!=NULL) {
        Value * v = help->value;
        int count = 0;
        while (v!=NULL) { v=v->next; count++; };
        printf(" ambient %d",count);
    }

    help = getField(FID_DIFFUSECOLOR);
    if (help!=NULL) {
        Value * v = help->value;
        int count = 0;
        while (v!=NULL) { v=v->next; count++; };
        printf(" diffuse %d",count);
    }

    help = getField(FID_SPECULARCOLOR);
    if (help!=NULL) {
        Value * v = help->value;
        int count = 0;
        while (v!=NULL) { v=v->next; count++; };
        printf(" specular %d",count);
    }

    help = getField(FID_EMISSIVECOLOR);
    if (help!=NULL) {
        Value * v = help->value;
        int count = 0;
        while (v!=NULL) { v=v->next; count++; };
        printf(" emissive %d",count);
    }

    help = getField(FID_SHININESS);
    if (help!=NULL) {
        Value * v = help->value;
        int count = 0;
        while (v!=NULL) { v=v->next; count++; };
        printf(" shininess %d",count);
    }

    help = getField(FID_TRANSPARENCY);
    if (help!=NULL) {
        Value * v = help->value;
        int count = 0;
        while (v!=NULL) { v=v->next; count++; };
        printf(" transparency %d",count);
    }
    printf(")");
}

//----------------------------------------------------------------------------
//   class MaterialBindingNode : Node
//----------------------------------------------------------------------------
//used for VRML-node MaterialBinding - overrides setupDepends

//---MaterialBindingNode::MaterialBindingNode---------------------------------
// Input : int aid           : init
//         char *afieldspec  : init
//         Field *afield     : init
//         int fline         : init
//         int lline         : init
// Output:

MaterialBindingNode::MaterialBindingNode(int aid, char *afieldspec, Field *afield,
                                         int fline, int lline) :
                     Node(aid,afieldspec,afield,fline,lline)
{
}

//---MaterialBindingNode::setupDepends----------------------------------------
// Input : State *astate  : current state
// Output: State *astate  : new state
// set materialbinding in astate to the binding defined in this node

void MaterialBindingNode::setupDepends(State *astate)
{
    Field *value = getField(FID_VALUE);
    if (value) astate->currMaterialBinding = ((FEnum *)value->value)->number;
          else astate->currMaterialBinding = VRML_DEFAULT;
}

//---MaterialBindingNode::joinable--------------------------------------------
// Input : Node *node        :
// Output: int               :
// compare 2 nodes

int MaterialBindingNode::joinable(Node * node)
{
    if (id!=node->id) return(0);
    if (used || node->used) return(0);

    if (!compareField(getField(FID_VALUE), node->getField(FID_VALUE)))
         return(0);

    return(1);
}

//---MaterialBindingNode::printTree-------------------------------------------
// Input : int indent     :
// Output:

void MaterialBindingNode::printTree(int indent)
{
    Node::printTree(indent);

    printf(" (");
    field->value->print(stdout);
    printf(")");
}

//----------------------------------------------------------------------------
//   class NormalNode : Node
//----------------------------------------------------------------------------
//used for VRML-node Normal
//The print-method inserts all normals defined in this node (Value-list
//of Field with field-id FID_VECTOR) in an NormalSet-object (created in
//constructor) and deletes the list. The NormalSet removes doublettes and 
//is also able to print the normals - the NormalSet has capabilities 
//required for index-searching (see NormalSet)

int NormalNode::verbose = 0;  //NormalNode prints out information
                              //to stdout if !=0

//---NormalNode::NormalNode---------------------------------------------------
// Input : int aid           : init
//         char *afieldspec  : init
//         Field *afield     : init
//         int fline         : init
//         int lline         : init
// Output:

NormalNode::NormalNode(int aid, char *afieldspec, Field *afield, int fline, int lline) :
            Node(aid,afieldspec,afield,fline,lline)
{
    normalSet = new NormalSet();
    lastShapeUser = NULL;
    lastOtherUser = NULL;
}

//---NormalNode::~NormalNode--------------------------------------------------
// Input :
// Output:

NormalNode::~NormalNode()
{
    if (normalSet) delete normalSet;
}

//---NormalNode::print--------------------------------------------------------
// Input : FILE *fd         : VRML output file descriptor
//         int indent       : indention of output
// Output:
//The print-method inserts all normals defined in this node (Value-list
//of Field with field-id FID_VECTOR) in an NormalSet-object (created in
//constructor) and deletes the list. The NormalSet removes doublettes and 
//is also able to print the normals - the NormalSet has capabilities 
//required for index-searching (see NormalSet)

void NormalNode::print(FILE *fd, int indent)
{
    if (lastShapeUser==NULL && lastOtherUser==NULL && used==0)
    {
        warning("node is useless (no references)\n");
        delete normalSet;
        normalSet=NULL;
        return;
    }

    printHead(fd,indent);

    Field *vector = getField(FID_VECTOR);

    if (vector) 
    {
        //---insert all points in the new NormalSet and delete List

        int lcount = 0;
        Normal *nor;
        FVec3f *vh = (FVec3f *)vector->value;
        while (vh!=NULL)
        {
            nor = new Normal(vh->x,vh->y,vh->z);
            normalSet->insert(&nor);
            deleteValues((Value **)(&vh),1);
            lcount++;
        }
        vector->value = NULL;

        normalSet->print(fd,indent);

        if (verbose) message("normal","list=%d set=%d uncompressed=%d\n",
                         lcount,normalSet->count,normalSet->countUncompressed);
    }
    else warning("no vectors\n");

    printTail(fd,indent);

    if (lastShapeUser==NULL)   //if no IndexedFaceSet or IndexedLineSet
    {                          //uses this normals they can be deleted
        delete normalSet;
        normalSet=NULL;
    }
}

//---NormalNode::setupDepends-------------------------------------------------
// Input : State *astate  : current state
// Output: State *astate  : new state
// sets astate->lastMaterialNode to this node

void NormalNode::setupDepends(State *astate)
{
    astate->lastNormalNode = this; 
}

//---NormalNode::joinable-----------------------------------------------------
// Input : Node *node       : second Node
// Output: int              : TRUE or FALSE
// compare 2 nodes

int NormalNode::joinable(Node * node)
{
    if (id!=node->id) return(0);
    if (used || node->used) return(0);

    if (!compareField(getField(FID_VECTOR),
                      node->getField(FID_VECTOR))) joinmode|=JM_NORMAL;

    return(1);
}

//---NormalNode::makeJoin-----------------------------------------------------
// Input : Node * node      : second node
// Output: 
// join 2 NormalNodes

void NormalNode::makeJoin(Node * node)
{
    if (joinmode&JM_COORD)
    {
        makeJoinField3D(node,FID_VECTOR);
        ((NormalNode *)node)->pointcount+=pointcount;
    }
}

//----------------------------------------------------------------------------
//   class NormalBindingNode : Node
//----------------------------------------------------------------------------
//used for VRML-node NormalBinding - overrides setupDepends

//---NormalBindingNode::NormalBindingNode-------------------------------------
// Input : int aid           : init
//         char *afieldspec  : init
//         Field *afield     : init
//         int fline         : init
//         int lline         : init
// Output:

NormalBindingNode::NormalBindingNode(int aid, char *afieldspec, Field *afield,
                                     int fline, int lline) :
                   Node(aid,afieldspec,afield,fline,lline)
{
}

//---NormalBindingNode::setupDepends------------------------------------------
// Input : State *astate  : current state
// Output: State *astate  : new state
// sets normalbinding in astate to the binding defined in this node

void NormalBindingNode::setupDepends(State *astate)
{
    Field *value = getField(FID_VALUE);
    if (value) astate->currNormalBinding = ((FEnum *)value->value)->number;
          else astate->currNormalBinding = VRML_DEFAULT;
}

//---NormalBindingNode::joinable----------------------------------------------
// Input : Node *node        : second node
// Output: int               : TRUE or FALSE
// compare 2 nodes

int NormalBindingNode::joinable(Node * node)
{
    if (id!=node->id) return(0);
    if (used || node->used) return(0);

    if (!compareField(getField(FID_VALUE),
                      node->getField(FID_VALUE))) return(0);

    return(1);
}

//----------------------------------------------------------------------------
//   class TextureCoordNode : Node
//----------------------------------------------------------------------------
//used for VRML-node TextureCoordinate2
//The print-method inserts all texturecoords defined in this node (Value-list
//of Field with field-id FID_POINT) in an TextureCoordSet-object (created in
//constructor) and deletes the list. The TextureCoordSet removes doublettes and 
//is also able to print the normals - the textureCoordSet has capabilities 
//required for index-searching (see TextureCoordSet)

int TextureCoordNode::verbose = 0;  //TextureCoordNode prints out information
                                    //to stdout if !=0

//---TextureCoordNode::TextureCoordNode---------------------------------------
// Input : int aid           : init
//         char *afieldspec  : init
//         Field *afield     : init
//         int fline         : init
//         int lline         : init
// Output:

TextureCoordNode::TextureCoordNode(int aid, char *afieldspec, Field *afield,
                                   int fline, int lline) :
                  Node(aid,afieldspec,afield,fline,lline)
{
    state = State();
    textureCoordSet = new TextureCoordSet();
    lastShapeUser = NULL;
    lastOtherUser = NULL;
}

//---TextureCoordNode::~TextureCoordNode--------------------------------------
// Input :
// Output:

TextureCoordNode::~TextureCoordNode()
{
    if (textureCoordSet) delete textureCoordSet;
}

//---TextureCoordNode::print--------------------------------------------------
// Input : FILE *fd      : VRML output file descriptor
//         int indent    : indention of output
// Output:
//The print-method inserts all texturecoords defined in this node (Value-list
//of Field with field-id FID_POINT) in an TextureCoordSet-object (created in
//constructor) and deletes the list. The TextureCoordSet removes doublettes and 
//is also able to print the normals - the textureCoordSet has capabilities 
//required for index-searching (see TextureCoordSet)

void TextureCoordNode::print(FILE *fd, int indent)
{
    if (lastShapeUser==NULL && lastOtherUser==NULL && used==0)
    {
        warning("node is useless (no references)\n");
        delete textureCoordSet;
        textureCoordSet=NULL;
        return;
    }

    printHead(fd,indent);

    Field *point = getField(FID_POINT);

    if (point) 
    {
        //---insert all points in the TextureCoordSet and delete List

        int lcount = 0;
        TextureCoord *tex;
        FVec3f *vh = (FVec3f *)point->value;
        while (vh!=NULL)
        {
            tex = new TextureCoord(vh->x,vh->y);
            textureCoordSet->insert(&tex);
            deleteValues((Value **)(&vh),1);
            lcount++;
        }
        point->value = NULL;

        textureCoordSet->print(fd,indent);

        if (verbose) message("texcoord","list=%d set=%d uncompressed=%d\n",
                         lcount,textureCoordSet->count,
                         textureCoordSet->countUncompressed);
    }
    else warning("no points\n");

    if (lastShapeUser==NULL)   //if no IndexedFaceSet or IndexedLineSet
    {                          //uses this texturecoords they can be deleted
        delete textureCoordSet;
        textureCoordSet=NULL;
    }

    printTail(fd,indent);
}

//---TextureCoordNode::setupDepends-------------------------------------------
// Input : State *astate  : current state
// Output: State *astate  : new state
// sets astate->lastTextureCoordNode to this node
// sets TextureCoordNode::state to astate
// checks for unsupported combinations

void TextureCoordNode::setupDepends(State *astate)
{
    //---if this method is called again from a USE-Node
    //---check for node combinations which are not supported by this program

    if (state.lastTextureNode!=NULL)
    {
        if (state.lastTextureNode!=astate->lastTextureNode)
            error("DEF and USE statements are using different Texture2-Nodes\n\
    sorry can't handle this\n");
    }

    astate->lastTextureCoordNode = this;
    state=*astate;
}

//---TextureCoordNode::joinable-----------------------------------------------
// Input : Node *node      : second node
// Output: int             : TRUE or FALSE
// compare 2 nodes

int TextureCoordNode::joinable(Node * node)
{
    if (id!=node->id) return(0);
    if (used || node->used) return(0);

    if (state.lastTextureNode!=
        ((TextureCoordNode *)node)->state.lastTextureNode) return 0;

    if (!compare3DField(getField(FID_POINT),
                        node->getField(FID_POINT))) joinmode|=JM_TEXTURE;

    return(1);
}

//----------------------------------------------------------------------------
//   class TextureNode : Node
//----------------------------------------------------------------------------
//used for VRML-node TextureNode - overrides setupDepends

//---TextureNode::TextureNode-------------------------------------------------
// Input : int aid           : init
//         char *afieldspec  : init
//         Field *afield     : init
//         int fline         : init
//         int lline         : init
// Output:

TextureNode::TextureNode(int aid, char *afieldspec, Field *afield,
                         int fline, int lline) :
             Node(aid,afieldspec,afield,fline,lline)
{
}

//---TextureNode::setupDepends------------------------------------------------
// Input : State *astate  : current state
// Output: State *astate  : new state
// sets astate->lastTextureNode to this node

void TextureNode::setupDepends(State *astate)
{
    astate->lastTextureNode = this;
}

//----------------------------------------------------------------------------
//   class ShapeHintsNode : Node
//----------------------------------------------------------------------------
//used for VRML-node ShapeHints - overrides setupDepends

//---ShapeHintsNode::ShapeHintsNode-------------------------------------------
// Input : int aid           : init
//         char *afieldspec  : init
//         Field *afield     : init
//         int fline         : init
//         int lline         : init
// Output:

ShapeHintsNode::ShapeHintsNode(int aid, char *afieldspec, Field *afield,
                               int fline, int lline):
                Node(aid,afieldspec,afield,fline,lline)
{
}

//---ShapeHintsNode::setupDepends---------------------------------------------
// Input : State *astate  : current state
// Output: State *astate  : new state
// sets vertexOrdering and faceType in astate to the values
// defined in this node

void ShapeHintsNode::setupDepends(State *astate)
{
    Field *vertexOrdering = getField(FID_VERTEXORDERING);
    Field *faceType = getField(FID_FACETYPE);

    if (vertexOrdering) 
         astate->currVertexOrdering = ((FEnum *)vertexOrdering->value)->number;
    else astate->currVertexOrdering = VRML_UNKNOWN_ORDERING;

    if (faceType)
         astate->currFaceType = ((FEnum *)faceType->value)->number;
    else astate->currFaceType = VRML_CONVEX;
}

//----------------------------------------------------------------------------

