// ===========================================================================
//  Copyright 1996 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       snodes
//  TYPE:       c++ code
//  PROJECT:    lodestar
//  CONTENT:    implementation of class ShapeNode, 
//              class FaceNode, class LineNode, class PointNode,
//              class ConeNode, class CubeNode, class CylinderNode, 
//              class SphereNode
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rs      Reinhard Sainitzer
//		hb      Herbert Buchegger
// ===========================================================================
//  HISTORY:
//
// ===========================================================================

#include "nodes.h"
#include "global.h"
#include "enum.h"
#include "tri.h"
#include <limits.h>
#include <assert.h>

//----------------------------------------------------------------------------
//   class ShapeNode : Node
//----------------------------------------------------------------------------
//This class is a (abstract) superclass of class FaceNode and class LineNode.
//It contains methods which are common to the two subclasses

//---ShapeNode::ShapeNode-----------------------------------------------------
// Input : int aid             : init
//         char *afieldspec    : init
//         Field *afield       : init
//         int fline           : init
//         int lline           : init
// Output:

ShapeNode::ShapeNode(int aid, char *afieldspec, Field *afield, int fline, int lline) :
           Node(aid,afieldspec,afield,fline,lline)
{
    state = State();
}

//---ShapeNode::setupDepends--------------------------------------------------
// Input : State *astate  : current state
// Output: State *astate  : state is not changed !
//
// if this method is called first time (ShapeNode::state is uninitialized)
// it will perform following actions:
//    -ShapeNode::state is set to astate
//    -if the value of state.currNormalBinding or state.currMaterialBinding is
//     VRML_DEFAULT then this value is changed to the real binding
//    -state.lastCoordNode->lastShapeUser = this;
//    -state.lastMaterialNode->lastShapeUser = this;
//    -state.lastNormalNode->lastShapeUser = this;
//    -state.lastTextureCoordNode->lastShapeUser=this;
//    -some bindings are not acceptable so ShapeNode::newMaterialBinding and
//     ShapeNode::newNormalBinding are set to the bindings used by this node
//
// if this method is called again from a USE-Node (ShapeNode::state is set)
// then it checks for unsupported combinations

void ShapeNode::setupDepends(State *astate)
{
    State s = *astate;

    if (s.currNormalBinding==VRML_DEFAULT)
        s.currNormalBinding=VRML_PER_VERTEX_INDEXED;

    if (s.currMaterialBinding==VRML_DEFAULT)
        s.currMaterialBinding=VRML_OVERALL;

    //---this method is called again from a USE-Node
    //---check for node combinations which are not supported by this program

    if (state.lastCoordNode!=NULL)
    {
        if (state.lastCoordNode!=s.lastCoordNode)
            error("DEF and USE statements are using different Coordinate3-Nodes\n\
    sorry can't handle this\n");

        if (state.currMaterialBinding!=s.currMaterialBinding)
            error("DEF and USE statements are using different Material-Bindings\n\
    sorry can't handle this\n");

        if (state.lastMaterialNode!=s.lastMaterialNode && 
            s.currMaterialBinding!=VRML_OVERALL)
            error("DEF and USE statements are using different Material-Nodes\n\
    and Material-Binding != OVERALL\n\
    sorry can't handle this\n");

        if (state.currNormalBinding!=s.currNormalBinding)
            error("DEF and USE statements are using different Normal-Bindings\n\
    sorry can't handle this\n");

        if (state.lastNormalNode!=s.lastNormalNode &&
            s.currNormalBinding!=VRML_OVERALL)
            error("DEF and USE statements are using different Normal-Nodes\n\
    and Normal-Binding != OVERALL\n\
    sorry can't handle this\n");

        if (state.lastTextureCoordNode!=s.lastTextureCoordNode)
            error("DEF and USE statements are using different TextureCoordinate2-Nodes\n\
    sorry can't handle this\n");
    }

    //---this method is called first time

    else
    {
        s.lastCoordNode->lastShapeUser = this;
        s.lastMaterialNode->lastShapeUser = this;
        s.lastNormalNode->lastShapeUser = this;
        s.lastTextureCoordNode->lastShapeUser=this;

        newMaterialBinding = s.currMaterialBinding;
        if (newMaterialBinding==VRML_PER_FACE) newMaterialBinding=VRML_PER_FACE_INDEXED;
        if (newMaterialBinding==VRML_PER_PART) newMaterialBinding=VRML_PER_PART_INDEXED;
        if (newMaterialBinding==VRML_PER_VERTEX) newMaterialBinding=VRML_PER_VERTEX_INDEXED;

        newNormalBinding = s.currNormalBinding;
        if (newNormalBinding==VRML_PER_FACE) newNormalBinding=VRML_PER_FACE_INDEXED;
        if (newNormalBinding==VRML_PER_PART) newNormalBinding=VRML_PER_PART_INDEXED;
        if (newNormalBinding==VRML_PER_VERTEX) newNormalBinding=VRML_PER_VERTEX_INDEXED;

        state = s;
    }
}

//---ShapeNode::cleanUpLast---------------------------------------------------
// Input : 
// Output:
// checks if this node is the last user of the CoordNode, MaterialNode,
// NormalNode and TextureCoordNode in the current scope - if this is true
// than the data of these nodes will be deleted

void ShapeNode::cleanUpLast()
{
    if (state.lastCoordNode->lastShapeUser==this)
    { 
        delete state.lastCoordNode->coordSet;
        state.lastCoordNode->coordSet = NULL;
    }
    if (state.lastMaterialNode->lastShapeUser==this)
    {
        delete state.lastMaterialNode->materialSet;
        state.lastMaterialNode->materialSet = NULL;
    }
    if (state.lastNormalNode->lastShapeUser==this)
    {
       delete state.lastNormalNode->normalSet;
       state.lastNormalNode->normalSet = NULL;
    }
    if (state.lastTextureCoordNode->lastShapeUser==this)
    {
        delete state.lastTextureCoordNode->textureCoordSet;
        state.lastTextureCoordNode->textureCoordSet = NULL;
    }
}

//---ShapeNode::getCoordIndex-------------------------------------------------
// Input : CoordSet *cs      : pointer to a CoordSet-object
//         int index         : index of the wanted element
// Output: Coord *           : pointer to the wanted element
// gets a pointer to an element of CoordSet cs by index (insertion order !)
// an error occures if the index is out of range

Coord *ShapeNode::getCoordIndex(CoordSet *cs, int index)
{
    if (index<0 || index>cs->getMaxIndex())
        error("index >%d< out of range in coordIndex\n",index);
    return (Coord *)cs->getIndex(index); 
}

//---ShapeNode::getMaterialIndex----------------------------------------------
// Input : MaterialSet *ms   : pointer to a MaterialSet-object
//         int index         : index of the wanted element
// Output: Material *        : pointer to the wanted element
// gets a pointer to an element of MaterialSet ms by index (insertion order !)
// an error occures if the index is out of range (<0)
// index if taken modulo by MaterialSet::getIndex

Material *ShapeNode::getMaterialIndex(MaterialSet *ms, int index)
{
    if (index<0)
        error("index >%d< out of range in materialIndex\n",index);
    return (Material *)ms->getIndex(index); 
}

//---ShapeNode::getNormalIndex------------------------------------------------
// Input : NormalSet *ns    : pointer to a NormalSet-object
//         int index        : index of the wanted element
// Output: Normal *         : pointer to the wanted element
// gets a pointer to an element of NormalSet ns by index (insertion order !)
// an error occures if the index is out of range

Normal *ShapeNode::getNormalIndex(NormalSet *ns, int index)
{
    if (index<0 || index>ns->getMaxIndex())
        error("index >%d< out of range in normalIndex\n",index);
    return (Normal *)ns->getIndex(index); 
}

//---ShapeNode::getTextureCoordIndex------------------------------------------
// Input : TextureCoordSet *ts  : pointer to a TextureCoordSet-object
//         int index            : index of the wanted element
// Output: TextureCoord *       : pointer to the wanted element
// gets a pointer to an element of TextureCoordSet ts by index (insertion order !)
// an error occures if the index is out of range

TextureCoord *ShapeNode::getTextureCoordIndex(TextureCoordSet *ts, int index)
{
    if (index<0 || index>ts->getMaxIndex())
        error("index >%d< out of range in textureCoordIndex\n",index);
    return (TextureCoord *)ts->getIndex(index); 
}

//---ShapeNode::joinable------------------------------------------------------
// Input : Node *node        : second node
// Output: int               : TRUE or FALSE
// compare 2 nodes

int ShapeNode::joinable(Node * node)
{
    if (id!=node->id) return(0);
    if (used || node->used) return(0);

    Field * f1,* f2;
    // check coord index (2 cases: there are 2 pointer or 0) -----------------
    f1=getField(FID_COORDINDEX);
    f2=node->getField(FID_COORDINDEX);
    if ((f1==NULL && f2!=NULL) || (f1!=NULL && f2==NULL)) return 0;


    // check material index --------------------------------------------------
    f1=getField(FID_MATERIALINDEX);
    f2=node->getField(FID_MATERIALINDEX);

    //if (f1==NULL && f2==NULL)
    //{
    //    everything is allowed: every binding, materialIndex and no materialIndex
    //}
    if (f1==NULL && f2!=NULL)
    {
        return 0;              // I don't support this
    }
    else if (f1!=NULL && f2==NULL)          // node 1 is prepared for join, node 2 not
    {
        if (joinmode&JM_MATERIAL)
        {
            if (((ShapeNode *)node)->state.currMaterialBinding!=VRML_OVERALL) return 0;
            if (state.currMaterialBinding!=VRML_PER_FACE_INDEXED &&
                state.currMaterialBinding!=VRML_OVERALL) return 0;
        }
        else return 0;         // new 29.05.1996
    }
    //else if (f1!=NULL && f2!=NULL)
    //{
    //    there must be the same material binding - see MaterialBindingNode::joinable();
    //}

    // I hope this case is impossible
    if (state.currMaterialBinding==VRML_PER_VERTEX_INDEXED && f1==NULL) return 0;
    // I hope this case is impossible
    if (((ShapeNode *)node)->state.currMaterialBinding==VRML_PER_VERTEX_INDEXED &&
        f2==NULL) return 0;


    // check normal index ----------------------------------------------------
    f1=getField(FID_NORMALINDEX);
    f2=node->getField(FID_NORMALINDEX);
    if ((f1==NULL && f2!=NULL) || (f1!=NULL && f2==NULL)) return 0;
    if (state.currNormalBinding==VRML_PER_VERTEX_INDEXED && f1==NULL &&
        state.lastNormalNode->firstline!=-1) return 0;
    if (((ShapeNode *)node)->state.currNormalBinding==VRML_PER_VERTEX_INDEXED &&
        f2==NULL && ((ShapeNode *)node)->state.lastNormalNode->firstline!=-1) return 0;


    // check texture coord index ---------------------------------------------
    f1=getField(FID_TEXTURECOORDINDEX);
    f2=node->getField(FID_TEXTURECOORDINDEX);

    if ((f1==NULL && f2!=NULL) || (f1!=NULL && f2==NULL)) return 0;
    if (state.lastTextureNode!=((ShapeNode *)node)->state.lastTextureNode) return 0;

    return 1;
}

//---ShapeNode::makeJoin------------------------------------------------------
// Input : Node *node       : second node
// Output:
// join to ShapeNodes

void ShapeNode::makeJoin(Node * node)
{
    // two CoordNodes will be joined, so the indices of the second list are
    // increment by the count of the first list
    if (joinmode&JM_COORD)
    {
        Field * f = node->getField(FID_COORDINDEX);
        if (f!=NULL)
        {
            FLong * v1 = (FLong *)(f->value);
            while (v1!=NULL)
            {
                if (v1->number!=-1) v1->number+=state.lastCoordNode->pointcount;
                v1=(FLong *)(v1->next);
            }
        }
    }

    // append the list of the second node to the list of the first node
    // use this new list as list of the second node
    joinFields(getField(FID_COORDINDEX),node->getField(FID_COORDINDEX));


    // if joinmode is material, the indices of the list of the second node,
    // are maybee 0, so I must set a new one (new index=(max of first list)+1)
    if (joinmode&JM_MATERIAL)
    {
        Field * f1 = getField(FID_MATERIALINDEX);
        Field * f2 = node->getField(FID_MATERIALINDEX);

        if ((joinmode&JM_MATERIAL) && (!f1 || !f2))
        {
            error("join mode is material - second materiaIndex == NULL\n");
            exit(1);
        }

        if (f1 && f2 && f1->value && f2->value)
        {
            FLong * v1 = (FLong *)(f1->value);
            FLong * v2 = (FLong *)(f2->value);

            if (joinmode&JM_MATERIAL)
            {
                FLong * help = v1;
                int max = 0;
                while (help!=NULL)
                {
                    if (help->number>max) max=help->number;
                    help=(FLong *)(help->next);
                }
                help = v2;
                max++;
                while (help!=NULL)
                {
                    help->number+=max;
                    help=(FLong *)(help->next);
                }
            }
        }
    }
    // append the list of the second node to the list of the first node
    // use this new list as list of the second node
    joinFields(getField(FID_MATERIALINDEX),node->getField(FID_MATERIALINDEX));


    // two NormalNodes will be joined, so the indices of the second list are
    // increment by the count of the first list
    if (joinmode&JM_NORMAL)
    {
        Field * f = node->getField(FID_NORMALINDEX);
        if (f!=NULL)
        {
            FLong * v1 = (FLong *)(f->value);
            while (v1!=NULL)
            {
                if (v1->number!=-1) v1->number+=state.lastNormalNode->pointcount;
                v1=(FLong *)(v1->next);
            }
        }
    }
    // append the list of the second node to the list of the first node
    // use this new list as list of the second node
    joinFields(getField(FID_NORMALINDEX),node->getField(FID_NORMALINDEX));


    // two TextureNodes will be joined, so the indices of the second list are
    // increment by the count of the first list
    if (joinmode&JM_TEXTURE)
    {
        Field * f = node->getField(FID_TEXTURECOORDINDEX);
        if (f!=NULL)
        {
            FLong * v1 = (FLong *)(f->value);
            while (v1!=NULL)
            {
                if (v1->number!=-1) v1->number+=state.lastTextureCoordNode->pointcount;
                v1=(FLong *)(v1->next);
            }
        }
    }
    // append the list of the second node to the list of the first node
    // use this new list as list of the second node
    joinFields(getField(FID_TEXTURECOORDINDEX),node->getField(FID_TEXTURECOORDINDEX));
}

//---ShapeNode::printTree-----------------------------------------------------
// Input : int indent     : indention of output
// Output:

void ShapeNode::printTree(int indent)
{
    Node::printTree(indent);
    printf(" (");

    Field * help;

    help = getField(FID_COORDINDEX);
    if (help)
    {
        Value * v = help->value;
        int count=0;
        while (v!=NULL) { v=v->next; count++; }
        printf(" coordIndex %d",count);
    }
    help = getField(FID_MATERIALINDEX);
    if (help)
    {
        Value * v = help->value;
        int count=0;
        while (v!=NULL) { v=v->next; count++; }
        printf(" materialIndex %d",count);
    }
    help = getField(FID_NORMALINDEX);
    if (help)
    {
        Value * v = help->value;
        int count=0;
        while (v!=NULL) { v=v->next; count++; }
        printf(" normalIndex %d",count);
    }
    help = getField(FID_TEXTURECOORDINDEX);
    if (help)
    {
        Value * v = help->value;
        int count=0;
        while (v!=NULL) { v=v->next; count++; }
        printf(" textureCoordIndex %d",count);
    }
    printf(")");
}

//----------------------------------------------------------------------------
//   class FaceNode : ShapeNode : Node
//----------------------------------------------------------------------------
//used for VRML-node IndexedFaceSet
//It is able to transform the index-lists to a FaceSet and produce lods
//and print out the lods - these actions are performed in the print-method.
//The print-method calls the method generateFaceSet wich reads the index-lists
//and generates a FaceSet. Then the method generateLODs will be called.
//Then the print-method prints out the lods - additionally some nodes are
//created temporary and printed out if required (e.g. MaterialBinding,
//ShapeHints,Separator, ...)

int FaceNode::verbose = 0;  //FaceNode prints out informations
                            //to stdout if !=0

//---FaceNode::FaceNode-------------------------------------------------------
// Input : int aid              : init
//         char *afieldspec     : init
//         Field *afield        : init
//         int fline            : init
//         int lline            : init
// Output:

FaceNode::FaceNode(int aid, char *afieldspec, Field *afield, int fline, int lline) :
          ShapeNode(aid,afieldspec,afield,fline,lline)
{
    faceSet = NULL;
}

//---FaceNode::~FaceNode------------------------------------------------------
// Input :
// Output:

FaceNode::~FaceNode()
{
    deleteFaceSets();
}

//---FaceNode::print----------------------------------------------------------
// Input : FILE *fd       : VRML output file descriptor
//         int indent     : indention of output
// Output:
//The print-method calls the method generateFaceSet wich reads the index-lists
//and generates a FaceSet. Then the method generateLODs will be called.
//Then the print-method prints out the lods - additionally some nodes are
//created temporary and printed out if required (e.g. MaterialBinding,
//ShapeHints,Separator, ...)

void FaceNode::print(FILE *fd, int indent)
{
    FaceSet *fsh;

    generateFaceSet();

    if (faceSet==NULL || faceSet->count==0)
    {
        printHead(fd,indent);
        printIndent(fd,indent);
        fprintf(fd,"%s [ ]\n",fieldIdString[FID_COORDINDEX]);
        printTail(fd,indent);
        warning("empty node\n");
        cleanUpLast();
        deleteFaceSets();
        return;               //<-- return !
    }

    if (nolods==0 && extractlod!=0) generateLODs();

    //---evt. create separator and binding nodes

    SeparatorNode *sep = NULL;
    MaterialBindingNode *mbn = NULL;
    NormalBindingNode *nbn =NULL;
    ShapeHintsNode *shn =NULL;

    if (newMaterialBinding != state.currMaterialBinding)
    {
        if (sep==NULL) sep = new SeparatorNode(NID_SEPARATOR,NULL,NULL,NULL,0,0);
        mbn = new MaterialBindingNode(NID_MATERIALBINDING,
                  NULL, new Field(FID_VALUE,new FEnum(newMaterialBinding)),0,0);
        ((FEnum *)mbn->field->value)->type = FTvalue;
    }

    if (newNormalBinding != state.currNormalBinding)
    {
        if (sep==NULL) sep = new SeparatorNode(NID_SEPARATOR,NULL,NULL,NULL,0,0);
        nbn = new NormalBindingNode(NID_NORMALBINDING,
                  NULL, new Field(FID_VALUE,new FEnum(newNormalBinding)),0,0);
        ((FEnum *)nbn->field->value)->type = FTvalue;
    }

    if (state.currFaceType==VRML_UNKNOWN_FACE_TYPE)
    {
        if (sep==NULL) sep = new SeparatorNode(NID_SEPARATOR,NULL,NULL,NULL,0,0);
        shn = new ShapeHintsNode(NID_SHAPEHINTS,
                  NULL, new Field(FID_FACETYPE,new FEnum(VRML_CONVEX)),0,0);
        ((FEnum *)shn->field->value)->type = FTfaceType;
    }

    //----print heads

    if (nodespec) 
    {
        fprintf(fd,"DEF %s ",nodespec);
        delete nodespec;
        nodespec=NULL;
    }
    if (sep) sep->printHead(fd,indent);
    if (mbn) mbn->print(fd,indent);
    if (nbn) nbn->print(fd,indent);
    if (shn) shn->print(fd,indent);

    if (nolods==0 && extractlod==-1)
    {
        printIndent(fd,indent);
        fprintf(fd,"%s {\n",nodeIdString[NID_LOD]);
        indent+=global_indent;
    }

    //---ranges

    if (nolods==0 && extractlod==-1)
    {
        printIndent(fd,indent);
        fprintf(fd,"%s [",fieldIdString[FID_RANGE]);
        float scale = state.getMaxScale();
        if (verbose) message("face","scale=%f\n",scale);
        fsh = faceSet;
        while (fsh && fsh->next)
        {
            float sphere = fsh->getSphere();
            float range = (fsh->range+sphere)*scale;
            printFloat(fd,range);
            if (verbose) message("face","range of lod %d = %f\n",fsh->lod,range);
            if (fsh->next->next) fprintf(fd,",");
            fsh=fsh->next;
        }
        fprintf(fd,"]\n");
    }

    //---center

    if (nolods==0 && extractlod==-1)
    {
        printIndent(fd,indent);
        fprintf(fd,"%s ",fieldIdString[FID_CENTER]);
        float cx,cy,cz;
        faceSet->getCenter(cx,cy,cz);
        printFloat(fd,cx); fprintf(fd," ");
        printFloat(fd,cy); fprintf(fd," ");
        printFloat(fd,cz); fprintf(fd,"\n");
    }

    //---print IndexedFaceSet-Nodes - one for each LOD

    SeparatorNode lodsep = SeparatorNode(NID_SEPARATOR,NULL,NULL,NULL,0,0);
    fsh = faceSet;
    for (int i=0; i<extractlod; i++) if (fsh) fsh=fsh->next;
    while (fsh)
    {
        if (fsh->polycount>0 && fsh->linecount>0) lodsep.printHead(fd,indent);
        if (fsh->polycount==0 && fsh->linecount==0) lodsep.printHead(fd,indent);

        if (fsh->polycount>0)
        {
            printHead(fd,indent);
            fsh->print(fd,indent,FACE_PRINT_POLY);
            printTail(fd,indent);
        } 

        if (fsh->linecount>0)
        {
            int oldid = id;
            id = NID_INDEXEDLINESET;  
            printHead(fd,indent);
            id = oldid;
            fsh->print(fd,indent,FACE_PRINT_LINE);
            printTail(fd,indent);
        }

        if (fsh->polycount>0 && fsh->linecount>0) lodsep.printTail(fd,indent);
        if (fsh->polycount==0 && fsh->linecount==0) lodsep.printTail(fd,indent);
        fsh=fsh->next;
    }

    if (nolods==0 && extractlod==-1)
    {
        indent-=global_indent;
        printIndent(fd,indent);
        fprintf(fd,"}\n");
    }

    if (sep) sep->printTail(fd,indent);

    //---clean up

    if (sep) delete sep; 
    if (mbn) delete mbn; 
    if (nbn) delete nbn; 
    if (shn) delete shn; 

    cleanUpLast();
    deleteFaceSets();
}

//---FaceNode::generateFaceSet------------------------------------------------
// Input : 
// Output:
//The method creates a FaceSet - reads the index-lists of the fields with
//field-id FID_COORDINDEX, FID_MATERIALINDEX, FID_NORMALINDEX, 
//FID_TEXTURECOORDINDEX creates Faces using the current bindings
//(this a very complicated task because there are sooo many bindings)
//polygons with more than 4 vertices will be triangulated with an object of
//class Triangulator. Before this Faces will be inserted into the FaceSet
//the Faces are stored in temporary lists because faces with more vertices
//must be inserted in the FaceSet before such with less vertices.
//the index-lists will be deleted !

void FaceNode::generateFaceSet()
{
    CoordSet *csh        = state.lastCoordNode->coordSet;
    MaterialSet *msh     = state.lastMaterialNode->materialSet;
    NormalSet *nsh       = state.lastNormalNode->normalSet;
    TextureCoordSet *tsh = state.lastTextureCoordNode->textureCoordSet;

    csh->deleteShapeRefs();

    //---find out which fields are defined    

    Field *coordIndex = getField(FID_COORDINDEX);
    Field *materialIndex = getField(FID_MATERIALINDEX);
    Field *normalIndex = getField(FID_NORMALINDEX);
    Field *textureCoordIndex = getField(FID_TEXTURECOORDINDEX);

    if (verbose) 
    {
        message("face","accesses:");

        if (state.lastCoordNode->firstline>=0)
            printf(" coo(%d-%d)",state.lastCoordNode->firstline,
                                 state.lastCoordNode->lastline);
        else printf(" coo(default)");

        if (state.lastMaterialNode->firstline>=0)
            printf(" mat(%d-%d)",state.lastMaterialNode->firstline,
                                 state.lastMaterialNode->lastline);
        else printf(" mat(default)");

        if (state.lastNormalNode->firstline>=0)
            printf(" nor(%d-%d)",state.lastNormalNode->firstline,
                                 state.lastNormalNode->lastline);
        else printf(" nor(default)");

        if (state.lastTextureCoordNode->firstline>=0)
            printf(" tex(%d-%d)",state.lastTextureCoordNode->firstline,
                                 state.lastTextureCoordNode->lastline);
        else printf(" tex(default)");

        printf("\n");
    }

    FLong *cooV = NULL;
    FLong *matV = NULL;
    FLong *norV = NULL;
    FLong *texV = NULL;
    if (coordIndex) cooV = (FLong *)coordIndex->value;
    if (materialIndex) matV = (FLong *)materialIndex->value;
    if (normalIndex) norV = (FLong *)normalIndex->value;
    if (textureCoordIndex) texV = (FLong *)textureCoordIndex->value;

    //---check for dumb and useless index combinations

    if (materialIndex)
        if (state.currMaterialBinding==VRML_OVERALL ||
            state.currMaterialBinding==VRML_PER_PART ||
            state.currMaterialBinding==VRML_PER_FACE ||
            state.currMaterialBinding==VRML_PER_VERTEX)
                warning("useless materialIndex\n");

    if (normalIndex)
        if (state.currNormalBinding==VRML_OVERALL ||
            state.currNormalBinding==VRML_PER_PART ||
            state.currNormalBinding==VRML_PER_FACE ||
            state.currNormalBinding==VRML_PER_VERTEX)
                warning("useless normalIndex\n");

    //---if the MaterialSet in the current scope contains <= 1 elements ->
    //---change the binding to OVERALL - this causes this node to ingore the 
    //---materialIndex

    if (msh->count<=1) newMaterialBinding = VRML_OVERALL;

    //---create a FaceSet for LOD 0

    faceSet = new FaceSet(0,csh,
                          newMaterialBinding, 
                          newNormalBinding, 
                          state.currVertexOrdering);

    //---create Faces store it temporary in lists and delete index-lists

    int lcount = 0;
    int faceIndex = 0;
    int vertexIndex = 0;

    int i;
    int vertices;
    FLong *vh;
    Vertex *vertex;

    Face *face;
    Face *firstTemp4Face = NULL;
    Face *lastTemp4Face = NULL;
    Face *firstTemp3Face = NULL;
    Face *lastTemp3Face = NULL;
    Face *firstTemp2Face = NULL;
    Face *lastTemp2Face = NULL;

    Triangulator *triang = new Triangulator();

    while (cooV)
    {
        //---determine count of vertices of next polygon

        vh=cooV;
        vertices=0;
        while (vh!=NULL && vh->number!=-1)
        { 
            vertices++;
            vh=(FLong *)(vh->next);
        }
        if (vertices==0) error("polygon with no vertices\n");
         
        //---allocate memory for vertices

        vertex = new Vertex[vertices];

        //---material

        if (msh->count<=1)
        {
            for (i=0; i<vertices; i++) vertex[i].material = NULL;
        }
        else switch (state.currMaterialBinding)
        {
            case VRML_OVERALL:
                for (i=0; i<vertices; i++) vertex[i].material = NULL;
                break;
                
            case VRML_PER_PART:
            case VRML_PER_FACE:
                for (i=0; i<vertices; i++)
                    vertex[i].material = getMaterialIndex(msh,faceIndex);
                break;  

            case VRML_PER_VERTEX:
                for (i=0; i<vertices; i++)
                    vertex[i].material = getMaterialIndex(msh,vertexIndex+i); 
                break;


            case VRML_PER_PART_INDEXED:
            case VRML_PER_FACE_INDEXED:
                for (i=0; i<vertices; i++)
                    if (matV) vertex[i].material = getMaterialIndex(msh,matV->number);
                        else error("incomplete materialIndex\n");
                if (matV) deleteValues((Value **)(&matV),1);
                break;

            case VRML_PER_VERTEX_INDEXED:
                if (materialIndex==NULL)      //!!! undocumented feature ???
                {
                    FLong *cooH = cooV;
                    for (i=0; i<vertices; i++)
                    {
                        vertex[i].material = getMaterialIndex(msh,cooH->number);
                        cooH=(FLong *)cooH->next;
                    }
                }
                else
                {
                    for (i=0; i<vertices; i++)
                        if (matV)
                        {
                            vertex[i].material = getMaterialIndex(msh,matV->number);
                            deleteValues((Value **)(&matV),1);
                        }
                        else error("incomplete materialIndex\n");
                    if (matV && matV->number==-1) deleteValues((Value **)(&matV),1);
                }
                break;

            default: assert(0);
        }

        //---normal

        if (nsh->count==0)
        {
            for (i=0; i<vertices; i++) vertex[i].normal = NULL;
        }
        else switch (state.currNormalBinding)
        {
            case VRML_OVERALL:
                for (i=0; i<vertices; i++) vertex[i].normal = NULL;
                break;

            case VRML_PER_PART:
            case VRML_PER_FACE:
                for (i=0; i<vertices; i++)
                    if (faceIndex<nsh->getMaxIndex())
                        vertex[i].normal = getNormalIndex(nsh,faceIndex);
                    else error("not enough normals defined\n");
                break;  

            case VRML_PER_VERTEX:
                for (i=0; i<vertices; i++)
                    if (vertexIndex+i<nsh->getMaxIndex())
                        vertex[i].normal = getNormalIndex(nsh,vertexIndex+i);
                    else error("not enough normals defined\n");
                break;

            case VRML_PER_PART_INDEXED:
            case VRML_PER_FACE_INDEXED:
                for (i=0; i<vertices; i++)
                    if (norV) vertex[i].normal = getNormalIndex(nsh,norV->number);
                        else error("incomplete normalIndex\n");
                if (norV) deleteValues((Value **)(&norV),1);
                break;

            case VRML_PER_VERTEX_INDEXED:
                if (normalIndex==NULL)      //!!! undocumented feature ???
                {
                    FLong *cooH = cooV;
                    for (i=0; i<vertices; i++)
                    {
                        vertex[i].normal = getNormalIndex(nsh,cooH->number);
                        cooH=(FLong *)cooH->next;
                    }
                }
                else
                {
                    for (i=0; i<vertices; i++)
                        if (norV)
                        {
                            vertex[i].normal = getNormalIndex(nsh,norV->number);
                            deleteValues((Value **)(&norV),1);
                        }
                        else error("incomplete normalIndex\n");
                    if (norV && norV->number==-1) deleteValues((Value **)(&norV),1);
                }
                break;

            default: assert(0);
        }

        //---textureCoord

        if (textureCoordIndex)
        {
            if (tsh->count==0)
            {
                for (i=0; i<vertices; i++) vertex[i].textureCoord = NULL;
            } 
            else
            {
                for (i=0; i<vertices; i++) 
                    if (texV) 
                    {
                        vertex[i].textureCoord = getTextureCoordIndex(tsh,texV->number);
                        deleteValues((Value **)(&texV),1);
                    }
                    else error("incomplete textureCoordIndex\n");
                if (texV && texV->number==-1) deleteValues((Value **)(&texV),1);
            }
        }

        //---coord

        for (i=0; i<vertices; i++)
        {
            vertex[i].coord = getCoordIndex(csh,cooV->number);
            deleteValues((Value **)(&cooV),1);
        }
        if (cooV && cooV->number==-1) deleteValues((Value **)(&cooV),1);

        //---create new face and insert it in the temporary list

        Vertex nullVertex;
        Coord *coord;

        if (vertices==2) 
        {
            face = new Face(vertex[0],vertex[1],nullVertex,nullVertex);
            if (firstTemp2Face)
            {
                lastTemp2Face->next=face;
                lastTemp2Face=face;
            }
            else firstTemp2Face=lastTemp2Face=face;
        }

        else if (vertices==3) 
        {
            face = new Face(vertex[0],vertex[1],vertex[2],nullVertex);
            if (firstTemp3Face)
            {
                lastTemp3Face->next=face;
                lastTemp3Face=face;
            }
            else firstTemp3Face=lastTemp3Face=face;
        }

        else if (vertices==4)
        {
            face = new Face(vertex[0],vertex[1],vertex[2],vertex[3]);
            if (firstTemp4Face)
            {
                lastTemp4Face->next=face;
                lastTemp4Face=face;
            }
            else firstTemp4Face=lastTemp4Face=face;
        }

        else if (vertices>4)
        {
            triang->reset();

            for (i=0; i<vertices; i++) 
            {
                coord=vertex[i].coord;                 
                triang->addPoint(coord->x,coord->y,coord->z);
            }

            if (triang->run())
            {
                int i1,i2,i3;
                int t;
                for (t=0; t<triang->getCount(); t++)
                { 
                    i1=triang->getPoint(t,0);
                    i2=triang->getPoint(t,1);
                    i3=triang->getPoint(t,2);
                    face = new Face(vertex[i1],vertex[i2],vertex[i3],nullVertex);
                    if (firstTemp3Face)
                    {
                        lastTemp3Face->next=face;
                        lastTemp3Face=face;
                    }
                    else firstTemp3Face=lastTemp3Face=face;
                }
            }
        }

        //---prepare for the next iteration

        delete vertex;
        lcount++;
        faceIndex++;
        vertexIndex+=vertices;

    } //end while (main-loop)

    //First all quadrangles will be inserted in the FaceSet
    //then all triangles and finally all edges. This ensures that
    //no polygon which is already inserted must be removed from the set

    Face *faceh;

    face=firstTemp4Face;
    while (face)
    {
        faceh=(Face *)face->next;
        face->next=NULL;
        faceSet->insert(&face);
        face=faceh;
    }

    face=firstTemp3Face;
    while (face)
    {
        faceh=(Face *)face->next;
        face->next=NULL;
        faceSet->insert(&face);
        face=faceh;
    }

    face=firstTemp2Face;
    while (face)
    {
        faceh=(Face *)face->next;
        face->next=NULL;
        faceSet->insert(&face);
        face=faceh;
    }

    //---clean up

    if (verbose) message("face","list=%d set=%d (quad=%d tri=%d line=%d)\n",
                         lcount,faceSet->count,faceSet->quadcount,
                         faceSet->tricount,faceSet->linecount);

    if (cooV) deleteAllValues((Value **)(&cooV));
    if (matV) deleteAllValues((Value **)(&matV));
    if (norV) deleteAllValues((Value **)(&norV));
    if (texV) deleteAllValues((Value **)(&texV));
    if (coordIndex) coordIndex->value=NULL;
    if (materialIndex) materialIndex->value=NULL;
    if (normalIndex) normalIndex->value=NULL;
    if (textureCoordIndex) textureCoordIndex->value=NULL;
    delete triang;
}

//---FaceNode::generateLODs---------------------------------------------------
// Input :
// Output:
// generate the lods for this FaceNode !
// the lods are stored in separate FaceSets building a linear list
// the lods are generated by alternating calling generateNextLOD of the 
// CoordSet and the current FaceSet:
// 
//csh->generateNextLOD generate coord-lod 1 using the shapeRefs of
//                     coord-lod 0 to face-lod 0
//fsh->generateNextLOD generate face-lod 1 - setup shapeRefs of coord-lod 1
//                     to face-lod 1
//fsh=fsh->next   fsh is now face-lod 1
//
//next iteration analog (with lods 1 and 2) and so on

void FaceNode::generateLODs()
{
    FaceSet *fsh = faceSet;
    CoordSet *csh = state.lastCoordNode->coordSet;

    csh->resetLODs();                     //set CoordSet to lod 0
    csh->setShapeType(SHAPE_TYPE_FACE);   //the CoordSet is used by a FaceSet
    int newcount = csh->getCountUsed();

    if (verbose) message("face","uses %d of %d coord(s)\n", newcount,csh->count); 
        else message("","lod - 0 ");

    while ((newcount>1 && fsh->count>0) && 
           (extractlod==-1 || csh->getCurrentLOD()<extractlod))
    {
        csh->generateNextLOD();
        fsh->generateNextLOD();
        fsh->range = csh->getRange();          //set the range of the old face-lod
                                               //to getRange of the new coord-lod
        fsh=fsh->next;
        newcount=csh->getCountUsed();

        if (verbose) message("face","lod %d: set=%d (quad=%d tri=%d line=%d)\n",
                         csh->getCurrentLOD(),fsh->count,
                         fsh->quadcount,fsh->tricount,fsh->linecount);
        else { printf("%d ",csh->getCurrentLOD()); fflush(stdout); }

    }
    if (!verbose) printf("\n");
}

//---FaceNode::deleteFaceSets-------------------------------------------------
// Input :
// Output:
//delete all FaceSets (==lods) of this node

void FaceNode::deleteFaceSets()
{
    FaceSet *fsh;
    while (faceSet)
    {
        fsh=faceSet;
        faceSet=faceSet->next;
        delete fsh;
    }
}

//----------------------------------------------------------------------------
//   class LineNode : ShapeNode : Node
//----------------------------------------------------------------------------
//used for VRML-node IndexedLineSet
//It is able to transform the index-lists to a LineSet and produce lods
//and print out the lods - these actions are performed in the print-method.
//The print-method calls the method generateLineSet wich reads the index-lists
//and generates a LineSet. Then the method generateLODs will be called.
//Then the print-method prints out the lods - additionally some nodes are
//created temporary and printed out if required (e.g. MaterialBinding,
//ShapeHints,Separator, ...)

int LineNode::verbose = 0;  //LineNode prints out informations
                            //to stdout if !=0

//---LineNode::LineNode-------------------------------------------------------
// Input : int aid              : init
//         char *afieldspec     : init
//         Field *afield        : init
//         int fline            : init
//         int lline            : init
// Output:

LineNode::LineNode(int aid, char *afieldspec, Field *afield, int fline, int lline) :
          ShapeNode(aid,afieldspec,afield,fline,lline)
{
    lineSet = NULL;
}

//---LineNode::~LineNode------------------------------------------------------
// Input :
// Output:

LineNode::~LineNode()
{
    deleteLineSets();
}

//---LineNode::print----------------------------------------------------------
// Input : FILE *fd         : VRML output file descriptor
//         int indent       : indention of output
// Output:
//The print-method calls the method generateLineSet wich reads the index-lists
//and generates a LineSet. Then the method generateLODs will be called.
//Then the print-method prints out the lods - additionally some nodes are
//created temporary and printed out if required (e.g. MaterialBinding,
//ShapeHints,Separator, ...)

void LineNode::print(FILE *fd, int indent)
{
    LineSet *lsh;

    generateLineSet();

    if (lineSet==NULL || lineSet->count==0)
    {
        printHead(fd,indent);
        printIndent(fd,indent);
        fprintf(fd,"%s [ ]\n",fieldIdString[FID_COORDINDEX]);
        printTail(fd,indent);
        warning("empty node\n");
        cleanUpLast();
        deleteLineSets();
        return;                   //<-- return !
    }

    if (nolods==0 && extractlod!=0) generateLODs();

    //---evt. create separator and binding nodes

    SeparatorNode *sep = NULL;
    MaterialBindingNode *mbn = NULL;
    NormalBindingNode *nbn =NULL;

    if (newMaterialBinding != state.currMaterialBinding)
    {
        if (sep==NULL) sep = new SeparatorNode(NID_SEPARATOR,NULL,NULL,NULL,0,0);
        mbn = new MaterialBindingNode(NID_MATERIALBINDING,
                  NULL, new Field(FID_VALUE,new FEnum(newMaterialBinding)),0,0);
        ((FEnum *)mbn->field->value)->type = FTvalue;
    }

    if (newNormalBinding != state.currNormalBinding)
    {
        if (sep==NULL) sep = new SeparatorNode(NID_SEPARATOR,NULL,NULL,NULL,0,0);
        nbn = new NormalBindingNode(NID_NORMALBINDING,
                  NULL, new Field(FID_VALUE,new FEnum(newNormalBinding)),0,0);
        ((FEnum *)nbn->field->value)->type = FTvalue;
    }

    //----print heads

    if (nodespec) 
    {
        fprintf(fd,"DEF %s ",nodespec);
        delete nodespec;
        nodespec=NULL;
    }
    if (sep) sep->printHead(fd,indent);
    if (mbn) mbn->print(fd,indent);
    if (nbn) nbn->print(fd,indent);

    if (nolods==0 && extractlod==-1)
    {
        printIndent(fd,indent);
        fprintf(fd,"%s {\n",nodeIdString[NID_LOD]);
        indent+=global_indent;
    }

    //---ranges

    if (nolods==0 && extractlod==-1)
    {
        printIndent(fd,indent);
        fprintf(fd,"%s [",fieldIdString[FID_RANGE]);
        float scale = state.getMaxScale();
        if (verbose) message("line","scale=%f\n",scale);
        lsh = lineSet;
        while (lsh && lsh->next)
        {
            float sphere = lsh->getSphere();
            float range = (lsh->range+sphere)*scale;
            printFloat(fd,range);
            if (verbose) message("line","range of lod %d = %f\n",lsh->lod,range);
            if (lsh->next->next) fprintf(fd,",");
            lsh=lsh->next;
        }
        fprintf(fd,"]\n");
    }

    //---center

    if (nolods==0 && extractlod==-1)
    {
        printIndent(fd,indent);
        fprintf(fd,"%s ",fieldIdString[FID_CENTER]);
        float cx,cy,cz;
        lineSet->getCenter(cx,cy,cz);
        printFloat(fd,cx); fprintf(fd," ");
        printFloat(fd,cy); fprintf(fd," ");
        printFloat(fd,cz); fprintf(fd,"\n");
    }
 
    //---print IndexedLineSet-Nodes - one for each LOD

    lsh = lineSet;
    for (int i=0; i<extractlod; i++) if (lsh) lsh=lsh->next;
    while (lsh)
    {
        printHead(fd,indent);
        lsh->print(fd,indent);
        printTail(fd,indent);
        lsh=lsh->next;
    }

    //---print tails

    if (nolods==0 && extractlod==-1)
    {
        indent-=global_indent;
        printIndent(fd,indent);
        fprintf(fd,"}\n");
    }

    if (sep) sep->printTail(fd,indent);

    //---clean up

    if (sep) delete sep; 
    if (mbn) delete mbn; 
    if (nbn) delete nbn; 

    cleanUpLast();
    deleteLineSets();
}

//---LineNode::generateLineSet------------------------------------------------
// Input :
// Output:
//The method creates a LineSet - reads the index-lists of the fields with
//field-id FID_COORDINDEX, FID_MATERIALINDEX, FID_NORMALINDEX, 
//FID_TEXTURECOORDINDEX creates Lines using the current bindings
//(this a very complicated task because there are sooo many bindings)
//Polylines will be split into single lines. This Lines will be inserted
//into the LineSet - the index-lists will be deleted !

void LineNode::generateLineSet()
{
    CoordSet *csh        = state.lastCoordNode->coordSet;
    MaterialSet *msh     = state.lastMaterialNode->materialSet;
    NormalSet *nsh       = state.lastNormalNode->normalSet;
    TextureCoordSet *tsh = state.lastTextureCoordNode->textureCoordSet;

    csh->deleteShapeRefs();

    //---find out which fields are defined    

    Field *coordIndex = getField(FID_COORDINDEX);
    Field *materialIndex = getField(FID_MATERIALINDEX);
    Field *normalIndex = getField(FID_NORMALINDEX);
    Field *textureCoordIndex = getField(FID_TEXTURECOORDINDEX);

    if (verbose) 
    {
        message("line","accesses:");

        if (state.lastCoordNode->firstline>=0)
            printf(" coo(%d-%d)",state.lastCoordNode->firstline,
                                 state.lastCoordNode->lastline);
        else printf(" coo(default)");

        if (state.lastMaterialNode->firstline>=0)
            printf(" mat(%d-%d)",state.lastMaterialNode->firstline,
                                 state.lastMaterialNode->lastline);
        else printf(" mat(default)");

        if (state.lastNormalNode->firstline>=0)
            printf(" nor(%d-%d)",state.lastNormalNode->firstline,
                                 state.lastNormalNode->lastline);
        else printf(" nor(default)");

        if (state.lastTextureCoordNode->firstline>=0)
            printf(" tex(%d-%d)",state.lastTextureCoordNode->firstline,
                                 state.lastTextureCoordNode->lastline);
        else printf(" tex(default)");

        printf("\n");
    }

    FLong *cooV = NULL;
    FLong *matV = NULL;
    FLong *norV = NULL;
    FLong *texV = NULL;
    if (coordIndex) cooV = (FLong *)coordIndex->value; 
    if (materialIndex) matV = (FLong *)materialIndex->value;
    if (normalIndex) norV = (FLong *)normalIndex->value;
    if (textureCoordIndex) texV = (FLong *)textureCoordIndex->value;

    //---check for dumb and useless index combinations

    if (materialIndex)
        if (state.currMaterialBinding==VRML_OVERALL ||
            state.currMaterialBinding==VRML_PER_PART ||
            state.currMaterialBinding==VRML_PER_FACE ||
            state.currMaterialBinding==VRML_PER_VERTEX)
                warning("useless materialIndex\n");

    if (normalIndex)
        if (state.currNormalBinding==VRML_OVERALL ||
            state.currNormalBinding==VRML_PER_PART ||
            state.currNormalBinding==VRML_PER_FACE ||
            state.currNormalBinding==VRML_PER_VERTEX)
                warning("useless normalIndex\n");

    //---if the MaterialSet in the current scope contains <= 1 elements ->
    //---change the binding to OVERALL - this causes this node to ingore the 
    //---materialIndex

    if (msh->count<=1) newMaterialBinding = VRML_OVERALL;

    //---create a LineSet for LOD 0

    lineSet = new LineSet(0,csh,
                          newMaterialBinding, 
                          newNormalBinding);

    //---create Lines insert it in the LineSet and delete index-lists

    int lcount = 0;
    int faceIndex = 0;
    int partIndex = 0;
    int vertexIndex = 0;

    int i;
    int vertices,nv;
    FLong *vh;
    Vertex *vertex;
    Line *line;

    while (cooV)
    {
        //---determine count of vertices of next polyline

        vh=cooV;
        vertices=0;
        while (vh!=NULL && vh->number!=-1)
        { 
            vertices++;
            vh=(FLong *)(vh->next);
        }
        if (vertices==0) error("polyline with no vertices\n");
         
        //---allocate memory for vertices

        vertex = new Vertex[vertices];

        //---material

        if (msh->count<=1)
        {
            for (i=0; i<vertices; i++) vertex[i].material = NULL;
        }
        else switch (state.currMaterialBinding)
        {
            case VRML_OVERALL:
                for (i=0; i<vertices; i++) vertex[i].material = NULL;
                break;
                
            case VRML_PER_FACE:
                for (i=0; i<vertices; i++)
                    vertex[i].material = getMaterialIndex(msh,faceIndex);
                break;  

            case VRML_PER_PART:
                for (i=0; i<vertices-1; i++)
                    vertex[i].material = getMaterialIndex(msh,partIndex+i);                
                break;

            case VRML_PER_VERTEX:
                for (i=0; i<vertices; i++)
                    vertex[i].material = getMaterialIndex(msh,vertexIndex+i); 
                break;

            case VRML_PER_FACE_INDEXED:
                for (i=0; i<vertices; i++)
                    if (matV) vertex[i].material = getMaterialIndex(msh,matV->number);
                        else error("incomplete materialIndex\n");
                if (matV) deleteValues((Value **)(&matV),1);
                break;

            case VRML_PER_PART_INDEXED:
            case VRML_PER_VERTEX_INDEXED:
                nv=vertices;
                if (state.currMaterialBinding==VRML_PER_PART_INDEXED) nv--;

                if (materialIndex==NULL)      //!!! undocumented feature ???
                {
                    FLong *cooH = cooV;
                    for (i=0; i<nv; i++)
                    {
                        vertex[i].material = getMaterialIndex(msh,cooH->number);
                        cooH=(FLong *)cooH->next;
                    }
                }
                else
                {
                    for (i=0; i<nv; i++)
                        if (matV)
                        {
                            vertex[i].material = getMaterialIndex(msh,matV->number);
                            deleteValues((Value **)(&matV),1);
                        }
                        else error("incomplete materialIndex\n");
                    if (matV && matV->number==-1) deleteValues((Value **)(&matV),1);
                }
                break;

            default: assert(0);
        }

        //---normal

        if (nsh->count==0)
        {
            for (i=0; i<vertices; i++) vertex[i].normal = NULL;
        }
        else switch (state.currNormalBinding)
        {
            case VRML_OVERALL:
                for (i=0; i<vertices; i++) vertex[i].normal = NULL;
                break;

            case VRML_PER_FACE:
                for (i=0; i<vertices; i++)
                    if (faceIndex<nsh->getMaxIndex())
                        vertex[i].normal = getNormalIndex(nsh,faceIndex);
                    else error("not enough normals defined\n");
                break;  

            case VRML_PER_PART:
                for (i=0; i<vertices-1; i++)
                    if (faceIndex<nsh->getMaxIndex())
                        vertex[i].normal = getNormalIndex(nsh,partIndex);
                    else error("not enough normals defined\n");
                break;  

            case VRML_PER_VERTEX:
                for (i=0; i<vertices; i++)
                    if (vertexIndex+i<nsh->getMaxIndex())
                        vertex[i].normal = getNormalIndex(nsh,vertexIndex+i);
                    else error("not enough normals defined\n");
                break;

            case VRML_PER_FACE_INDEXED:
                for (i=0; i<vertices; i++)
                    if (norV) vertex[i].normal = getNormalIndex(nsh,norV->number);
                        else error("incomplete normalIndex\n");
                if (norV) deleteValues((Value **)(&norV),1);
                break;

            case VRML_PER_PART_INDEXED:
            case VRML_PER_VERTEX_INDEXED:
                nv=vertices;
                if (state.currNormalBinding==VRML_PER_PART_INDEXED) nv--;

                if (normalIndex==NULL)      //!!! undocumented feature ???
                {
                    FLong *cooH = cooV;
                    for (i=0; i<nv; i++)
                    {
                        vertex[i].normal = getNormalIndex(nsh,cooH->number);
                        cooH=(FLong *)cooH->next;
                    }
                }
                else
                {
                    for (i=0; i<nv; i++)
                        if (norV)
                        {
                            vertex[i].normal = getNormalIndex(nsh,norV->number);
                            deleteValues((Value **)(&norV),1);
                        }
                        else error("incomplete normalIndex\n");
                    if (norV && norV->number==-1) deleteValues((Value **)(&norV),1);
                }
                break;

            default: assert(0);
        }

        //---textureCoord

        if (textureCoordIndex)
        {
            if (tsh->count==0)
            {
                for (i=0; i<vertices; i++) vertex[i].textureCoord = NULL;
            } 
            else
            {
                for (i=0; i<vertices; i++) 
                    if (texV) 
                    {
                        vertex[i].textureCoord = getTextureCoordIndex(tsh,texV->number);
                        deleteValues((Value **)(&texV),1);
                    }
                    else error ("incomplete textureCoordIndex\n");
                if (texV && texV->number==-1) deleteValues((Value **)(&texV),1);
            }
        }

        //---coord

        for (i=0; i<vertices; i++)
        {
            vertex[i].coord = getCoordIndex(csh,cooV->number);
            deleteValues((Value **)(&cooV),1);
        }
        if (cooV && cooV->number==-1) deleteValues((Value **)(&cooV),1);

        //---create new line and insert it in the lineSet

        int p;
        for (p=0; p<vertices-1; p++) 
        {
            line = new Line(vertex[p],vertex[p+1]);
            lineSet->insert(&line);
            lcount++;
        }

        //---prepare for the next iteration

        delete vertex;
        faceIndex++;
        partIndex+=(vertices-1);
        vertexIndex+=vertices;

    } //end while (main-loop)

    //---clean up

    if (verbose) message("line","list=%d set=%d\n", lcount,lineSet->count);

    if (cooV) deleteAllValues((Value **)(&cooV));
    if (matV) deleteAllValues((Value **)(&matV));
    if (norV) deleteAllValues((Value **)(&norV));
    if (texV) deleteAllValues((Value **)(&texV));
    if (coordIndex) coordIndex->value=NULL;
    if (materialIndex) materialIndex->value=NULL;
    if (normalIndex) normalIndex->value=NULL;
    if (textureCoordIndex) textureCoordIndex->value=NULL;
}

//---LineNode::generateLODs---------------------------------------------------
// Input :
// Output:
// generate the lods for this LineNode !
// the lods are stored in separate LineSets building a linear list
// the lods are generated by alternating calling generateNextLOD of the 
// CoordSet and the current LineSet:
// 
//csh->generateNextLOD generate coord-lod 1 using the shapeRefs of
//                     coord-lod 0 to line-lod 0
//lsh->generateNextLOD generate line-lod 1 - setup shapeRefs of coord-lod 1
//                     to line-lod 1
//lsh=lsh->next   lsh is now line-lod 1
//
//next iteration analog (with lods 1 and 2) and so on


void LineNode::generateLODs()
{
    LineSet *lsh = lineSet;
    CoordSet *csh = state.lastCoordNode->coordSet;

    csh->resetLODs();                    //set the CoordSet to lod 0
    csh->setShapeType(SHAPE_TYPE_LINE);  //the CoordSet is used by a LineSet
    int newcount = csh->getCountUsed();

    if (verbose) message("line","uses %d of %d coord(s)\n", newcount,csh->count); 
        else message("","lod - ");

    while ((newcount>1 && lsh->count>0) && 
           (extractlod==-1 || csh->getCurrentLOD()<extractlod))
    {
        csh->generateNextLOD();
        lsh->generateNextLOD();
        lsh->range = csh->getRange();          //set the range of the old line-lod
                                               //to getRange of the new coord-lod
        lsh=lsh->next;
        newcount=csh->getCountUsed();

        if (verbose) message("line","lod %d: set=%d\n",
                         csh->getCurrentLOD(),lsh->count);
            else { printf("%d ",csh->getCurrentLOD()); fflush(stdout); }
    }
    if (!verbose) printf("\n");
}

//---LineNode::deleteLineSets-------------------------------------------------
// Input :
// Output:
//delete all LineSets (==lods) of this node

void LineNode::deleteLineSets()
{
    LineSet *lsh;
    while (lineSet)
    {
        lsh=lineSet;
        lineSet=lineSet->next;
        delete lsh;
    }
}

//----------------------------------------------------------------------------
//   class PointNode : Node
//----------------------------------------------------------------------------
//used for VRML-node PointSet - overrides setupDepends
//this method evt. increases the variable countUncompressed of the DataSets
//in the nodes CoordNode, MaterialNode, NormalNode, TextureCoordNode in
//the current scope

//---PointNode::PointNode-----------------------------------------------------
// Input : int aid              : init
//         char *afieldspec     : init
//         Field *afield        : init
//         int fline            : init
//         int lline            : init
// Output:

PointNode::PointNode(int aid, char *afieldspec, Field *afield, int fline, int lline) :
           Node(aid,afieldspec,afield,fline,lline)
{
}

//---PointNode::setupDepends--------------------------------------------------
// Input : State *astate : current state
// Output:
//this method evt. increases the variable countUncompressed of the DataSets
//in the nodes CoordNode, MaterialNode, NormalNode, TextureCoordNode in
//the current scope - this is necessary because these DataSets removes
//doublettes what can have strange effects if an PointSet uses this Datas
//The variable countUncompressed prevents the DataSet from removing doublettes
//within the first countUncompressed elements
//e.g. 
//PointSet {
//    startIndex 50
//    numPoints 50
//}
//-> CoordSet::countUncompressed = max(CoordSet::countUncompressed,100)
//MaterialSet,NormalSet,TextureCoordSet analog if binding != OVERALL
//
//additionally the method sets the variable lastOtherUser of the nodes CoordNode,
//MaterialNode, NormalNode, TextureCoordNode in the current scope to this

void PointNode::setupDepends(State *astate)
{
    Field *startIndex = getField(FID_STARTINDEX);
    Field *numPoints = getField(FID_NUMPOINTS);

    int si = 0;
    if (startIndex) si = ((FLong *)startIndex->value)->number;

    int np = -1;
    if (numPoints) np = ((FLong *)numPoints->value)->number;

    if (np==0 || np<-1) return;

    int unc;
    if (np==-1) unc=INT_MAX ; else unc = si+np;  //<- si+np-1 +1


    if(astate->lastCoordNode->coordSet->countUncompressed<unc)
        astate->lastCoordNode->coordSet->countUncompressed=unc;

    if (astate->currMaterialBinding != VRML_OVERALL &&
        astate->currMaterialBinding != VRML_DEFAULT)
    {
        if(astate->lastMaterialNode->materialSet->countUncompressed<unc)
            astate->lastMaterialNode->materialSet->countUncompressed=unc;
    }

    if (astate->currNormalBinding != VRML_OVERALL)
    {
        if(astate->lastNormalNode->normalSet->countUncompressed<unc)
            astate->lastNormalNode->normalSet->countUncompressed=unc;
    }

    if(astate->lastTextureCoordNode->textureCoordSet->countUncompressed<unc)
        astate->lastTextureCoordNode->textureCoordSet->countUncompressed=unc;


    astate->lastCoordNode->lastOtherUser = this;
    astate->lastMaterialNode->lastOtherUser = this;
    astate->lastNormalNode->lastOtherUser = this;
    astate->lastTextureCoordNode->lastOtherUser = this;
}

//----------------------------------------------------------------------------
//   class ConeNode : Node
//----------------------------------------------------------------------------
//used for VRML-node Cone - overrides setupDepends 

//---ConeNode::ConeNode-------------------------------------------------------
// Input : int aid              : init
//         char *afieldspec     : init
//         Field *afield        : init
//         int fline            : init
//         int lline            : init
// Output:

ConeNode::ConeNode(int aid, char *afieldspec, Field *afield, int fline, int lline) :
          Node(aid,afieldspec,afield,fline,lline)
{
}

//---ConeNode::setupDepends---------------------------------------------------
// Input : State *astate        : current state
// Output:
// if the current material binding is PER_PART or PER_PART_INDEXED the
// method sets the variable MaterialSet::countUncompressed of the MaterialNode
// in the current scope to the maximum of 2 and the former value - this prevents
// the MaterialSet from removing doublettes within the first 2 values
// additionally the method sets the variable lastOtherUser of the
// MaterialNode in the current scope to this

void ConeNode::setupDepends(State *astate)
{
    if (astate->currMaterialBinding == VRML_PER_PART || 
        astate->currMaterialBinding == VRML_PER_PART_INDEXED)
    {
        if(astate->lastMaterialNode->materialSet->countUncompressed<2)
            astate->lastMaterialNode->materialSet->countUncompressed=2;
    }

    astate->lastMaterialNode->lastOtherUser = this;
}

//----------------------------------------------------------------------------
//   class CubeNode : Node
//----------------------------------------------------------------------------
//used for VRML-node Cube - overrides setupDepends 

//---CubeNode::CubeNode-------------------------------------------------------
// Input : int aid              : init
//         char *afieldspec     : init
//         Field *afield        : init
//         int fline            : init
//         int lline            : init
// Output:

CubeNode::CubeNode(int aid, char *afieldspec, Field *afield, int fline, int lline) :
          Node(aid,afieldspec,afield,fline,lline)
{
}

//---CubeNode::setupDepends---------------------------------------------------
// Input : State *astate        : current state
// Output:
// if the current material binding is PER_PART(_INDEXED), PER_FACE(_INDEXED) the
// method sets the variable MaterialSet::countUncompressed of the MaterialNode
// in the current scope to the maximum of 6 and the former value - this prevents
// the MaterialSet from removing doublettes within the first 6 values
// additionally the method sets the variable lastOtherUser of the
// MaterialNode in the current scope to this

void CubeNode::setupDepends(State *astate)
{
    if (astate->currMaterialBinding == VRML_PER_PART ||
        astate->currMaterialBinding == VRML_PER_PART_INDEXED ||
        astate->currMaterialBinding == VRML_PER_FACE ||
        astate->currMaterialBinding == VRML_PER_FACE_INDEXED)
    {
        if(astate->lastMaterialNode->materialSet->countUncompressed<6)
            astate->lastMaterialNode->materialSet->countUncompressed=6;
    }

    astate->lastMaterialNode->lastOtherUser = this;
}

//----------------------------------------------------------------------------
//   class CylinderNode : Node
//----------------------------------------------------------------------------
//used for VRML-node Cylinder - overrides setupDepends 

//---CylinderNode::CylinderNode-----------------------------------------------
// Input : int aid              : init
//         char *afieldspec     : init
//         Field *afield        : init
//         int fline            : init
//         int lline            : init
// Input :
// Output:

CylinderNode::CylinderNode(int aid, char *afieldspec, Field *afield,
                           int fline, int lline) :
              Node(aid,afieldspec,afield,fline,lline)
{
}

//---CylinderNode::setupDepends-----------------------------------------------
// Input : State *astate        : current state
// Output:
// if the current material binding is PER_PART or PER_PART_INDEXED the
// method sets the variable MaterialSet::countUncompressed of the MaterialNode
// in the current scope to the maximum of 3 and the former value - this prevents
// the MaterialSet from removing doublettes within the first 3 values
// additionally the method sets the variable lastOtherUser of the
// MaterialNode in the current scope to this

void CylinderNode::setupDepends(State *astate)
{
    if (astate->currMaterialBinding == VRML_PER_PART ||
        astate->currMaterialBinding == VRML_PER_PART_INDEXED)
    {
        if(astate->lastMaterialNode->materialSet->countUncompressed<3)
            astate->lastMaterialNode->materialSet->countUncompressed=3;
    }

    astate->lastMaterialNode->lastOtherUser = this;
}

//----------------------------------------------------------------------------
//   class SphereNode : Node
//----------------------------------------------------------------------------
//used for VRML-node Sphere - overrides setupDepends 

//---SphereNode::SphereNode---------------------------------------------------
// Input : int aid              : init
//         char *afieldspec     : init
//         Field *afield        : init
//         int fline            : init
//         int lline            : init
// Input :
// Output:

SphereNode::SphereNode(int aid, char *afieldspec, Field *afield,
                       int fline, int lline) :
            Node(aid,afieldspec,afield,fline,lline)
{
}

//---SphereNode::setupDepends-------------------------------------------------
// Input : State *astate        : current state
// Output:
// the method sets the variable lastOtherUser of the
// MaterialNode in the current scope to this

void SphereNode::setupDepends(State *astate)
{
    astate->lastMaterialNode->lastOtherUser = this;
}

//----------------------------------------------------------------------------

