// ===========================================================================
//  Copyright 1996 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       tnodes
//  TYPE:       c++ code
//  PROJECT:    lodestar
//  CONTENT:    implementation of class RotationNode, class ScaleNode,
//              class TransformNode, class MatrixTransformNode,
//              class TranslationNode
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rs      Reinhard Sainitzer
//		hb      Herbert Buchegger
// ===========================================================================
//  HISTORY:
//
// ===========================================================================

#include "nodes.h"
#include "global.h"
#include "enum.h"
#include "tri.h"
#include <limits.h>
#include <assert.h>

//----------------------------------------------------------------------------
//   class RotationNode : Node
//----------------------------------------------------------------------------
//used for VRML-node Rotation - overrides setupDepends

//---RotationNode::RotationNode-----------------------------------------------
// Input : int aid                  : init
//         char *afieldspec         : init
//         Field *afield            : init
//         int fline                : init
//         int lline                : init
// Output:

RotationNode::RotationNode(int aid, char *afieldspec, Field *afield,
                           int fline, int lline) :
              Node(aid,afieldspec,afield,fline,lline)
{
}

//---RotationNode::setupDepends-----------------------------------------------
// Input : State *astate  : current state
// Output: State *astate  : new state
// The Field with field-id FID_ROTATION defines a 3D rotation about an
// arbitrary axis through the origin. The rotation is accumulated into
// the current transformation: 
// astate->currTransMatrix = rotation-matrix*astate->currTransMatrix

void RotationNode::setupDepends(State *astate)
{
    Field *rotation = getField(FID_ROTATION);
    if (rotation)
    {
        FRotation *r = (FRotation *)rotation->value;
        if (r->angle!=0)
        {
            Vec a = Vec(r->x, r->y, r->z);
            float l = a.getLength();
            if (l==0) error("arbitrary axis of rotation is null-vector\n");
            a /= l;

            float c = cos(r->angle);
            float s = sin(r->angle);
            float t = 1-c;

            astate->currTransMatrix =
                Matrix(t*a.x*a.x+c,     t*a.x*a.y+s*a.z, t*a.x*a.z-s*a.y, 0,
                       t*a.x*a.y-s*a.z, t*a.y*a.y+c,     t*a.y*a.z+s*a.x, 0,
                       t*a.x*a.z+s*a.y, t*a.y*a.z-s*a.x, t*a.z*a.z+c,     0,
                       0,               0,               0,               1) *
                astate->currTransMatrix;
        }
    }

    /*
    for (int row=0; row<4; row++)
    {
        message("","%f %f %f %f\n",
            astate->currTransMatrix.e[row][0],
            astate->currTransMatrix.e[row][1],
            astate->currTransMatrix.e[row][2],
            astate->currTransMatrix.e[row][3]);
    }
    */
}

//----------------------------------------------------------------------------
//   class ScaleNode : Node
//----------------------------------------------------------------------------
//used for VRML-node Scale - overrides setupDepends

//---ScaleNode::ScaleNode-----------------------------------------------------
// Input : int aid                  : init
//         char *afieldspec         : init
//         Field *afield            : init
//         int fline                : init
//         int lline                : init
// Output:

ScaleNode::ScaleNode(int aid, char *afieldspec, Field *afield, int fline, int lline) :
           Node(aid,afieldspec,afield,fline,lline)
{
}

//---ScaleNode::setupDepends--------------------------------------------------
// Input : State *astate  : current state
// Output: State *satate  : new state
// The Field with field-id FID_SCALEFACTOR defines a 3D scaling about the
// origin. The scale-operation is accumulated into the current transformation: 
// astate->currTransMatrix = scale-matrix * astate->currTransMatrix

void ScaleNode::setupDepends(State *astate)
{
    Field *scaleFactor = getField(FID_SCALEFACTOR);
    if (scaleFactor!=NULL)
    {
        FVec3f *s = (FVec3f *)scaleFactor->value;

        astate->currTransMatrix =
            Matrix(s->x, 0,    0,    0,
                   0,    s->y, 0,    0,
                   0,    0,    s->z, 0,
                   0,    0,    0,    1) *
            astate->currTransMatrix;
    }

    /*
    for (int row=0; row<4; row++)
    {
        message("","%f %f %f %f\n",
            astate->currTransMatrix.e[row][0],
            astate->currTransMatrix.e[row][1],
            astate->currTransMatrix.e[row][2],
            astate->currTransMatrix.e[row][3]);
    }
    */
}

//----------------------------------------------------------------------------
//   class TransformNode : Node
//----------------------------------------------------------------------------
//used for VRML-node Transform - overrides setupDepends

//---TransformNode::TransformNode---------------------------------------------
// Input : int aid                  : init
//         char *afieldspec         : init
//         Field *afield            : init
//         int fline                : init
//         int lline                : init
// Output:

TransformNode::TransformNode(int aid, char *afieldspec, Field *afield,
                             int fline, int lline) :
               Node(aid,afieldspec,afield,fline,lline)
{
}

//---TransformNode::setupDepends----------------------------------------------
// Input : State *astate  : current state
// Output: State *satate  : new state
//This fields of this node (FID_TRANSLATION,FID_ROTATION,FID_SCALEFACTOR,
//FID_SCALEORIENTATION,FID_CENTER) defines a geometric 3D transformation
//consisting of (in order) a (possibly) non-uniform scale about an arbitrary
//point, a rotation about an arbitrary point and axis, and a translation.
//The operations are accumulated into the current transformation:
//
//Transform {
//    translation T1
//    rotation R1
//    scaleFactor S
//    scaleOrientation R2
//    center T2
//}
//
// astate->currTransMatrix = (-T2)*(-R2)*S*R2*R1*T2*T1*astate->currTransMatrix

void TransformNode::setupDepends(State *astate)
{
    Field *translation = getField(FID_TRANSLATION);
    if (translation)
    {
        FVec3f *t = (FVec3f *)translation->value;
        
        astate->currTransMatrix =
            Matrix(1,    0,    0,    0,
                   0,    1,    0,    0,
                   0,    0,    1,    0,
                   t->x, t->y, t->z, 1) *
            astate->currTransMatrix;
    }

    //----------

    Field *center = getField(FID_CENTER);
    if (center)
    {
        FVec3f *c = (FVec3f *)center->value;
        
        astate->currTransMatrix =
            Matrix(1,    0,    0,    0,
                   0,    1,    0,    0,
                   0,    0,    1,    0,
                   c->x, c->y, c->z, 1) *
            astate->currTransMatrix;
    }

    //----------

    Field *rotation = getField(FID_ROTATION);
    if (rotation)
    {
        FRotation *r = (FRotation *)rotation->value;
        if (r->angle!=0)
        {
            Vec a = Vec(r->x, r->y, r->z);
            float l = a.getLength();
            if (l==0) error("arbitrary axis of rotation is null-vector\n");
            a /= l;

            float c = cos(r->angle);
            float s = sin(r->angle);
            float t = 1-c;

            astate->currTransMatrix =
                Matrix(t*a.x*a.x+c,     t*a.x*a.y+s*a.z, t*a.x*a.z-s*a.y, 0,
                       t*a.x*a.y-s*a.z, t*a.y*a.y+c,     t*a.y*a.z+s*a.x, 0,
                       t*a.x*a.z+s*a.y, t*a.y*a.z-s*a.x, t*a.z*a.z+c,     0,
                       0,               0,               0,               1) *
                astate->currTransMatrix;
        }
    }

    //----------

    Field *scalefactor = getField(FID_SCALEFACTOR);
    if (scalefactor)
    {
        FVec3f *s = (FVec3f *)scalefactor->value;
        if (s->x!=1 || s->y!=1 || s->z!=1)
        {
            Field *scaleorientation = getField(FID_SCALEORIENTATION);
            if (scaleorientation)
            {
                FRotation *r = (FRotation *)scaleorientation->value;
                if (r->angle!=0)
                {
                    Vec a = Vec(r->x, r->y, r->z);
                    float l = a.getLength();
                    if (l==0) error("arbitrary axis of scaleorientation is null-vector\n");
                    a /= l;

                    float c = cos(r->angle);
                    float s = sin(r->angle);
                    float t = 1-c;

                    astate->currTransMatrix =
                        Matrix(t*a.x*a.x+c,     t*a.x*a.y+s*a.z, t*a.x*a.z-s*a.y, 0,
                               t*a.x*a.y-s*a.z, t*a.y*a.y+c,     t*a.y*a.z+s*a.x, 0,
                               t*a.x*a.z+s*a.y, t*a.y*a.z-s*a.x, t*a.z*a.z+c,     0,
                               0,               0,               0,               1) *
                        astate->currTransMatrix;

                }
            }

            //---

            astate->currTransMatrix =
                Matrix(s->x, 0,    0,    0,
                       0,    s->y, 0,    0,
                       0,    0,    s->z, 0,
                       0,    0,    0,    1) *
                astate->currTransMatrix;
 
            //---

            if (scaleorientation)
            {
                FRotation *r = (FRotation *)scaleorientation->value;
                if (r->angle!=0)
                {
                    Vec a = Vec(r->x, r->y, r->z);
                    a /= a.getLength();

                    float c = cos(-r->angle);
                    float s = sin(-r->angle);
                    float t = 1-c;

                    astate->currTransMatrix =
                        Matrix(t*a.x*a.x+c,     t*a.x*a.y+s*a.z, t*a.x*a.z-s*a.y, 0,
                               t*a.x*a.y-s*a.z, t*a.y*a.y+c,     t*a.y*a.z+s*a.x, 0,
                               t*a.x*a.z+s*a.y, t*a.y*a.z-s*a.x, t*a.z*a.z+c,     0,
                               0,               0,               0,               1) *
                        astate->currTransMatrix;
                }
            }
        }
    }

    //----------

    if (center)
    {
        FVec3f *c = (FVec3f *)center->value;
        
        astate->currTransMatrix =
            Matrix(1,     0,     0,     0,
                   0,     1,     0,     0,
                   0,     0,     1,     0,
                   -c->x, -c->y, -c->z, 1) *
            astate->currTransMatrix;
    }

    /*
    for (int row=0; row<4; row++)
    {
        message("","%f %f %f %f\n",
            astate->currTransMatrix.e[row][0],
            astate->currTransMatrix.e[row][1],
            astate->currTransMatrix.e[row][2],
            astate->currTransMatrix.e[row][3]);
    }
    */
}

//----------------------------------------------------------------------------
//   class MatrixTransformNode : Node
//----------------------------------------------------------------------------
//used for VRML-node MatrixTransform

//---MatrixTransformNode::MatrixTransformNode---------------------------------
// Input : int aid                  : init
//         char *afieldspec         : init
//         Field *afield            : init
//         int fline                : init
//         int lline                : init
// Output:

MatrixTransformNode::MatrixTransformNode(int aid, char *afieldspec,
                                         Field *afield, int fline, int lline) :
                     Node(aid,afieldspec,afield,fline,lline)
{
}

//---MatrixTransformNode::setupDepends----------------------------------------
// Input : State *astate  : current state
// Output: State *astate  : new state
// The field with field-id FID_MATRIX defines a geometric 3D transformation
// with a 4 by 4 matrix. The matrix is accumulated into the current
// transformation: 
// astate->currTransMatrix = matrix * astate->currTransMatrix

void MatrixTransformNode::setupDepends(State *astate)
{
    Field *matrix = getField(FID_MATRIX);
    if (matrix)
    {
        FMatrix *m = (FMatrix *)matrix->value;

        astate->currTransMatrix =
            Matrix(m->number11,m->number12,m->number13,m->number14,
                   m->number21,m->number22,m->number23,m->number24,
                   m->number31,m->number32,m->number33,m->number34,
                   m->number41,m->number42,m->number43,m->number44) *
            astate->currTransMatrix;
    }

    /*
    for (int row=0; row<4; row++)
    {
        message("","%f %f %f %f\n",
            astate->currTransMatrix.e[row][0],
            astate->currTransMatrix.e[row][1],
            astate->currTransMatrix.e[row][2],
            astate->currTransMatrix.e[row][3]);
    }
    */
}

//----------------------------------------------------------------------------
//   class TranslationNode : Node
//----------------------------------------------------------------------------
//used for VRML-node Tanslation

//---TranslationNode::TranslationNode-----------------------------------------
// Input : int aid                  : init
//         char *afieldspec         : init
//         Field *afield            : init
//         int fline                : init
//         int lline                : init
// Output:

TranslationNode::TranslationNode(int aid, char *afieldspec, Field *afield,
                                 int fline, int lline) :
                 Node(aid,afieldspec,afield,fline,lline)
{
}

//---TranslationNode::setupDepends--------------------------------------------
// Input : State *astate  : current state
// Output: State *astate  : new state
// The Field with field-id FID_TRANSLATION defines a translation by a
// 3D vector. The translation is accumulated into the current transformation: 
// astate->currTransMatrix = translation-matrix*astate->currTransMatrix

void TranslationNode::setupDepends(State *astate)
{
    Field *translation = getField(FID_TRANSLATION);
    if (translation)
    {
        FVec3f *t = (FVec3f *)translation->value;
        
        astate->currTransMatrix =
            Matrix(1,    0,    0,    0,
                   0,    1,    0,    0,
                   0,    0,    1,    0,
                   t->x, t->y, t->z, 1) *
            astate->currTransMatrix;
     }

    /*
    for (int row=0; row<4; row++)
    {
        message("","%f %f %f %f\n",
            astate->currTransMatrix.e[row][0],
            astate->currTransMatrix.e[row][1],
            astate->currTransMatrix.e[row][2],
            astate->currTransMatrix.e[row][3]);
    }
    */
}

//----------------------------------------------------------------------------
