//  Copyright 1996 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       tri
//  TYPE:       c++ header
//  PROJECT:    lodestar
//  CONTENT:    class Triangulator
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rs      Reinhard Sainitzer
//		hb      Herbert Buchegger
// ===========================================================================
//  HISTORY:
//
// ===========================================================================
//  MANUAL:     The class Triangulator is a 3d extended version of the source
//              published in Graphic Gems V (Seidel's algorithmn). This is the
//              header file for tri.C and tri2.C. In tri2.C is the oo version
//              of the Graphic Gems source. In tri.C is the 3d extension.
//              
// ===========================================================================

#ifndef __TRI_H
#define __TRI_H

//---includes-------------------------------------------------------------------

#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

//------------------------------------------------------------------------------
// type definitions, used from Graphic Gems source

typedef struct {
  double x, y;
} point_t,vector_t;

typedef struct {
  point_t v0, v1;
  int is_inserted;
  int root0, root1;
} segment_t;

typedef struct {
  int lseg, rseg;
  point_t hi, lo;
  int u0, u1;
  int d0, d1;
  int sink;
  int usave, uside;
  int state;
} trap_t;

typedef struct {
  int nodetype;
  int segnum;
  point_t yval;
  int trnum;
  int parent;
  int left, right;
} node_t;

typedef struct {
  int vnum;
  int next;
  int prev;
} monchain_t;

typedef struct {
  point_t pt;
  int vnext[4];			/* next vertices for the 4 chains */
  int vpos[4];			/* position of v in the 4 chains */
  int nextfree;
} vertexchain_t;

struct global_s {
  int nseg;
};

#define QSIZE   800		/* maximum table sizes */
#define TRSIZE  400		/* max# trapezoids */
#define SEGSIZE 100		/* max# of segments */

//------------------------------------------------------------------------------
// the 5937 definition of a 3d vector

struct VecXYZ
{
    double x,y,z;
};

//------------------------------------------------------------------------------
// return values for checkCrossing function

#define C_NO_CROSSING 0
#define C_CROSSING    1
#define C_PARALLEL    2

//------------------------------------------------------------------------------
// Triangulator
//------------------------------------------------------------------------------
// Manual:
// To triangulate a set of polygons, you must create a Triangulator object
// and use the addPoint function to set the vertices of one polygon into the
// vec3d array. To triangulate, call the run() function. If the
// return value is TRUE, you can read the result with the getPoint() functions.
// To triangulate the next polygon, call reset() and use addPoint(), run(), 
// getPoint() and for the next polygon call reset() and use addPoint(), ...
// And when you are ready you can delete the object.
// Description:
// To use the 2d triangulator, I must reduce the dimension from 3 to 2. I 
// determinate a correct x,y pair with the function checkForXYZ() and set the
// xok, yok and zok variables to TRUE or FALSE. If xok and yok, I have found
// a correct pair and can use this new 2D polygon. Now I can determinate the 
// clock wise order of the polygon and reorder the vertices if it is necessary.
// When I have a 2d list with the right clock wise order and can use the 2d 
// triangulator.

class Triangulator
{
public:
    static int verbose;             // debug messages on / off

    Triangulator();
    ~Triangulator();

    void reset();                   // the constructor call this automatically
    void addPoint(double x,double y,double z);
    void getPoint(int row,int col,double & x,double & y,double & z);
    int getPoint(int row,int col);
    int run();
    void printList();               // prints debug message to stdout
    void print2DList();             // prints debug message and write the file tri_2D
                                    // you can use this file and gnuplot to draw the 2d polygon
    void print3DList();             // prints debug message and write the file tri_3D
                                    // you can use this file and gnuplot to draw the 3d polygon
    void printResult();
    int getCount();                 // result count of triangles
    int checkForXYZ();              // checks if the triangulator is able to handle
                                    // the polygon
protected:
    VecXYZ vec3D[SEGSIZE];          // store the 3d vertices list
    int count;                      // count of vertices

    int curr;                       // an index variable
    int clockwise;                  // the polygon is in clock wise order or not
                                    // clockWiseOrder determinate this
    int error;
    int xok,yok,zok;

    // variables from the Graphic Gems source code
    node_t qs[QSIZE];	              /* Query structure */
    trap_t tr[TRSIZE];		  /* Trapezoid structure */
    segment_t seg[SEGSIZE];		/* Segment table */
    int triangles[SEGSIZE+2][3];
    struct global_s global;

    monchain_t mchain[TRSIZE];

    vertexchain_t vert[SEGSIZE];
    int mon[SEGSIZE];
    int visited[TRSIZE];
    int chain_idx, op_idx, mon_idx;

    int q_idx;
    int tr_idx;

    int triangulate();              // called from run();
    void addPoint(double x,double y); // add a 2d point
    int clockWiseOrder();
    int checkCrossing(double p1x,double p1y,double p12x,double p12y,
                      double p2x,double p2y,double p22x,double p22y);

    // functions from the Graphic Gems source code
    int generate_random_ordering(int n);
    int choose_segment();
    int inserted(int segnum,int whichpt);
    int read_segments(FILE * infile);
    int math_logstar_n(int n);
    int math_N(int n,int h);

    int inside_polygon(trap_t * t);
    int newmon();
    int new_chain_element();
    double get_angle(point_t * vp0,point_t * vpnext,point_t * vp1);
    int get_vertex_positions(int v0,int v1,int * ip,int * iq);
    int make_new_monotone_poly(int mcur,int v0,int v1);
    int monotonate_trapezoids(int n);
    int traverse_polygon(int mcur,int trnum,int from,int dir);
    void triangulate_monotone_polygons(int nmonpoly,int op[][3]);
    int triangulate_single_polygon(int posmax,int side,int op[][3]);

    int newnode();
    int newtrap();
    int _max(point_t * yval,point_t * v0,point_t * v1);
    int _min(point_t *yval,point_t * v0,point_t * v1);
    int _greater_than(point_t * v0,point_t * v1);
    int _equal_to(point_t * v0,point_t * v1);
    int _greater_than_equal_to(point_t * v0,point_t * v1);
    int _less_than(point_t * v0,point_t * v1);
    int init_query_structure(int segnum);
    int is_left_of(int segnum,point_t * v);
    int locate_endpoint(point_t * v,point_t * vo,int r);
    int merge_trapezoids(int segnum,int tfirst,int tlast,int side);
    int add_segment(int segnum);
    int find_new_roots(int segnum);
    int construct_trapezoids(int nseg,segment_t * seg);
};


#endif
