// ===========================================================================
//  Copyright 1996 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       yacc
//  TYPE:       yacc parser
//  PROJECT:    lodestar
//  CONTENT:    yacc parser for VRML files
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rs      Reinhard Sainitzer
//              hb      Herbert Buchegger
// ===========================================================================
//  HISTORY:
//  based on a yacc parser written by Emmanuel Frcon and Olof Hagsand
//
// ===========================================================================
//  MANUAL:
//
// ===========================================================================

%{

#include "nodes.h"
#include "fields.h"
#include "enum.h"
#include <stdio.h>

#define YYMAXDEPTH 100000

extern "C++" int yylex(void);
extern int lex_line;
extern int lex_idline;

void yyerror(const char *s);

/*==========================================================================
  variables
  ==========================================================================*/

Node *yacc_tree = NULL; //root of the complete parse tree
int yacc_nolods = 0;    //if a LOD-node or a LevelOfDetail-node is encountered
                        //and yacc_nolods=0 and error ocurres 
                        //else this nodes will be treated as GroupNodes


static Field * lastfield;  //help-variable for setting up Field-lists
static Value * lastvalue;  //help-variable for setting up Value-lists

%} 


%union {
  int intval;
  float floatval;
  double doubleval;
  char *string;
  Node *node;
  Field *field;
  Value *value;
  int line;
}


%token <line> ASCIITEXT CONE COORDINATE3 CUBE CYLINDER DIRECTIONALLIGHT FONTSTYLE
%token <line> GROUP INDEXEDFACESET INDEXEDLINESET INFO LOD MATERIAL MATERIALBINDING
%token <line> LEVELOFDETAIL MATRIXTRANSFORM
%token <line> NORM NORMALBINDING ORTHOGRAPHICCAMERA PERSPECTIVECAMERA
%token <line> POINTLIGHT POINTSET ROTATION SCALE SEPARATOR SHAPEHINTS SPHERE
%token <line> SPOTLIGHT SWITCH TEXTURE2 TEXTURE2TRANSFORM TEXTURECOORDINATE2
%token <line> TRANSFORM TRANSFORMSEPARATOR TRANSLATION WWWANCHOR WWWINLINE

%token PARTS BOTTOMRADIUS HEIGHT POINT WIDTH DEPTH STRING SPACING
%token JUSTIFICATION RADIUS ON INTENSITY COLOR DIRECTION SIZE FAMILY STYLE
%token COORDINDEX MATERIALINDEX NORMALINDEX TEXTURECOORDINDEX RANGE
%token SCREENAREA TRANSLATION_F
%token CENTER AMBIENTCOLOR DIFFUSECOLOR SPECULARCOLOR EMISSIVECOLOR SHININESS
%token TRANSPARENCY VALUE MATRIX VECTOR POSITION ORIENTATION FOCALDISTANCE
%token HEIGHTANGLE LOCATION STARTINDEX NUMPOINTS ROTATION_F SCALEFACTOR
%token RENDERCULLING VERTEXORDERING SHAPETYPE FACETYPE CREASEANGLE DROPOFFRATE
%token CUTOFFANGLE WHICHCHILD FILENAME IMAGE WRAPS WRAPT SCALEORIENTATION
%token NAME DESCRIPTION MAP BBOXSIZE BBOXCENTER

%token DEF
%token <line> USE

%token FIELDS

%token <intval> NUMBER
%token <floatval> FLOAT
%token <string> VRMLSTRING VRMLWORD FIELDSPEC UNKNOWNNODE UNKNOWNNODELIST


%type <floatval> FloatVal
%type <intval> bitmasklist
%type <string> NodeSpec FieldSpecs

%type <node> Nodes Node NodeBody
%type <node> AsciiText Cone Coordinate3 Cube Cylinder Dlight FontStyle Group
%type <node> IdxFaceSet IdxLineSet Info Lod LevelOfDetail Material
%type <node> MaterialBinding MatrixTransform Normal NormalBinding
%type <node> OrthographicCamera PerspectiveCamera PointLight PointSet Rotation
%type <node> Scale Separator ShapeHints Sphere SpotLight Switch Texture2
%type <node> Tex2Transform TexCoordinate2 Transform TransSeparator Translation
%type <node> WWWAnchor WWWInline UnknownNode

%type <field> asciitextfield asciitextfields
%type <field> conefield conefields
%type <field> coordinatefield coordinatefields
%type <field> cubefield cubefields
%type <field> cylinderfield cylinderfields
%type <field> dlightfield dlightfields
%type <field> fstylefield fstylefields
%type <field> ifacesetfield ifacesetfields
%type <field> ilinesetfield ilinesetfields
%type <field> lodfield lodfields
%type <field> levofdetfield levofdetfields
%type <field> infofield infofields
%type <field> materialfield materialfields
%type <field> mbindingfield mbindingfields
%type <field> mxtransformfield mxtransformfields
%type <field> normalfield normalfields
%type <field> nbindingfield nbindingfields
%type <field> ocamerafield ocamerafields
%type <field> pcamerafield pcamerafields
%type <field> plightfield plightfields
%type <field> psetfield psetfields
%type <field> rotationfield rotationfields
%type <field> scalefield scalefields
%type <field> separatorfield separatorfields
%type <field> shapehintsfield shapehintsfields
%type <field> spherefield spherefields
%type <field> spotlightfield spotlightfields
%type <field> switchfield switchfields
%type <field> texture2field texture2fields
%type <field> tex2transfield tex2transfields
%type <field> texcoord2field texcoord2fields
%type <field> transformfield transformfields
%type <field> translationfield translationfields
%type <field> WWWanchorfield WWWanchorfields
%type <field> WWWinlinefield WWWinlinefields

%type <value> SFBitMask SFBool SFColor SFEnum SFFloat SFImage SFLong
%type <value> SFMatrix SFRotation SFString SFVec2f SFVec3f
%type <value> MFColor MFLong MFFloat MFString MFVec2f MFVec3f
%type <value> imglist colorlist longlist floatlist
%type <value> stringlist vec2flist vec3flist


%start vrml


%%
/*****************************************************************************/
/* syntactical definitions
/*****************************************************************************/

vrml		: Nodes { yacc_tree = $1;  YYACCEPT;}
		| error {YYABORT;}
		;

Nodes		: Nodes Node
                  {
                      if ($1==NULL) { $2->last=$2; $$=$2; }
                      else { $1->last->next=$2; $1->last=$2; }
                  }
		| { $$=NULL; }
		;

Node		: NodeSpec NodeBody 	{ $2->nodespec = $1; $$ = $2; $$->registerName(); }
		| USE VRMLWORD		{ $$ = new UseNode($2,$1,$1); }
		;

NodeSpec	: DEF VRMLWORD		{ $$ = $2; }
		|			{ $$ = NULL; }
		;

NodeBody	: AsciiText		{ $$ = $1; }
		| Cone			{ $$ = $1; }
		| Coordinate3		{ $$ = $1; }
		| Cube			{ $$ = $1; }
		| Cylinder		{ $$ = $1; }
		| Dlight		{ $$ = $1; }
		| FontStyle		{ $$ = $1; }
		| Group			{ $$ = $1; }
		| IdxFaceSet		{ $$ = $1; }
		| IdxLineSet		{ $$ = $1; }
		| Info			{ $$ = $1; }
		| Lod			{ $$ = $1; }
		| LevelOfDetail		{ $$ = $1; }
		| Material		{ $$ = $1; }
		| MaterialBinding	{ $$ = $1; }
		| MatrixTransform	{ $$ = $1; }
		| Normal		{ $$ = $1; }
		| NormalBinding		{ $$ = $1; }
		| OrthographicCamera	{ $$ = $1; }
		| PerspectiveCamera	{ $$ = $1; }
		| PointLight		{ $$ = $1; }
		| PointSet		{ $$ = $1; }
		| Rotation		{ $$ = $1; }
		| Scale			{ $$ = $1; }
		| Separator		{ $$ = $1; }
		| ShapeHints		{ $$ = $1; }
		| Sphere		{ $$ = $1; }
		| SpotLight		{ $$ = $1; }
		| Switch		{ $$ = $1; }
		| Texture2		{ $$ = $1; }
		| Tex2Transform		{ $$ = $1; }
		| TexCoordinate2	{ $$ = $1; }
		| Transform		{ $$ = $1; }
		| TransSeparator	{ $$ = $1; }
		| Translation		{ $$ = $1; }
		| WWWAnchor		{ $$ = $1; }
		| WWWInline		{ $$ = $1; }
		| UnknownNode		{ $$ = $1; }
		;		

/* Field specification */
FieldSpecs	: FIELDS FIELDSPEC	{ $$ = $2; }
		|			{ $$ = NULL; }
		;

/* Unknown node */
UnknownNode	: UNKNOWNNODE UNKNOWNNODELIST	{ $$ = new UnknownNode($1,$2,lex_idline,lex_line); } 
		;

/* Ascii Text */
asciitextfields	: asciitextfields asciitextfield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

asciitextfield	: STRING	MFString	{ $$ = new Field(FID_STRING,$2); }
		| SPACING	SFFloat		{ $$ = new Field(FID_SPACING,$2); }
		| JUSTIFICATION	SFEnum		{ $$ = new Field(FID_JUSTIFICATION,$2); ((FEnum *)$2)->type=FTjustification; }
		| WIDTH		MFFloat		{ $$ = new Field(FID_WIDTH,$2); }
		;
		
AsciiText	: ASCIITEXT '{' FieldSpecs asciitextfields '}'	{ $$ = new Node(NID_ASCIITEXT,$3,$4,$1,lex_line); }
		;

/* Cones */
conefields	: conefields conefield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

conefield	: PARTS		SFBitMask	{ $$ = new Field(FID_PARTS,$2); ((FBitMask *)$2)->type=FTparts; }
		| BOTTOMRADIUS	SFFloat		{ $$ = new Field(FID_BOTTOMRADIUS,$2); }
		| HEIGHT	SFFloat		{ $$ = new Field(FID_HEIGHT,$2); }
		;

Cone		: CONE '{' FieldSpecs conefields '}'  { $$ = new ConeNode(NID_CONE,$3,$4,$1,lex_line); }
		;


/* Coordinate 3 */
coordinatefields: coordinatefields coordinatefield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

coordinatefield :  POINT MFVec3f		{ $$ = new Field(FID_POINT,$2); }
		;

Coordinate3	: COORDINATE3 '{' FieldSpecs coordinatefields '}'  { $$ = new CoordNode(NID_COORDINATE3,$3,$4,$1,lex_line); }
		;

/* Cubes */
cubefields	: cubefields cubefield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

cubefield	: WIDTH		SFFloat		{ $$ = new Field(FID_WIDTH,$2); }
		| HEIGHT	SFFloat		{ $$ = new Field(FID_HEIGHT,$2); }
		| DEPTH		SFFloat		{ $$ = new Field(FID_DEPTH,$2); }
		;

Cube		: CUBE '{' FieldSpecs cubefields '}'  { $$ = new CubeNode(NID_CUBE,$3,$4,$1,lex_line); }
		;


/* Cylinder */
cylinderfields	: cylinderfields cylinderfield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

cylinderfield	: PARTS		SFBitMask	{ $$ = new Field(FID_PARTS,$2); ((FBitMask *)$2)->type=FTparts; }
		| RADIUS	SFFloat		{ $$ = new Field(FID_RADIUS,$2); }
		| HEIGHT	SFFloat		{ $$ = new Field(FID_HEIGHT,$2); }
		;

Cylinder	: CYLINDER '{' FieldSpecs cylinderfields '}'  { $$ = new CylinderNode(NID_CYLINDER,$3,$4,$1,lex_line); }
		;

/* Directional Light */
dlightfields	: dlightfields dlightfield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

dlightfield	: ON		SFBool		{ $$ = new Field(FID_ON,$2); }
		| INTENSITY	SFFloat		{ $$ = new Field(FID_INTENSITY,$2); }
		| COLOR		SFColor		{ $$ = new Field(FID_COLOR,$2); }
		| DIRECTION	SFVec3f		{ $$ = new Field(FID_DIRECTION,$2); }
		;

Dlight		: DIRECTIONALLIGHT '{' FieldSpecs dlightfields '}'  { $$ = new Node(NID_DIRECTIONALLIGHT,$3,$4,$1,lex_line); }
		;

/* Font Style */
fstylefields	: fstylefields fstylefield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

fstylefield	: SIZE		SFFloat		{ $$ = new Field(FID_SIZE,$2); }
		| FAMILY	SFEnum		{ $$ = new Field(FID_FAMILY,$2); ((FEnum *)$2)->type=FTfamily; }
		| STYLE		SFBitMask	{ $$ = new Field(FID_STYLE,$2); }
		;

FontStyle	: FONTSTYLE '{' FieldSpecs fstylefields '}'  { $$ = new Node(NID_FONTSTYLE,$3,$4,$1,lex_line); }
		;

/* Groups */
Group		: GROUP '{' Nodes '}'	{ $$ = new GroupNode(NID_GROUP,NULL,NULL,$3,$1,lex_line); }
		;

/* Indexed Face Set */
ifacesetfields	: ifacesetfields ifacesetfield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

ifacesetfield	: COORDINDEX		MFLong	{ $$ = new Field(FID_COORDINDEX,$2); }
		| MATERIALINDEX		MFLong	{ $$ = new Field(FID_MATERIALINDEX,$2); }
		| NORMALINDEX		MFLong	{ $$ = new Field(FID_NORMALINDEX,$2); }
		| TEXTURECOORDINDEX	MFLong	{ $$ = new Field(FID_TEXTURECOORDINDEX,$2); }
		;

IdxFaceSet	: INDEXEDFACESET '{' FieldSpecs ifacesetfields '}'  { $$ = new FaceNode(NID_INDEXEDFACESET,$3,$4,$1,lex_line); }
		;

/* Indexed Line Set */
ilinesetfields	: ilinesetfields ilinesetfield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

ilinesetfield	: COORDINDEX		MFLong	{ $$ = new Field(FID_COORDINDEX,$2); }
		| MATERIALINDEX		MFLong	{ $$ = new Field(FID_MATERIALINDEX,$2); }
		| NORMALINDEX		MFLong	{ $$ = new Field(FID_NORMALINDEX,$2); }
		| TEXTURECOORDINDEX	MFLong	{ $$ = new Field(FID_TEXTURECOORDINDEX,$2); }
		;

IdxLineSet	: INDEXEDLINESET '{' FieldSpecs ilinesetfields '}'  { $$ = new LineNode(NID_INDEXEDLINESET,$3,$4,$1,lex_line); }
		;

/* Info */
infofields	: infofields infofield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

infofield	: STRING	SFString	{ $$ = new Field(FID_STRING,$2); }
		;

Info		: INFO '{' FieldSpecs infofields '}'  { $$ = new Node(NID_INFO,$3,$4,$1,lex_line); }
		;

/* LOD */ 
lodfields	: lodfields lodfield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

lodfield 	: RANGE		MFFloat { $$ = new Field(FID_RANGE,$2); }
                | CENTER	MFVec3f { $$ = new Field(FID_CENTER,$2); }
		;

Lod		: LOD '{' FieldSpecs lodfields Nodes '}'  
                      { 
                          if (yacc_nolods) $$ = new GroupNode(NID_LOD,$3,$4,$5,$1,lex_line);
                          else
                          { 
                      	      fprintf(stderr,"error: this file already contains Levels of Detail\n"); 
			      YYABORT;
                          }
                      }
		;

/* LevelOfDetail */
levofdetfields	: levofdetfields levofdetfield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

levofdetfield 	: SCREENAREA		MFFloat { $$ = new Field(FID_SCREENAREA,$2); }
		| CENTER		MFVec3f { $$ = new Field(FID_CENTER,$2); }
		;

LevelOfDetail	: LEVELOFDETAIL '{' FieldSpecs levofdetfields Nodes '}'  
                      { 
                          if (yacc_nolods) $$ = new GroupNode(NID_LEVELOFDETAIL,$3,$4,$5,$1,lex_line);
                          else
                          { 
                      	      fprintf(stderr,"error: this file already contains Levels of Detail\n"); 
			      YYABORT;
                          }
                      }
		;

/* Material */
materialfields	: materialfields materialfield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

materialfield	: AMBIENTCOLOR		MFColor	{ $$ = new Field(FID_AMBIENTCOLOR,$2); }
		| DIFFUSECOLOR		MFColor	{ $$ = new Field(FID_DIFFUSECOLOR,$2); }
		| SPECULARCOLOR		MFColor	{ $$ = new Field(FID_SPECULARCOLOR,$2); }
		| EMISSIVECOLOR		MFColor	{ $$ = new Field(FID_EMISSIVECOLOR,$2); }
		| SHININESS		MFFloat	{ $$ = new Field(FID_SHININESS,$2); }
		| TRANSPARENCY		MFFloat	{ $$ = new Field(FID_TRANSPARENCY,$2); }
		;

Material	: MATERIAL '{' FieldSpecs materialfields '}'  { $$ = new MaterialNode(NID_MATERIAL,$3,$4,$1,lex_line); }
		;

/* Material Binding */
mbindingfields	: mbindingfields mbindingfield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

mbindingfield	: VALUE			SFEnum	{ $$ = new Field(FID_VALUE,$2); ((FEnum *)$2)->type=FTvalue; }
		;

MaterialBinding	: MATERIALBINDING '{' FieldSpecs mbindingfields '}'  { $$ = new MaterialBindingNode(NID_MATERIALBINDING,$3,$4,$1,lex_line); }
		;

/* Matrix Transform */
mxtransformfields	: mxtransformfields mxtransformfield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
			| { lastfield=NULL; $$=NULL; }
			;

mxtransformfield	: MATRIX	SFMatrix  { $$ = new Field(FID_MATRIX,$2); }
			;

MatrixTransform	: MATRIXTRANSFORM '{' FieldSpecs mxtransformfields '}'  { $$ = new MatrixTransformNode(NID_MATRIXTRANSFORM,$3,$4,$1,lex_line); }
		;

/* Normal */
normalfields	: normalfields normalfield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

normalfield	: VECTOR	MFVec3f		{ $$ = new Field(FID_VECTOR,$2); }
		;

Normal		: NORM '{' FieldSpecs normalfields '}'  { $$ = new NormalNode(NID_NORMAL,$3,$4,$1,lex_line); }
		;

/* Normal Binding */
nbindingfields	: nbindingfields nbindingfield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

nbindingfield	: VALUE			SFEnum	{ $$ = new Field(FID_VALUE,$2); ((FEnum *)$2)->type=FTvalue; }
		;

NormalBinding	: NORMALBINDING '{' FieldSpecs nbindingfields '}'  { $$ = new NormalBindingNode(NID_NORMALBINDING,$3,$4,$1,lex_line); }
		;

/* Orthographic camera */
ocamerafields	: ocamerafields ocamerafield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

ocamerafield	: POSITION	SFVec3f		{ $$ = new Field(FID_POSITION,$2); }
		| ORIENTATION	SFRotation	{ $$ = new Field(FID_ORIENTATION,$2); }
		| FOCALDISTANCE	SFFloat		{ $$ = new Field(FID_FOCALDISTANCE,$2); }
		| HEIGHT	SFFloat		{ $$ = new Field(FID_HEIGHT,$2); }
		;

OrthographicCamera	: ORTHOGRAPHICCAMERA '{' FieldSpecs ocamerafields '}'  { $$ = new OrthographicCameraNode(NID_ORTHOGRAPHICCAMERA,$3,$4,$1,lex_line); }
			;

/* Perspective camera */
pcamerafields	: pcamerafields pcamerafield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

pcamerafield	: POSITION	SFVec3f		{ $$ = new Field(FID_POSITION,$2); }
		| ORIENTATION	SFRotation	{ $$ = new Field(FID_ORIENTATION,$2); }
		| FOCALDISTANCE	SFFloat		{ $$ = new Field(FID_FOCALDISTANCE,$2); }
		| HEIGHTANGLE	SFFloat		{ $$ = new Field(FID_HEIGHTANGLE,$2); }
		;

PerspectiveCamera	: PERSPECTIVECAMERA '{' FieldSpecs pcamerafields '}'  { $$ = new PerspectiveCameraNode(NID_PERSPECTIVECAMERA,$3,$4,$1,lex_line); }
			;

/* Point Light */
plightfields	: plightfields plightfield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

plightfield	: ON		SFBool		{ $$ = new Field(FID_ON,$2); }
		| INTENSITY	SFFloat		{ $$ = new Field(FID_INTENSITY,$2); }
		| COLOR		SFColor		{ $$ = new Field(FID_COLOR,$2); }
		| LOCATION	SFVec3f		{ $$ = new Field(FID_LOCATION,$2); }
		;	

PointLight	: POINTLIGHT '{' FieldSpecs plightfields '}'  { $$ = new Node(NID_POINTLIGHT,$3,$4,$1,lex_line); }
		;

/* Point Set */
psetfields	: psetfields psetfield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

psetfield	: STARTINDEX	SFLong		{ $$ = new Field(FID_STARTINDEX,$2); }
		| NUMPOINTS	SFLong		{ $$ = new Field(FID_NUMPOINTS,$2); }
		;

PointSet	: POINTSET '{' FieldSpecs psetfields '}'  { $$ = new PointNode(NID_POINTSET,$3,$4,$1,lex_line); }
		;

/* Rotation */
rotationfields	: rotationfields rotationfield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

rotationfield	: ROTATION_F	SFRotation	{ $$ = new Field(FID_ROTATION,$2); }
		;

Rotation	: ROTATION '{' FieldSpecs rotationfields '}'  { $$ = new RotationNode(NID_ROTATION,$3,$4,$1,lex_line); }
		;

/* Scale */
scalefields	: scalefields scalefield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

scalefield	: SCALEFACTOR	SFVec3f		{ $$ = new Field(FID_SCALEFACTOR,$2); }
		;

Scale		: SCALE '{' FieldSpecs scalefields '}'  { $$ = new ScaleNode(NID_SCALE,$3,$4,$1,lex_line); }
		;

/* Separator */
separatorfields	: separatorfields separatorfield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

separatorfield	: RENDERCULLING	SFEnum		{ $$ = new Field(FID_RENDERCULLING,$2); ((FEnum *)$2)->type=FTrenderCulling; }
		;

Separator	: SEPARATOR '{' FieldSpecs separatorfields Nodes '}' { $$ = new SeparatorNode(NID_SEPARATOR,$3,$4,$5,$1,lex_line); }
		;

/* Shape Hints */

shapehintsfields	: shapehintsfields shapehintsfield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
			| { lastfield=NULL; $$=NULL; }
			;

shapehintsfield	: VERTEXORDERING	SFEnum	{ $$ = new Field(FID_VERTEXORDERING,$2); ((FEnum *)$2)->type=FTvertexOrdering; }
		| SHAPETYPE		SFEnum	{ $$ = new Field(FID_SHAPETYPE,$2); ((FEnum *)$2)->type=FTshapeType; }
		| FACETYPE		SFEnum	{ $$ = new Field(FID_FACETYPE,$2); ((FEnum *)$2)->type=FTfaceType; }
		| CREASEANGLE		SFFloat	{ $$ = new Field(FID_CREASEANGLE,$2); }
		;

ShapeHints	: SHAPEHINTS '{' FieldSpecs shapehintsfields '}'  { $$ = new ShapeHintsNode(NID_SHAPEHINTS,$3,$4,$1,lex_line); }
		;

/* Sphere */
spherefields	: spherefields spherefield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

spherefield	: RADIUS	SFFloat		{ $$ = new Field(FID_RADIUS,$2); }
		;

Sphere		: SPHERE '{' FieldSpecs spherefields '}'  { $$ = new SphereNode(NID_SPHERE,$3,$4,$1,lex_line); }
		;

/* Spotlight */
spotlightfields	: spotlightfields spotlightfield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

spotlightfield	: ON		SFBool		{ $$ = new Field(FID_ON,$2); }
		| INTENSITY	SFFloat		{ $$ = new Field(FID_INTENSITY,$2); }
		| COLOR		SFColor		{ $$ = new Field(FID_COLOR,$2); }
		| LOCATION	SFVec3f		{ $$ = new Field(FID_LOCATION,$2); }
		| DIRECTION	SFVec3f		{ $$ = new Field(FID_DIRECTION,$2); }
		| DROPOFFRATE	SFFloat		{ $$ = new Field(FID_DROPOFFRATE,$2); }
		| CUTOFFANGLE	SFFloat		{ $$ = new Field(FID_CUTOFFANGLE,$2); }
		;

SpotLight	: SPOTLIGHT '{' FieldSpecs spotlightfields '}'  { $$ = new Node(NID_SPOTLIGHT,$3,$4,$1,lex_line); }
		;

/* Switch */
switchfields	: switchfields switchfield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

switchfield	: WHICHCHILD	SFLong		{ $$ = new Field(FID_WHICHCHILD,$2); }
		;

Switch		: SWITCH '{' FieldSpecs switchfields Nodes '}'  { $$ = new GroupNode(NID_SWITCH,$3,$4,$5,$1,lex_line); }
		;

/* Texture2 */
texture2fields	: texture2fields texture2field { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

texture2field	: FILENAME	SFString	{ $$ = new Field(FID_FILENAME,$2); }
		| IMAGE		SFImage		{ $$ = new Field(FID_IMAGE,$2); }
		| WRAPS		SFEnum		{ $$ = new Field(FID_WRAPS,$2); ((FEnum *)$2)->type=FTwrapS; }
		| WRAPT		SFEnum		{ $$ = new Field(FID_WRAPT,$2); ((FEnum *)$2)->type=FTwrapT; }
		;

Texture2	: TEXTURE2 '{' FieldSpecs texture2fields '}'  { $$ = new TextureNode(NID_TEXTURE2,$3,$4,$1,lex_line); }
		;

/* Texture transformation */
tex2transfields	: tex2transfields tex2transfield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

tex2transfield	: TRANSLATION_F	SFVec2f		{ $$ = new Field(FID_TRANSLATION,$2); }
		| ROTATION_F	SFFloat		{ $$ = new Field(FID_ROTATION,$2); }
		| SCALEFACTOR	SFVec2f		{ $$ = new Field(FID_SCALEFACTOR,$2); }
		| CENTER	SFVec2f		{ $$ = new Field(FID_CENTER,$2); }
		;

Tex2Transform	: TEXTURE2TRANSFORM '{' FieldSpecs tex2transfields '}'  { $$ = new Node(NID_TEXTURE2TRANSFORM,$3,$4,$1,lex_line); }
		;

/* Texcoord2 */
texcoord2fields	: texcoord2fields texcoord2field { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

texcoord2field	: POINT		MFVec2f		{ $$ = new Field(FID_POINT,$2); }
		;

TexCoordinate2	: TEXTURECOORDINATE2 '{' FieldSpecs texcoord2fields '}'  { $$ = new TextureCoordNode(NID_TEXTURECOORDINATE2,$3,$4,$1,lex_line); }
		;

/* Transformation */
transformfields	: transformfields transformfield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

transformfield	: TRANSLATION_F		SFVec3f		{ $$ = new Field(FID_TRANSLATION,$2); }
		| ROTATION_F		SFRotation	{ $$ = new Field(FID_ROTATION,$2); }
		| SCALEFACTOR		SFVec3f		{ $$ = new Field(FID_SCALEFACTOR,$2); }
		| SCALEORIENTATION	SFRotation	{ $$ = new Field(FID_SCALEORIENTATION,$2); }
		| CENTER		SFVec3f		{ $$ = new Field(FID_CENTER,$2); }
		;

Transform	: TRANSFORM '{' FieldSpecs transformfields '}'  { $$ = new TransformNode(NID_TRANSFORM,$3,$4,$1,lex_line); }
		;

/* Transform Separator */
TransSeparator	: TRANSFORMSEPARATOR '{' FieldSpecs Nodes '}'  { $$ = new TransformSeparatorNode(NID_TRANSFORMSEPARATOR,$3,NULL,$4,$1,lex_line); }
		;

/* Translation */
translationfields	: translationfields translationfield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
			| { lastfield=NULL; $$=NULL; }
			;

translationfield	: TRANSLATION_F	SFVec3f  { $$ = new Field(FID_TRANSLATION,$2); }
			;

Translation	: TRANSLATION '{' FieldSpecs translationfields '}'  { $$ = new TranslationNode(NID_TRANSLATION,$3,$4,$1,lex_line); }
		;

/* WWWAnchor */
WWWanchorfields	: WWWanchorfields WWWanchorfield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

WWWanchorfield	: NAME		SFString	{ $$ = new Field(FID_NAME,$2); }
		| DESCRIPTION	SFString	{ $$ = new Field(FID_DESCRIPTION,$2); }
		| MAP		SFEnum		{ $$ = new Field(FID_MAP,$2); ((FEnum *)$2)->type=FTmap; }
		;

WWWAnchor	: WWWANCHOR '{' FieldSpecs WWWanchorfields Nodes '}'  { $$ = new SeparatorNode(NID_WWWANCHOR,$3,$4,$5,$1,lex_line); }
		;

/* WWWinline */
WWWinlinefields	: WWWinlinefields WWWinlinefield { if (lastfield==NULL) { lastfield=$2; $$=$2; } else { lastfield->next=$2; lastfield=$2; } }
		| { lastfield=NULL; $$=NULL; }
		;

WWWinlinefield	: NAME		SFString	{ $$ = new Field(FID_NAME,$2); }
		| BBOXSIZE	SFVec3f 	{ $$ = new Field(FID_BBOXSIZE,$2); }
		| BBOXCENTER	SFVec3f 	{ $$ = new Field(FID_BBOXCENTER,$2); }
		;

WWWInline	: WWWINLINE '{' FieldSpecs WWWinlinefields '}'  { $$ = new Node(NID_WWWINLINE,$3,$4,$1,lex_line); }
		;





/* ========== Classes of Fields ========== */

FloatVal	: FLOAT				{ $$ = $1; }
		| NUMBER			{ $$ = (float)$1; }
		;

/* Single Value Fields */

bitmasklist	: NUMBER			{ $$ = $1; }
		| bitmasklist '|' NUMBER 	{ $$ = ($1 | $3); }
		;

SFBitMask	: NUMBER			{ $$ = new FBitMask($1); }
		| '(' bitmasklist ')'		{ $$ = new FBitMask($2); }
		;

SFBool		: NUMBER			{ $$ = new FBool($1); }
		;

SFColor		: FloatVal FloatVal FloatVal 	{ $$ = new FVec3f($1,$2,$3); }
		;

SFEnum		: NUMBER			{ $$ = new FEnum($1); }
		;

SFFloat		: FloatVal 			{ $$ = new FFloat($1); }
		;

imglist		: NUMBER imglist		{ $$ = new FImage($1,$2); }
		|				{ $$ = NULL; }
		;

SFImage		: NUMBER NUMBER NUMBER imglist	{ $$ = new FImage($1,
						       new FImage($2,
						       new FImage($3,$4))); }
		;

SFLong		: NUMBER			{ $$ = new FLong($1); }
		;

SFMatrix	: FloatVal FloatVal FloatVal FloatVal
		  FloatVal FloatVal FloatVal FloatVal
		  FloatVal FloatVal FloatVal FloatVal
		  FloatVal FloatVal FloatVal FloatVal  { $$ = new FMatrix($1 ,$2 ,$3 ,$4 ,
								          $5 ,$6 ,$7 ,$8 ,
								          $9 ,$10,$11,$12,
								          $13,$14,$15,$16); }
		;

SFRotation	: FloatVal FloatVal FloatVal FloatVal  { $$ = new FRotation($1,$2,$3,$4); }
		;

SFString	: VRMLSTRING			{ $$ = new FString($1); }
		;

SFVec2f		: FloatVal FloatVal		{ $$ = new FVec2f($1,$2); }
		;

SFVec3f		: FloatVal FloatVal FloatVal	{ $$ = new FVec3f($1,$2,$3); }
		;

/* Multiple Value Field */
colorlist	: SFColor			{ $$ = $1; lastvalue=$1; }
		| colorlist ',' SFColor		{ lastvalue->next=$3; lastvalue=$3; }
		;

MFColor		: SFColor 			{ $$ = $1; }
		| '[' colorlist ']'		{ $$ = $2; }
		| '[' colorlist ',' ']'		{ $$ = $2; }
		| '[' ']'			{ $$ = NULL; }
		;

longlist	: SFLong			{ $$ = $1; lastvalue=$1; }
		| longlist ',' SFLong		{ lastvalue->next=$3; lastvalue=$3; }
		;

MFLong		: SFLong			{ $$ = $1; }
		| '[' longlist ']'		{ $$ = $2; }
		| '[' longlist ',' ']'		{ $$ = $2; }
		| '[' ']'			{ $$ = NULL; }
		; 

floatlist	: SFFloat			{ $$ = $1; lastvalue=$1; }
		| floatlist ',' SFFloat		{ lastvalue->next=$3; lastvalue=$3; }
		;

MFFloat		: SFFloat			{ $$ = $1; }
		| '[' floatlist ']'		{ $$ = $2; }
		| '[' floatlist ',' ']'		{ $$ = $2; }
		| '[' ']'			{ $$ = NULL; }
		; 

stringlist	: SFString			{ $$ = $1; lastvalue=$1; }
		| stringlist ',' SFString	{ lastvalue->next=$3; lastvalue=$3; }
		;

MFString	: SFString			{ $$ = $1; }
		| '[' stringlist ']'		{ $$ = $2; }
		| '[' stringlist ',' ']'	{ $$ = $2; }
		| '[' ']'			{ $$ = NULL; }
		; 

vec2flist	: SFVec2f			{ $$ = $1; lastvalue=$1; }
		| vec2flist ',' SFVec2f		{ lastvalue->next=$3; lastvalue=$3; }
		;

MFVec2f		: SFVec2f			{ $$ = $1; }
		| '[' vec2flist ']'		{ $$ = $2; }
		| '[' vec2flist ',' ']'		{ $$ = $2; }
		| '[' ']'			{ $$ = NULL; }
		; 

vec3flist	: SFVec3f			{ $$ = $1; lastvalue=$1; }
		| vec3flist ',' SFVec3f		{ lastvalue->next=$3; lastvalue=$3; }
		;

MFVec3f		: SFVec3f			{ $$ = $1; }
		| '[' vec3flist ']'		{ $$ = $2; }
		| '[' vec3flist ',' ']'		{ $$ = $2; }
		| '[' ']'			{ $$ = NULL; }
		; 

%%

extern char *yytext;

void yyerror(const char *s) 
{ 
  fprintf(stderr, "error: %d: %s at or before '%s'\n", lex_line, s, yytext); 
}

